include("0_common.jl")
include("0_load_spectral.jl")
include("load_spectral_data_InSe.jl")


function apply_gaussian_filter!(As, iT, data)
    # Resolution of 20 meV and 0.03 Å.
    resolution_e = (20.0 * unit_to_aru(:meV)) / (data.ωs_plot[2] - data.ωs_plot[1]) * 1.0
    resolution_k = (0.03 / unit_to_aru(:Å)) / ((data.xplot[end] - data.xplot[1]) / length(data.xplot)) * 1.0
    As[:, 1, :, iT] .= imfilter(As[:, 1, :, iT], Kernel.gaussian((resolution_e, resolution_k)))
    As
end

for blur in [false, true]
    # Plot RS vs Cumulant vs FM vs scFM vs ARPES
    # blur = true
    # blur = false

    material = "InSe"

    data = data_InSe
    iT = 3  # 100 K
    vmax = 5.0


    T_Kelvin = data.Tlist[iT] / unit_to_aru(:K)
    e_unit_type = :eV
    _e_to_eV(e) = (e - data.e_ref) / unit_to_aru(:eV)

    # ------------------------------------------------------------------------------------
    # 2d spectral function

    left   = 0.07
    right  = 0.96
    bottom = 0.11
    top    = 0.96

    fig = figure(figsize=(7.0, 1.6))
    gridspec = fig.add_gridspec(; nrows = 1, ncols = 6, hspace = 0.32, wspace = 0.05,
        left, right, top, bottom, width_ratios = [1., 1., 1., 1., 1., 0.05])
    plotaxes_A_2d = [fig.add_subplot(gridspec[i, j]) for i in 1:1, j in [1,2,3,4,5]]
    cax = fig.add_subplot(gridspec[1, 6])

    ik0 = argmax(data.es_plot[1, :])

    As_list = copy.([data.As_dict["Cumulant"], data.As_dict["FM"], data.As_dict["scFM"]])
    if blur
        apply_gaussian_filter!.(As_list, iT, Ref(data))
    end

    As_labels = ["Cumulant", "\$G_0 D_0\$", "sc\$G D_0\$"]
    plot_spectral(data, plotaxes_A_2d[1:4], nothing; As_list, As_labels, cax, iT, vmax, e_unit_type, ymin, ymax, c_onshell = "b", plot_horizontal_lines = false, ik0, logscale = false, vmin = 0)

    kwargs_A = (; origin="lower", aspect="auto", cmap="Blues", vmin = 0, vmax = vmax)
    ax = plotaxes_A_2d[5]
    inds = xplot_ARPES .<= 0
    extent_ = [extent_ARPES[1], 0, extent_ARPES[3], extent_ARPES[4]]
    img = ax.imshow(z_ARPES[:, inds]; extent = extent_, origin="lower", aspect="auto", kwargs_A...)

    inds = xplot_ARPES .>= 0
    extent_ = [0, extent_ARPES[2], extent_ARPES[3], extent_ARPES[4]]
    img = ax.imshow(z_ARPES[:, inds]; extent = extent_, origin="lower", aspect="auto", kwargs_A...)

    # Arrow indicating gap size
    if blur
        x = -0.25
        y0 = -0.14
        dy = 0.17
        plotaxes_A_2d[3].annotate("", xy=(x, y0), xytext=(x, y0 + dy), arrowprops=Dict("arrowstyle" => "<->", "lw" => 1, "color" => "darkviolet"))
        plotaxes_A_2d[3].text(x + 0.02, y0 - 0.09, "170\$~\$meV"; c = "darkviolet")
    end

    # Plot DFT bands
    for ib in ib_rng
        ax.plot(xplot, _e_to_eV.(data.es_plot[ib, :]), "k--"; lw = 1)
    end

    # 2d plot x labels
    for (iax, ax) in enumerate(plotaxes_A_2d)
        ax.axvline(0; c="k", lw = 0.5, ls="-")

        xk_ΓK = 0.5424990054552199
        xk_ΓM = 0.46981792025202973
        ax.set_xticks([-xk_ΓK/2, 0, xk_ΓM/2])
        ax.set_xticklabels(["K/2", "\$\\Gamma\$", "M/2"])
        ax.set_ylim([-0.5, 0.39])
        ax.set_xlim([-0.38, 0.35])
    end
    for ax in plotaxes_A_2d[2:end]
        ax.set_yticklabels([])
    end

    # Global settings

    plotaxes_A_2d[1].set_ylabel("\$\\varepsilon - \\varepsilon_{\\mathrm{VBM}}\$ (eV)"; labelpad = -1)

    for ax in vcat(plotaxes_A_2d[2:end])
        ax.set_yticklabels([])
    end

    for (iax, ax) in enumerate(plotaxes_A_2d)
        label = ["RS", "Cumulant", "\$G_0 D_0\$", "sc\$G D_0\$", "ARPES"][iax]
        α = iax <= 4 ? 1.0 : 1.0
        ax.text(0.02, 0.99, "($(('a':'z')[iax])) $label"; ha="left", va="top", transform=ax.transAxes, bbox = Dict("facecolor" => "w", "edgecolor" => "none", "pad" => 0.5, "alpha" => α))
    end

    fig.align_xlabels(plotaxes_A_2d)

    if blur
        do_savefig(fig, "./figures/spectral_$material.with_exp.$(T_Kelvin)K"; transparent = false)
    else
        do_savefig(fig, "./figures/spectral_$material.with_exp.no_blur.$(T_Kelvin)K"; transparent = false)
    end

    display(fig); close(fig)
end

begin
    # Plot for FM with self-consistent broadening vs scFM
    material = "InSe"

    data = data_InSe
    iT = 3
    vmax = 5.0

    T_Kelvin = data.Tlist[iT] / unit_to_aru(:K)
    e_unit_type = :eV
    _e_to_eV(e) = (e - data.e_ref) / unit_to_aru(:eV)

    # ------------------------------------------------------------------------------------
    # 2d spectral function

    left   = 0.07 + 0.23
    right  = 0.96 - 0.23
    bottom = 0.11
    top    = 0.96

    fig = figure(figsize=(3.4, 1.6))
    gridspec = fig.add_gridspec(; nrows = 1, ncols = 2, hspace = 0.32, wspace = 0.05,
        left, right, top, bottom, width_ratios = [1., 0.05])
    plotaxes_A_2d = [fig.add_subplot(gridspec[i, j]) for i in 1:1, j in [1]]
    cax = fig.add_subplot(gridspec[1, 2])

    ik0 = argmax(data.es_plot[1, :])

    As_list = [data.As_dict["FM_scγ"]]
    As_labels = [""]
    plot_spectral(data, [plotaxes_A_2d[1]], nothing; As_list, As_labels, cax, iT, vmin=0, vmax=5, e_unit_type, ymin, ymax, c_onshell = "b", plot_horizontal_lines = false, ik0, logscale = false, plot_RSPT = false)
    cax.set_yticks(0:5)

    # 2d plot x labels
    for (iax, ax) in enumerate(plotaxes_A_2d)
        ax.axvline(0; c="k", lw = 0.5, ls="-")

        xk_ΓK = 0.5424990054552199
        xk_ΓM = 0.46981792025202973
        ax.set_xticks([-xk_ΓK/2, 0, xk_ΓM/2])
        ax.set_xticklabels(["K/2", "\$\\Gamma\$", "M/2"])
        ax.set_ylim([-0.5, 0.39])
        ax.set_xlim([-0.38, 0.35])
    end
    for ax in plotaxes_A_2d[2:end]
        ax.set_yticklabels([])
    end

    # Global settings

    plotaxes_A_2d[1].set_ylabel("\$\\varepsilon - \\varepsilon_{\\mathrm{VBM}}\$ (eV)"; labelpad = -1)

    for ax in vcat(plotaxes_A_2d[2:end])
        ax.set_yticklabels([])
    end

    γ_sc_meV_list = [12.90018695603513, 22.163953319736414, 32.91736597797363, 49.50341633771752, 61.432239809775965, 72.68868345271916, 116.23905439817864]
    γ_sc_text = @sprintf "%.1f" γ_sc_meV_list[iT]

    for (iax, ax) in enumerate(plotaxes_A_2d)
        label = ["\$G_0 D_0\$, \$\\eta\\!=\\!\\gamma^{\\mathrm{sc}}\\!=\\!$γ_sc_text~\\mathrm{meV}\$"][iax]
        α = iax <= 4 ? 1.0 : 1.0
        ax.text(0.02, 0.99, "$label"; ha="left", va="top", transform=ax.transAxes, bbox = Dict("facecolor" => "w", "edgecolor" => "none", "pad" => 0.5, "alpha" => α))
    end

    fig.align_xlabels(plotaxes_A_2d)
    # fig.align_ylabels([cbar.ax for cbar in cbars])

    do_savefig(fig, "./figures/spectral_$material.sc_broadening.$(T_Kelvin)K"; transparent = false)

    display(fig); close(fig)
end


begin
    # Plot scFM vs ARPES

    material = "InSe"

    data = data_InSe
    iT = 3
    vmax = 5.0

    T_Kelvin = data.Tlist[iT] / unit_to_aru(:K)
    e_unit_type = :eV
    _e_to_eV(e) = (e - data.e_ref) / unit_to_aru(:eV)

    # ------------------------------------------------------------------------------------
    # 2d spectral function

    left   = 0.18
    right  = 0.87
    bottom = 0.12
    top    = 0.90

    fig = figure(figsize=(3.4, 1.3))
    gridspec = fig.add_gridspec(; nrows = 1, ncols = 5, hspace = 0.32, wspace = 0.,
        left, right, top, bottom, width_ratios = [1., 0.3, 1., 0.1, 0.05])
    plotaxes_A_2d = [fig.add_subplot(gridspec[i, j]) for i in 1:1, j in [1,3]]
    cax = fig.add_subplot(gridspec[1, 5])

    ik0 = argmax(data.es_plot[1, :])

    As_list = copy.([data.As_dict["scFM"]])
    apply_gaussian_filter!(As_list[1], iT, data)

    As_labels = ["Cumulant", "FM", "scFM"]
    plot_spectral(data, plotaxes_A_2d[1:1], nothing; As_list, As_labels, cax, iT, vmin = 0, vmax, e_unit_type, ymin=-1, ymax=-1, c_onshell = "b", plot_horizontal_lines = false, ik0, logscale = false, plot_RSPT = false)

    kwargs_A = (; origin="lower", aspect="auto", cmap="Blues", vmin = 0, vmax)
    ax = plotaxes_A_2d[2]
    inds = xplot_ARPES .<= 0
    extent_ = [extent_ARPES[1], 0, extent_ARPES[3], extent_ARPES[4]]
    img = ax.imshow(z_ARPES[:, inds] ; extent = extent_, origin="lower", aspect="auto", kwargs_A...)

    inds = xplot_ARPES .>= 0
    extent_ = [0, extent_ARPES[2], extent_ARPES[3], extent_ARPES[4]]
    img = ax.imshow(z_ARPES[:, inds]; extent = extent_, origin="lower", aspect="auto", kwargs_A...)


    for ib in ib_rng
        ax.plot(xplot, _e_to_eV.(es_plot[ib, :]), "k--"; lw = 1)
    end

    # 2d plot x labels
    for (iax, ax) in enumerate(plotaxes_A_2d)
        ax.axvline(0; c="k", lw = 0.5, ls="-")

        xk_ΓK = 0.5424990054552199
        xk_ΓM = 0.46981792025202973
        ax.set_xticks([-xk_ΓK/2, 0, xk_ΓM/2])
        ax.set_xticklabels(["K/2", "\$\\Gamma\$", "M/2"])
        ax.set_ylim([-0.5, 0.39])
        ax.set_xlim([-0.38, 0.35])
        ax.set_yticks([-0.4, -0.2, 0.0, 0.2])
    end
    for ax in plotaxes_A_2d[2:end]
        ax.set_yticklabels([])
    end

    # Global settings

    plotaxes_A_2d[1].set_ylabel("\$\\varepsilon - \\varepsilon_{\\mathrm{VBM}}\$ (eV)"; labelpad = -1)

    for ax in vcat(plotaxes_A_2d[2:end])
        ax.set_yticklabels([])
    end

    for (iax, ax) in enumerate(plotaxes_A_2d)
        label = ["sc\$G D_0\$", "ARPES"][iax]
        α = iax <= 4 ? 1.0 : 1.0
        ax.set_title(label; pad = 0)
    end

    fig.align_xlabels(plotaxes_A_2d)

    do_savefig(fig, "./figures/spectral_$material.scFM_and_exp.$(T_Kelvin)K"; transparent = false)

    display(fig); close(fig)
end



begin
    # Plot 1d and 2d spectral functions for each material
    material = "InSe"

    data = data_InSe
    ibplot = 1
    iT = 1; vmax_Re = 0.25; vmax_Im = 0.20
    iT = 3; vmax_Re = 0.25; vmax_Im = 0.25
    iT = 7; vmax_Re = 0.25; vmax_Im = 0.25

    T_Kelvin = data.Tlist[iT] / unit_to_aru(:K)
    e_unit_type = :eV
    _e_to_eV(e) = (e - data.e_ref) / unit_to_aru(:eV)

    color1 = "#7570b3"
    color2 = "#1b9e77"
    color3 = "#d95f02"

    # ------------------------------------------------------------------------------------
    # 2d spectral function

    left   = 0.13
    right  = 0.88
    bottom = 0.08
    top    = 0.94

    fig = figure(figsize=(7.0, 2.2))
    gridspec = fig.add_gridspec(; nrows = 2, ncols = 10, hspace = 0.05, wspace = 0.,
        left, right, top, bottom, width_ratios = [1., 0.05, 1., 0.05, 0.3, 1., 0.05, 1., 0.05, 0.05])
    plotaxes = [fig.add_subplot(gridspec[i, j]) for i in 1:2, j in [1, 3, 6, 8]]
    plotaxes_arr = reshape(plotaxes, 2, 2, 2)
    caxes = [fig.add_subplot(gridspec[i, 10]) for i in 1:2]

    extent = [extrema(xplot)..., _e_to_eV.(extrema(ωs_plot))...]

    for (iiT, iT) in enumerate([3, 7]), iax in 1:2, jax in 1:2
        key = ["FM", "scFM"][jax]
        ax = plotaxes_arr[iax, jax, iiT]

        if iax == 1
            z = real.(Σs_dict[key][:, ibplot, :, iT]) ./ unit_to_aru(:eV)
            kwargs_A = (; cmap="RdBu_r", vmin = -vmax_Re, vmax = vmax_Re)
        else
            z = .-imag.(Σs_dict[key][:, ibplot, :, iT]) ./ unit_to_aru(:eV)
            kwargs_A = (; cmap="hot_r", vmin = 0, vmax = vmax_Im)
        end

        inds = xplot .<= 0
        extent_ = [extent[1], 0, extent[3], extent[4]]
        img = ax.imshow(z[:, inds] ; extent = extent_, origin="lower", aspect="auto", kwargs_A...)

        inds = xplot .>= 0
        extent_ = [0, extent[2], extent[3], extent[4]]
        img = ax.imshow(z[:, inds]; extent = extent_, origin="lower", aspect="auto", kwargs_A...)

        if jax == 1
            colorbar(img; cax = caxes[iax])
        end

        if iax == 1 && jax == 1
            T_Kelvin = data.Tlist[iT] / unit_to_aru(:K)
            plotaxes_arr[1, 1, iiT].set_title("\$G_0 D_0\$, \$T=$(Int(T_Kelvin))\\,\$K"; pad = 0)
            plotaxes_arr[1, 2, iiT].set_title("sc\$G D_0\$, \$T=$(Int(T_Kelvin))\\,\$K"; pad = 0)
        end
    end

    # 2d plot x labels
    for (iax, ax) in enumerate(plotaxes)
        ax.axvline(0; c="k", lw = 0.5, ls="-")

        xk_ΓK = 0.5424990054552199
        xk_ΓM = 0.46981792025202973
        ax.set_xticks([-xk_ΓK/2, 0, xk_ΓM/2])
        ax.set_xticklabels(["K/2", "\$\\Gamma\$", "M/2"])
        ax.set_ylim([-0.5, 0.39])
        ax.set_xlim([-0.38, 0.35])
        ax.set_yticks([-0.4, -0.2, 0.0, 0.2])

        for ib in data.ib_rng
            ax.plot(xplot, _e_to_eV.(data.es_plot[ib, :]), "k--"; lw = 1)
        end
    end

    for ax in plotaxes[1:end-1, :]
        ax.set_xticklabels([])
    end
    for ax in plotaxes[:, 2:end]
        ax.set_yticklabels([])
    end

    for ax in plotaxes[:, 1]
        ax.set_ylabel("\$\\varepsilon - \\varepsilon_{\\mathrm{VBM}}\$ (eV)"; labelpad = -1)
    end
    caxes[1].set_ylabel("\$\\mathrm{Re} \\Sigma\\ (\\mathrm{eV})\$"; labelpad = -4)
    caxes[2].set_ylabel("\$-\\mathrm{Im} \\Sigma\\ (\\mathrm{eV})\$"; labelpad = -4)
    # caxes[2].set_yticks([0, 0.1, 0.2])


    ax = fig.add_axes([0, 0, 1, 1])
    ax.set_axis_off()
    ax.set_xlim([0, 1])
    ax.set_ylim([0, 1])
    ax.axvline(0.5; c="k", lw = 0.5, ls="--")


    # Global settings

    for (iax, ax) in enumerate(permutedims(plotaxes_arr, (2, 1, 3)))
        ax.text(0.02, 0.98, "($(('a':'z')[iax]))"; ha="left", va="top", transform=ax.transAxes)
    end

    fig.align_xlabels(plotaxes[1, :])
    fig.align_ylabels(plotaxes[:, 1])
    fig.align_ylabels(caxes)

    do_savefig(fig, "./figures/selfen_$material"; transparent = false)

    display(fig); close(fig)
end



begin
    # Plot experiment only

    material = "InSe"

    data = data_InSe
    iT = 3  # 100 K
    vmax = 5.0

    T_Kelvin = data.Tlist[iT] / unit_to_aru(:K)
    e_unit_type = :eV
    _e_to_eV(e) = (e - data.e_ref) / unit_to_aru(:eV)

    # ------------------------------------------------------------------------------------
    # 2d spectral function

    left   = 0.20
    right  = 0.85
    bottom = 0.11
    top    = 0.96

    fig = figure(figsize=(2.4, 1.6))
    gridspec = fig.add_gridspec(; nrows = 1, ncols = 2, hspace = 0.32, wspace = 0.05,
        left, right, top, bottom, width_ratios = [1., 0.05])
    plotaxes_A_2d = [fig.add_subplot(gridspec[i, j]) for i in 1:1, j in [1,]]
    cax = fig.add_subplot(gridspec[1, 2])

    ik0 = argmax(data.es_plot[1, :])

    kwargs_A = (; origin="lower", aspect="auto", cmap="Blues", vmin = 0, vmax = vmax)
    ax = plotaxes_A_2d[1]
    inds = xplot_ARPES .<= 0
    extent_ = [extent_ARPES[1], 0, extent_ARPES[3], extent_ARPES[4]]
    img = ax.imshow(z_ARPES[:, inds]; extent = extent_, origin="lower", aspect="auto", kwargs_A...)

    inds = xplot_ARPES .>= 0
    extent_ = [0, extent_ARPES[2], extent_ARPES[3], extent_ARPES[4]]
    img = ax.imshow(z_ARPES[:, inds]; extent = extent_, origin="lower", aspect="auto", kwargs_A...)

    cbar = colorbar(img; cax)
    cbar.ax.set_ylabel("\$A_\\mathbf{k}\\ (1 / \\mathrm{eV})\$"; labelpad = 2)

    # Plot DFT bands
    for ib in ib_rng
        ax.plot(xplot, _e_to_eV.(data.es_plot[ib, :]), "k--"; lw = 1)
    end

    # 2d plot x labels
    for (iax, ax) in enumerate(plotaxes_A_2d)
        ax.axvline(0; c="k", lw = 0.5, ls="-")

        xk_ΓK = 0.5424990054552199
        xk_ΓM = 0.46981792025202973
        ax.set_xticks([-xk_ΓK/2, 0, xk_ΓM/2])
        ax.set_xticklabels(["K/2", "\$\\Gamma\$", "M/2"])
        ax.set_ylim([-0.5, 0.39])
        ax.set_xlim([-0.38, 0.35])
    end
    for ax in plotaxes_A_2d[2:end]
        ax.set_yticklabels([])
    end

    # Global settings
    plotaxes_A_2d[1].set_ylabel("\$\\varepsilon - \\varepsilon_{\\mathrm{VBM}}\$ (eV)"; labelpad = -1)

    for ax in vcat(plotaxes_A_2d[2:end])
        ax.set_yticklabels([])
    end

    fig.align_xlabels(plotaxes_A_2d)

    do_savefig(fig, "./figures/spectral_$material.exp_only.$(T_Kelvin)K"; transparent = false)

    display(fig); close(fig)
end
