using Lehmann
using PyPlot
using NLsolve

struct HolsteinSolver
    # Parameters
    emax :: Float64
    β    :: Float64
    ω₀   :: Float64
    t    :: Float64
    g    :: Float64
    μ    :: Float64

    # DLR grid
    dlrF :: DLRGrid{Float64, :none}
    dlrB :: DLRGrid{Float64, :none}

    # Bare Green functions
    G0 :: Array{ComplexF64, 3}
    D0 :: Array{ComplexF64, 3}

    # Self-energy
    Σ  :: Array{ComplexF64, 3}
    Π  :: Array{ComplexF64, 3}

    # Full Green functions
    G  :: Array{ComplexF64, 3}
    D  :: Array{ComplexF64, 3}

    function HolsteinSolver(; emax, β, ω₀, t, g, μ)
        @info "γ = $(ω₀/t), λ = $(2g^2 / t / ω₀)"

        # Setup DLR
        dlrF = DLRGrid(emax, β, 1e-6, true)
        dlrB = DLRGrid(emax, β, 1e-6, false)
        numF = length(dlrF)
        numB = length(dlrB)
        @info "DLR grid size: Fermion = $numF, Boson = $numB"

        # Bare electron Green function
        # Hartree shift
        ΣH = -2g^2 / ω₀
        # ΣH = 0.
        G0 = zeros(ComplexF64, 2, 2, numF)
        y1 = @. 1 / (im * dlrF.ωn + t + μ - ΣH)
        y2 = @. 1 / (im * dlrF.ωn - t + μ - ΣH)
        @. G0[1, 1, :] = (y1 + y2) / 2
        @. G0[2, 2, :] = (y1 + y2) / 2
        @. G0[1, 2, :] = (y1 - y2) / 2
        @. G0[2, 1, :] = (y1 - y2) / 2

        # Bare phonon Green function
        D0 = zeros(ComplexF64, 2, 2, numB)
        D0[1, 1, :] .= -ω₀ ./ ((dlrB.ωn).^2 .+ ω₀^2)
        D0[2, 2, :] .= -ω₀ ./ ((dlrB.ωn).^2 .+ ω₀^2)

        # Templates for the self-energies and full Green functions
        G = copy(G0)
        D = copy(D0)
        Σ = zeros(ComplexF64, 2, 2, numF)
        Π = zeros(ComplexF64, 2, 2, numB)

        new(emax, β, ω₀, t, g, μ, dlrF, dlrB, G0, D0, Σ, Π, G, D)
    end
end

function self_energy!(S :: HolsteinSolver)
    # Compute self-energy
    (; dlrF, dlrB, G, D, Σ, Π, g, ω₀) = S
    nspin = 2

    # Electron self-energy
    # Σ_ij(τ) = -2g² G_ij(τ) D_ij(τ)
    G_τ = matfreq2tau(dlrF, G, dlrF.τ; axis=3)
    D_τ = matfreq2tau(dlrB, D, dlrF.τ; axis=3)
    Σ_τ = -2g^2 .* G_τ .* D_τ
    Σ .= tau2matfreq(dlrF, Σ_τ; axis=3)

    # Phonon self-energy
    # Π_ij(τ) = 2g² G_ji(τ) G_ij(-τ)
    G_pτ = matfreq2tau(dlrF, G, dlrB.τ; axis=3)
    G_mτ = matfreq2tau(dlrF, G, dlrB.β .- dlrB.τ; axis=3) .* -1
    Π_τ = 2g^2 .* permutedims(G_pτ, (2, 1, 3)) .* G_mτ
    Π .= tau2matfreq(dlrB, Π_τ; axis=3)

    # Spin degree of freedom
    Π .*= nspin

    # # Hartree self-energy
    n_el = dropdims(matfreq2tau(dlrF, G, [-eps(Float64)]; axis=3), dims=3)
    n_el .= (n_el + n_el') ./ 2
    # u1 = -sqrt(2) * g / ω₀ * real(n_el[1, 1]) * nspin
    # u2 = -sqrt(2) * g / ω₀ * real(n_el[2, 2]) * nspin
    # Σ[1, 1, :] .+= sqrt(2) * S.g * u1
    # Σ[2, 2, :] .+= sqrt(2) * S.g * u2
    display(real.(n_el))

    return (; Σ, Π)
end

function Dyson!(S :: HolsteinSolver)
    # Solve Dyson equation
    (; G0, D0, G, D, Σ, Π) = S

    @views for iω in axes(G, 3)
        # G[:, :, iω] .= inv(inv(G0[:, :, iω]) - Σ[:, :, iω])
        G[:, :, iω] .= G0[:, :, iω] + G0[:, :, iω] * Σ[:, :, iω] * G[:, :, iω]
    end
    @views for iω in axes(D, 3)
        # D[:, :, iω] .= inv(inv(D0[:, :, iω]) - Π[:, :, iω])
        D[:, :, iω] .= D0[:, :, iω] + D0[:, :, iω] * Π[:, :, iω] * D[:, :, iω]
    end

    return (; G, D)
end

function flatten(S :: HolsteinSolver)
    vcat(vec(S.G), vec(S.D))
end

function flatten!(S :: HolsteinSolver, x :: Vector)
    nG = length(S.G)
    x[    1: nG] .= vec(S.G)
    x[ nG+1:2nG] .= vec(S.D)
    return x
end

function unflatten!(S :: HolsteinSolver, x :: Vector)
    nG = length(S.G)
    @views S.G .= reshape(x[   1:  nG], size(S.G))
    @views S.D .= reshape(x[ nG+1:2nG], size(S.D))
    return nothing
end

function energy(S :: HolsteinSolver)
    # Compute total energy
    (; dlrF, dlrB, G, D, Σ, t, ω₀, g) = S
    nspin = 2

    # Electron density matrix and energy
    n_el = dropdims(matfreq2tau(dlrF, G, [-eps(Float64)]; axis=3), dims=3)
    n_el .= (n_el + n_el') ./ 2
    E_el = -t * real(n_el[1, 2] + n_el[2, 1]) * nspin

    # Phonon density matrix and energy
    n_ph = dropdims(matfreq2tau(dlrB, D, [-eps(Float64)]; axis=3), dims=3)
    n_ph .= (n_ph + n_ph') ./ 2
    E_ph = - ω₀/2 * real(n_ph[1, 1] + n_ph[2, 2]) - ω₀/2

    # Position shift
    u1 = -sqrt(2) * g / ω₀ * real(n_el[1, 1]) * nspin
    u2 = -sqrt(2) * g / ω₀ * real(n_el[2, 2]) * nspin
    E_ph += ω₀ / 2 * (u1^2 + u2^2)

    # Electron-phonon coupling energy
    # E = 1/β ∑_ν Σ_{ij}(ν) G_{ji}(ν) = ∫dτ Σ_{ij}(-τ) G_{ji}(τ)
    Σ_mτ = matfreq2tau(dlrF, Σ, .-dlrB.τ; axis=3)
    G_pτ = matfreq2tau(dlrF, G, .+dlrB.τ; axis=3)
    E_elph = real(sum(tau2matfreq(dlrB, Σ_mτ .* permutedims(G_pτ, (2, 1, 3)), 0:0; axis=3))) * nspin

    E_elph += sqrt(2) * g * u1 * real(n_el[1, 1]) * nspin
    E_elph += sqrt(2) * g * u2 * real(n_el[2, 2]) * nspin


    E_tot = E_el + E_ph + E_elph

    return (; E_tot, E_el, E_ph, E_elph)
end

function iterate_solver!(S :: HolsteinSolver)
    self_energy!(S)
    Dyson!(S)
end

# fixed-point equation
function fixed_point!(
    R :: Vector{Q},
    x :: Vector{Q},
    S,
    ) :: Nothing where {Q}

    # update solver from input vector
    unflatten!(S, x)

    # Iterate the solver
    iterate_solver!(S)

    # calculate residue
    flatten!(S, R)
    R .-= x

    return nothing
end

# run the solver
function solve!(
    S,
    ;
    maxiter  :: Int64 = 100,
    tol      :: Float64 = 1e-4,
    δ        :: Float64 = 0.85,
    mem      :: Int64 = 20,
    verbose  :: Bool = true,
    )

    verbose && println("Converging parquet equations.")

    ti = time()
    res = nlsolve((R, x) -> fixed_point!(R, x, S), flatten(S),
        method=:anderson,
        iterations=maxiter,
        ftol=tol,
        beta=δ,
        m=mem,
        show_trace = verbose,
    )

    verbose && println("Done. Calculation took $(round(time() - ti, digits = 3)) seconds.")

    return res
end

begin
    # Parameters
    emax = 1e2
    β = 1e3

    ω₀ = 1.0
    t = 8.0
    g = 1.5
    μ = -2g^2 / ω₀

    S = HolsteinSolver(; emax, β, ω₀, t, g, μ);
    @info energy(S)[1] / S.t
    res = solve!(S, mem=1, δ = 0.1, maxiter=200);
    @info energy(S)[1] / S.t

    G_ = copy(S.G0)
    D_ = copy(S.D0)


    gs = range(0.0, 4., length=32)
    es = map(gs) do g
        μ = -2g^2 / ω₀
        S = HolsteinSolver(; emax, β, ω₀, t, g, μ);

        S.G .= G_
        S.D .= D_

        @time solve!(S, verbose=true, maxiter = 1);
        @info res.f_converged

        G_ .= S.G
        D_ .= S.D
        energy(S)[1]
    end
# end;

# begin
    plot(2 * gs.^2 ./ t ./ ω₀, -2 .- 2 * gs.^2 ./ t ./ ω₀)
    plot(2 * gs.^2 ./ t ./ ω₀, es ./ t)
    # ylim([-10, -2])
    xlim(left=0)
    fig = gcf(); display(fig); close(fig)
end


begin
    emax = 1e2
    β = 1e3

    ω₀ = 1.0
    t = 8.0
    g = 4.
    μ = -2g^2 / ω₀

    S = HolsteinSolver(; emax, β, ω₀, t, g, μ);
    dropdims(matfreq2tau(S.dlrF, S.G, [-eps(Float64)]; axis=3), dims=3) |> display

    solve!(S; maxiter = 1, mem=0);
    @info energy(S)[1] / S.t

    dropdims(matfreq2tau(S.dlrF, S.G, [-eps(Float64)]; axis=3), dims=3) |> display

    (; dlrF, dlrB, G, D) = S
    # G = S.Σ
    # D = S.Π
    Gpp = (G[1, 1, :] .+ G[1, 2, :] .+ G[2, 1, :] .+ G[2, 2, :]) ./ 2
    Gmm = (G[1, 1, :] .- G[1, 2, :] .- G[2, 1, :] .+ G[2, 2, :]) ./ 2

    D11 = D[1, 1, :]

    ns = -1001:10:1000
    νs = @. π * (2ns + 1) / S.β
    plot(νs, real.(matfreq2matfreq(dlrF, Gpp, ns)))
    plot(νs, imag.(matfreq2matfreq(dlrF, Gpp, ns)))
    plot(νs, real.(matfreq2matfreq(dlrF, Gmm, ns)), "--")
    plot(νs, imag.(matfreq2matfreq(dlrF, Gmm, ns)), "--")

    # axvline.(S.dlrF.ωn, c="k", lw=0.5)
    axhline(0, c="k", lw=1)

    xlim(extrema(νs))

    # ms = -1001:10:1000
    # ωs = @. π * (2ms) / S.β
    # plot(ωs, real.(matfreq2matfreq(dlrB, D11, ms)))
    # plot(ωs, imag.(matfreq2matfreq(dlrB, D11, ms)))

    fig = gcf(); display(fig); close(fig)
end



begin
    emax = 1e2
    β = 1e3

    ω₀ = 1.0
    t = 8.0
    g = 4.
    μ = -2g^2 / ω₀

    S = HolsteinSolver(; emax, β, ω₀, t, g, μ);
    dropdims(matfreq2tau(S.dlrF, S.G, [-eps(Float64)]; axis=3), dims=3) |> display

    solve!(S; maxiter = 1, mem=0);
    @info energy(S)[1] / S.t

    dropdims(matfreq2tau(S.dlrF, S.G, [-eps(Float64)]; axis=3), dims=3) |> display

    (; dlrF, dlrB, G, D) = S

    G_τ = matfreq2tau(dlrF, G, dlrF.τ; axis=3)
    D_τ = matfreq2tau(dlrB, D, dlrF.τ; axis=3)
    Σ_τ = -2g^2 .* G_τ .* D_τ
    Σ = tau2matfreq(dlrF, Σ_τ; axis=3)


    τs = range(0, 2β / 1000, length=1001)
    # plot(τs, real.(tau2tau(dlrF, G_τ[1,1,:], τs)))
    # plot(τs, real.(tau2tau(dlrF, D_τ[1,1,:], τs)))
    # plot(τs, real.(matfreq2tau(dlrB, D[1,1,:], τs)))
    # plot(νs, imag.(matfreq2matfreq(dlrF, Σ[1,1,:], ns)))

    D11 = D[1, 1, :]

    ns = -10001:10:10000
    νs = @. π * (2ns + 1) / S.β
    # plot(νs, real.(matfreq2matfreq(dlrF, S.Σ[1,1,:], ns)))
    # plot(νs, imag.(matfreq2matfreq(dlrF, S.Σ[1,1,:], ns)))
    # plot(νs, real.(matfreq2matfreq(dlrF, Σ[1,1,:], ns)), "--")
    # plot(νs, imag.(matfreq2matfreq(dlrF, Σ[1,1,:], ns)), "--")

    # axvline.(S.dlrF.ωn, c="k", lw=0.5)
    # axhline(0, c="k", lw=1)

    # xlim(extrema(νs))

    ms = -1001:10:1000
    ωs = @. π * (2ms) / S.β
    plot(ωs, real.(matfreq2matfreq(dlrB, S.Π[1,1,:], ms)))
    # plot(ωs, imag.(matfreq2matfreq(dlrB, S.Π[1,1,:], ms)))

    plot(ωs, 1 ./ real.(matfreq2matfreq(dlrB, S.D0[1,1,:], ms)))
    # plot(ωs, real.(matfreq2matfreq(dlrB, S.D[1,1,:], ms)))

    # plot(dlrB.ωn, real.(S.D[1,1,:]), "o-")

    axvline.(S.dlrB.ωn, c="k", lw=0.5)
    xlim(extrema(ωs))
    ylim([-2, 2])
    # xlim([-1, 1])

    # plot(τs, real.(matfreq2tau(dlrB, S.D0[1,1,:], τs)))
    # plot(τs, real.(matfreq2tau(dlrB, S.D[1,1,:], τs)))

    fig = gcf(); display(fig); close(fig)
end

begin
    # G = S.G0
    # G_pτ = matfreq2tau(dlrF, G, dlrB.τ; axis=3)
    # G_mτ = matfreq2tau(dlrF, G, β.-dlrB.τ; axis=3) .* -1
    # Π_τ = 2g^2 .* permutedims(G_pτ, (2, 1, 3)) .* G_mτ
    # Π = tau2matfreq(dlrB, Π_τ; axis=3)

    τs = range(0, S.β, length=1001)
    # plot(τs, real.(matfreq2tau(dlrF, G[1,1,:], τs)))
    # plot(τs, real.(matfreq2tau(dlrF, G[1,1,:], β.-τs) .* -1))
    # plot(τs, real.(tau2tau(dlrF, Π_τ[1,1,:], τs)))
    plot(τs, real.(matfreq2tau(S.dlrB, S.D0[1,1,:], τs)))
    # plot(τs, real.(matfreq2tau(dlrB, D[1,1,:], τs)))
    # plot(νs, imag.(matfreq2matfreq(dlrF, Σ[1,1,:], ns)))
    # xlim([0, β/100])
    # xlim(β .- [0, β/100])
    fig = gcf(); display(fig); close(fig)
end



function kramers_kronig(f, vs, vs_out = vs)
    dv = vs[2] - vs[1]
    map(vs_out) do v
        sum(@. imag(f) / (im * v - vs)) / π * dv
    end
end

begin
    β = 10.
    dlr = DLRGrid(10., β, 1e-6, false)

    ws = range(-10, 10, length=1000)
    e1 = 2.0
    e2 = 3.0
    g = 0.1
    y1 = @. e1 / ((ws + g * im)^2 - e1^2)
    y2 = @. e2 / ((ws + g * im)^2 - e2^2) * 7.0

    y3 = @. 1 / (1 / y1 - y2)
    # y3 = @. y1 + y1 * y2 * y1

    z1 = kramers_kronig(y1, ws)
    z2 = kramers_kronig(y2, ws)
    z3 = kramers_kronig(y3, ws)

    fig, plotaxes = subplots(2,1, figsize=(5,8))
    sca(plotaxes[1])
    plot(ws, imag.(y1))
    plot(ws, imag.(y2))
    plot(ws, imag.(y3))
    plot(ws, imag.(@. y1 + y1*y2*y3), "k--")
    ylim([-2, 3])
    axhline(0, c="k")
    axvline(0, c="k")

    z4 = @. 1 / (1 / z1 - z2)
    # z4 = @. z1 + z1 * z2 * z1

    τs = range(nextfloat(-β), β, length=1001) |> collect

    sca(plotaxes[2])
    # plot(τs, real.(matfreq2tau(dlr, (@. e1 / (dlr.ωn^2 + e1^2)), τs)), "--")

    # plot(τs, real.(matfreq2tau(dlr, (@. e2 / (dlr.ωn^2 + e2^2)), τs)) .* 7, "--")

    # z1 = kramers_kronig(y1, ws, dlr.ωn)
    # z2 = kramers_kronig(y2, ws, dlr.ωn) .* 7
    # plot(τs, real.(matfreq2tau(dlr, kramers_kronig(y1, ws, dlr.ωn), τs)))
    # plot(τs, real.(matfreq2tau(dlr, kramers_kronig(y2, ws, dlr.ωn), τs)))
    # plot(τs, real.(matfreq2tau(dlr, kramers_kronig(y3, ws, dlr.ωn), τs)))
    # ylim([0, 1])

    # plot(τs, real.(matfreq2tau(dlr, (@. 1 / (1 / z1 - z2)), τs)))

    # plot(τs, real.(matfreq2tau(dlr, (@. e2 / (dlr.ωn^2 + e2^2)), τs)) .* 7, "--")
    # plot(τs, real.(matfreq2tau(dlr, (@. e2 / (dlr.ωn^2 + e2^2)), τs)) .* 7, "--")


    plot(ws, real.(z1))
    plot(ws, real.(z2))
    plot(ws, real.(z3))
    # plot(ws, real.(@. z1 + z3 * z2 * z1), "k--")
    plot(ws, real.(z4), "--")
    ylim([-2, 3])
    # yscale("symlog")
    # plot(ws, 1 ./ real.(z1), "k--")
    fig = gcf(); display(fig); close(fig)
end

begin
    ws = range(-10, 10, length=1000)
    e1 = 2.0
    e2 = 3.0
    g = 0.1
    y1 = zeros(ComplexF64, 2, 2, length(ws))
    y2 = zeros(ComplexF64, 2, 2, length(ws))
    y1[1, 2, :] = @. -1 / (ws + g * im + e1)
    y1[2, 1, :] = @.  1 / (ws + g * im - e1)
    y20 = @. e2 / ((ws + g * im)^2 - e2^2) * 3.5
    y2[1, 1, :] .= y20
    y2[1, 2, :] .= y20
    y2[2, 1, :] .= y20
    y2[2, 2, :] .= y20

    y3 = zeros(ComplexF64, 2, 2, length(ws))
    z1 = zeros(ComplexF64, 2, 2, length(ws))
    z2 = zeros(ComplexF64, 2, 2, length(ws))
    z3 = zeros(ComplexF64, 2, 2, length(ws))
    z4 = zeros(ComplexF64, 2, 2, length(ws))
    for iw in 1:length(ws)
        y3[:, :, iw] .= inv(inv(y1[:, :, iw]) - y2[:, :, iw])
    end
    for i in 1:2, j in 1:2
        z1[i, j, :] = kramers_kronig(y1[i, j, :], ws)
        z2[i, j, :] = kramers_kronig(y2[i, j, :], ws)
        z3[i, j, :] = kramers_kronig(y3[i, j, :], ws)
    end
    for iw in 1:length(ws)
        z4[:, :, iw] .= inv(inv(z1[:, :, iw]) - z2[:, :, iw])
    end

    fig, plotaxes = subplots(2,1, figsize=(5,8))
    sca(plotaxes[1])
    plot(ws, imag.(y1[1,2,:]))
    plot(ws, imag.(y2[1,2,:]))
    plot(ws, imag.(y3[1,2,:]))
    ylim([-2, 3])

    # z4 = @. 1 / (1 / z1 - z2)
    # # z4 = @. z1 + z1 * z2 * z1

    # τs = range(nextfloat(-β), β, length=1001) |> collect

    sca(plotaxes[2])
    # # plot(τs, real.(matfreq2tau(dlr, (@. e1 / (dlr.ωn^2 + e1^2)), τs)), "--")

    # # plot(τs, real.(matfreq2tau(dlr, (@. e2 / (dlr.ωn^2 + e2^2)), τs)) .* 7, "--")

    # # z1 = kramers_kronig(y1, ws, dlr.ωn)
    # # z2 = kramers_kronig(y2, ws, dlr.ωn) .* 7
    # # plot(τs, real.(matfreq2tau(dlr, kramers_kronig(y1, ws, dlr.ωn), τs)))
    # # plot(τs, real.(matfreq2tau(dlr, kramers_kronig(y2, ws, dlr.ωn), τs)))
    # # plot(τs, real.(matfreq2tau(dlr, kramers_kronig(y3, ws, dlr.ωn), τs)))
    # # ylim([0, 1])

    # # plot(τs, real.(matfreq2tau(dlr, (@. 1 / (1 / z1 - z2)), τs)))

    # # plot(τs, real.(matfreq2tau(dlr, (@. e2 / (dlr.ωn^2 + e2^2)), τs)) .* 7, "--")
    # # plot(τs, real.(matfreq2tau(dlr, (@. e2 / (dlr.ωn^2 + e2^2)), τs)) .* 7, "--")


    plot(ws, real.(z1[1, 2, :]))
    plot(ws, real.(z2[1, 2, :]))
    plot(ws, real.(z3[1, 2, :]))
    plot(ws, real.(z4[1, 2, :]), "--")
    ylim([-2, 3])
    # yscale("symlog")
    # plot(ws, 1 ./ real.(z1), "k--")
    fig = gcf(); display(fig); close(fig)
end


begin
    using Interpolations
    using HCubature

    itp1 = linear_interpolation(τs, matfreq2tau(dlr, kramers_kronig(y1, ws, dlr.ωn), τs))
    itp2 = linear_interpolation(τs, matfreq2tau(dlr, kramers_kronig(y2, ws, dlr.ωn), τs))
    itp3 = linear_interpolation(τs, matfreq2tau(dlr, kramers_kronig(y3, ws, dlr.ωn), τs))

    y4 = map(τs[501:end]) do τ
        quadgk(x -> itp2(τ - x) * itp3(x), 0, β)[1]
    end
    y4 = vcat(y4[1:end-1], y4);
    itp4 = linear_interpolation(τs, y4)

    y5 = map(τs[501:end]) do τ
        quadgk(x -> itp1(τ - x) * itp4(x), 0, β)[1]
    end
    y5 = vcat(y5[1:end-1], y5);


    z1 = kramers_kronig(y1, ws, dlr.ωn)
    z2 = kramers_kronig(y2, ws, dlr.ωn) .* 7
    # plot(τs, real.(matfreq2tau(dlr, kramers_kronig(y1, ws, dlr.ωn), τs)))
    # plot(τs, real.(matfreq2tau(dlr, kramers_kronig(y2, ws, dlr.ωn), τs)))
    plot(τs, real.(matfreq2tau(dlr, kramers_kronig(y3, ws, dlr.ωn), τs)) .- real.(matfreq2tau(dlr, kramers_kronig(y1, ws, dlr.ωn), τs)))
    plot(τs, y4, "--")
    plot(τs, y5, "--")
    # ylim([0, 1])
    fig = gcf(); display(fig); close(fig)
end
