using EPSpectral
using LinearAlgebra
using StaticArrays
using Interpolations
using OhMyThreads
using PyPlot
using Base.Threads
using ChunkSplitters
using QuadGK
using Roots
using Printf

using EPSpectral: tmapreduce, retarded_self_energy_single_pole, chunks

# Holstein model for the electron-phonon interaction
#
# References
# [1] A. Nocera and M. Berciu, SciPost Phys. 15, 110 (2023)

struct HolsteinModel
    g  :: Float64
    ω₀ :: Float64
    t  :: Float64
    μ  :: Float64
end

function EPSpectral.update_chemical_potential!(S :: EPSpectral.ElectronPhononSolver{HolsteinModel})

    if S.occupation == 0
        μ_new = -Inf
    else
        μ_new = find_zero(μ -> EPSpectral.compute_occupation(S, μ) - S.occupation, extrema(S.ωs_dense))
    end

    S.model = HolsteinModel(S.model.g, S.model.ω₀, S.model.t, μ_new)

    return S
end


function EPSpectral.compute_occupation(S :: EPSpectral.ElectronPhononSolver{HolsteinModel}, μ)
    # Same as compute_occupation!, but do not update S.spectral_occ

    dω = S.ωs_dense[2] - S.ωs_dense[1]
    fermi = @. 1 / (exp((S.ωs_dense - μ) / S.T) + 1)

    spectral_occ = zero(S.spectral_occ)

    @views for (ik, k) in enumerate(S.ks_dense)
        # Compute occupation
        spectral_occ[ik] = sum(S.As_dense[:, ik] .* fermi) * dω
    end

    # Linearly interpolate the occupations
    occ_itp = linear_interpolation(S.ks_dense, spectral_occ)

    # Integrate over the Brillouin zone
    n = quadgk(k -> occ_itp(k) / 2π, extrema(S.ks_dense)...)[1]

    return n
end



Base.Broadcast.broadcastable(model::HolsteinModel) = Ref(model)

EPSpectral.get_εk(k, model::HolsteinModel) = -2 * model.t * cos(k)

function EPSpectral.get_Σ_mesh(k :: T, ωs :: AbstractVector, qpts :: EPSpectral.Kpoints{T}, param :: HolsteinModel; linewidth_on_electron = true) where {T}
    (; g, ω₀, μ) = param

    Σ = tmapreduce(.+, chunks(1:length(qpts); n = 2 * Threads.nthreads()); chunking = false) do iqs
        Σ_tmp = zeros(ComplexF64, length(ωs))

        for iq in iqs
            q, weight = qpts[iq]

            if norm(q) < sqrt(eps(Float64))
                continue
            end

            εkq = EPSpectral.get_εk(k + q, param)
            factor = g * weight

            for (iω, ω) in enumerate(ωs)
                if linewidth_on_electron
                    # imag(ω) is linewidth of electrons
                    Σ_tmp[iω] += retarded_self_energy_single_pole(real(ω), εkq - im * imag(ω), ω₀, μ) * factor

                else
                    # imag(ω) is linewidth of phonons
                    if εkq > μ
                        Σ_tmp[iω] += 1 / (ω - εkq - ω₀) * factor
                    elseif εkq < μ
                        Σ_tmp[iω] += 1 / (ω - εkq + ω₀) * factor
                    end
                end
            end
        end

        Σ_tmp
    end

    return Σ
end

function EPSpectral.compute_self_energy_self_consistent!(S :: ElectronPhononSolver{HolsteinModel}, qpts :: Kpoints)
    (; g, ω₀, μ) = S.model
    Σ_itp = linear_interpolation((S.ωs, S.ks), S.Σs .- im .* S.η; extrapolation_bc = Flat());

    for (ik, k) in enumerate(S.ks)
        Σs_imag = tmapreduce(.+, chunks(1:length(qpts); n = 2 * Threads.nthreads()); chunking = false) do iqs
            Σs_imag_q = zeros(length(S.ωs))

            for iq in iqs
                q, weight = qpts[iq]
                εkq = EPSpectral.get_εk(k + q, S.model)
                kq = norm(k + q)

                factor = weight * g

                for (iω, ω) in enumerate(S.ωs)
                    if real(ω) > μ + ω₀
                        Σs_imag_q[iω] += imag(1 / (ω - ω₀ - εkq - Σ_itp(real(ω) - ω₀, kq))) * factor
                    elseif real(ω) < μ - ω₀
                        Σs_imag_q[iω] += imag(1 / (ω + ω₀ - εkq - Σ_itp(real(ω) + ω₀, kq))) * factor
                    end
                end
            end

            Σs_imag_q
        end :: Vector{Float64}

        Σs_real = kramers_kronig(real.(S.ωs), Σs_imag)

        S.Σs[:, ik] .= Σs_real .+ im .* Σs_imag
    end

    return S
end

begin
    t = 1.0
    ω₀ = 1.0
    λ = 0.25
    g = sqrt(2 * ω₀ * t * λ)
    η = 0.02

    occupation = 0.15
    μ =-2.0
    model = HolsteinModel(g, ω₀, t, μ)

    nq = 300

    qpts = polynomial_grid_1d(π, nq, 1)
    pop!(qpts.vectors)
    pop!(qpts.weights)
    qpts.weights ./= 2π

    ωs = range(-15., 15., step=0.01)
    ks = range(-π, π, length=301)[1:end-1]


    k = 0.0

    # @time Σs_mesh = EPSpectral.get_Σ_mesh(k, ωs .+ im * η, qpts, model; linewidth_on_electron = false);

    # Σ_mesh_shift = real(get_Σ_analytic(k, get_εk(k, model) + im * η, model) - get_Σ_mesh(k, get_εk(k, model) + im * η, qpts, model))

    # Σs_mesh .+= Σ_mesh_shift

    # fig, plotaxes = subplots(1, 1, figsize=(4, 3), sharex=true)
    # plotaxes.plot(ωs, real.(Σs_mesh), "C0--"; label="Re Σ, mesh",)
    # plotaxes.plot(ωs, imag.(Σs_mesh), "C1--"; label="Im Σ, mesh",)
    # # plotaxes.plot(ωs, kramers_kronig(ωs, imag.(Σs_mesh); tail=false) .+ Σ_mesh_shift, "C2--"; label="Im Σ, mesh",)
    # # plotaxes.plot(ωs, .-kramers_kronig(ωs, real.(Σs_mesh)), "C3--"; label="Im Σ, mesh",)
    # axvline(get_εk(k, model), c="k", lw=1, ls="--")
    # axhline(0, c="k", lw=1)
    # xlim(extrema(ωs))
    # # legend()
    # # ylim([-1.5, 0.2])
    # display(fig); close(fig)
# end

# begin

    S = ElectronPhononSolver(model, ωs, ks; η, occupation)

    vmax = 5.0

    fig, plotaxes = subplots(4, 4, figsize=(12, 10), sharex=true)
    for ax in plotaxes[1:3, :]
        ax.set_ylim([-3.5, 1.0])
    end

    for iter in 1:4
        @time EPSpectral.compute_self_energy_self_consistent!(S, qpts)
        compute_spectral_function!(S)
        nocc = compute_occupation!(S)
        update_chemical_potential!(S)
        @info S.model.μ

        plot_self_energy!(plotaxes[1, iter], S, :real; cmap="RdBu_r", vmin=-10, vmax=10)
        plot_self_energy!(plotaxes[2, iter], S, :imag; cmap="RdBu_r", vmin=-10, vmax=10)
        plot_spectral_function!(plotaxes[3, iter], S; cmap="Blues", norm=matplotlib.colors.LogNorm(vmin=vmax/1e3, vmax=vmax))
        plotaxes[4, iter].plot(S.ks_dense, S.spectral_sum; lw=2, c="r")
        plotaxes[4, iter].plot(S.ks_dense, S.spectral_occ; lw=2)
        plotaxes[4, iter].axhline(1, c="k", lw=1, ls="--")
        plotaxes[4, iter].set_ylim([0, 1.2])
    end

    fig.suptitle(@sprintf "λ = %.3f, ω₀/t = %.3f" model.g^2/(2*model.ω₀*model.t) model.ω₀ / model.t )
    tight_layout()

    display(fig); close(fig)
end
