
function compute_self_energy_analytic!(S :: ElectronPhononSolver)
    for (ik, k) in enumerate(S.ks)
        S.Σs[:, ik] .= get_Σ_analytic.(k, S.ωs .+ im * S.η, S.model)
    end

    return S
end

function get_Σ_itp(S :: ElectronPhononSolver, η = S.η)
    Σs_with_η = copy(S.Σs)

    # Old version: Add -im * η to all elements
    # Σs_with_η .+= -im * η

    # New version: make |Im Σ| at least η
    inds = @. -imag(Σs_with_η) < η
    Σs_with_η[inds] .= real.(Σs_with_η[inds]) .- im * η

    linear_interpolation((S.ωs, S.ks), Σs_with_η; extrapolation_bc = Flat())
end

function get_Σ_itp_dense(S :: ElectronPhononSolver, η = S.η)
    Σs_with_η = copy(S.Σs_dense)
    if η != S.η
        # S.Σs_dense is set with S.η, so we need to recompute it.
        Σ_itp_coarse = get_Σ_itp(S, η)
        @views for (ik, k) in enumerate(S.ks_dense)
            @. Σs_with_η[:, ik] = Σ_itp_coarse(S.ωs_dense, k)
        end
    end

    # New version: make |Im Σ| at least η
    inds = @. -imag(Σs_with_η) < η
    Σs_with_η[inds] .= real.(Σs_with_η[inds]) .- im * η

    linear_interpolation((S.ωs_dense, S.ks_dense), Σs_with_η; extrapolation_bc = Flat())
end

"""
    compute_spectral_function!(S :: AbstractSolver; spectral_smearing = S.η)
"""
function compute_spectral_function!(S :: AbstractSolver; spectral_smearing = S.η)
    # Interpolate the self-energy from (ωs, ks) to (ωs_dense, ks_dense)
    Σ_itp = get_Σ_itp(S, spectral_smearing)

    # Compute spectral function
    @views for (ik, k) in enumerate(S.ks_dense)
        εk = get_εk(k, S.model)
        @. S.Σs_dense[:, ik] = Σ_itp(S.ωs_dense, k)
        @. S.As_dense[:, ik] = -imag(1 / (S.ωs_dense - εk - Σ_itp(S.ωs_dense, k))) / π
    end

    return S
end


function compute_occupation!(S :: ElectronPhononSolver)
    dω = S.ωs_dense[2] - S.ωs_dense[1]
    fermi = @. 1 / (exp((S.ωs_dense - S.model.μ) / S.model.T) + 1)

    @views for (ik, k) in enumerate(S.ks_dense)
        # Compute integral of the spectral function (should be 1 when exact)
        S.spectral_sum[ik] = sum(S.As_dense[:, ik]) * dω

        # Compute occupation
        S.spectral_occ[ik] = sum(S.As_dense[:, ik] .* fermi) * dω
    end

    # Compute total occupation
    if isfinite(S.model.μ)
        # Linearly interpolate the occupations
        occ_itp = linear_interpolation(S.ks_dense, S.spectral_occ)

        # Integrate over the Brillouin zone
        if S.model isa HolsteinLatticeModel || S.model isa PeierlsLatticeModel
            n = quadgk(occ_itp, -π, π)[1] / 2π
        elseif S.model isa FrohlichModel
            n = quadgk(k -> 1 / 2π^2 * k^2 * occ_itp(k), extrema(S.ks_dense)...)[1]
        else
            error("Model not supported")
        end
    else
        n = zero(S.model.μ)
    end

    return n
end


function compute_occupation(S :: ElectronPhononSolver, μ)
    # Same as compute_occupation!, but do not update S.spectral_occ

    dω = step(S.ωs_dense)
    fermi = @. occ_fermion(S.ωs_dense - μ, S.model.T)

    spectral_occ = zero(S.spectral_occ)

    @views for (ik, k) in enumerate(S.ks_dense)
        # Compute occupation
        spectral_occ[ik] = sum(S.As_dense[:, ik] .* fermi) * dω
    end

    # Linearly interpolate the occupations
    occ_itp = linear_interpolation(S.ks_dense, spectral_occ)

    # Integrate over the Brillouin zone
    if S.model isa HolsteinLatticeModel || S.model isa PeierlsLatticeModel
        n = quadgk(occ_itp, -π, π)[1] / 2π
    elseif S.model isa FrohlichModel
        n = quadgk(k -> 1 / 2π^2 * k^2 * occ_itp(k), extrema(S.ks_dense)...)[1]
    else
        error("Model not supported")
    end

    return n
end

function compute_occupation_MaxwellBoltzmann(S :: ElectronPhononSolver)
    # Same as compute_occupation!, but do not update S.spectral_occ

    dω = S.ωs_dense[2] - S.ωs_dense[1]
    fermi = @. exp(-S.ωs_dense / S.model.T)

    spectral_occ = zero(S.spectral_occ)

    @views for (ik, k) in enumerate(S.ks_dense)
        # Compute occupation
        spectral_occ[ik] = sum(S.As_dense[:, ik] .* fermi) * dω
    end

    # Linearly interpolate the occupations
    occ_itp = linear_interpolation(S.ks_dense, spectral_occ)

    # Integrate over the Brillouin zone
    if S.model isa HolsteinLatticeModel || S.model isa PeierlsLatticeModel
        n = quadgk(occ_itp, -π, π)[1] / 2π
    elseif S.model isa FrohlichModel
        n = quadgk(k -> 1 / 2π^2 * k^2 * occ_itp(k), extrema(S.ks_dense)...)[1]
    else
        error("Model not supported")
    end

    return n
end

function plot_spectral_function!(ax, S :: ElectronPhononSolver;
    bare_band = true,
    chemical_potential = true,
    yscale_fac = 1.0,
    kwargs_plot...
    )

    ks = S.ks_dense
    ωs = S.ωs_dense
    dk = ks[2] - ks[1]
    dω = ωs[2] - ωs[1]
    extent = [ks[1] - dk/2, ks[end] + dk/2, (ωs[1] - dω/2) * yscale_fac, (ωs[end] + dω/2) * yscale_fac]

    img = ax.imshow(S.As_dense; origin="lower", extent, aspect="auto", kwargs_plot...)

    if bare_band
        ax.plot(S.ks_dense, get_εk.(S.ks_dense, S.model) .* yscale_fac, c="grey", ls="--", lw=1)
    end
    if chemical_potential
        ax.axhline(S.model.μ * yscale_fac, c="r", ls="--", lw=1)
    end

    return img
end

function plot_self_energy!(ax, S :: ElectronPhononSolver, term = :real;
    bare_band = true,
    chemical_potential = true,
    yscale_fac = 1.0,
    kwargs_plot...
    )

    Σ_itp = get_Σ_itp(S, 0)

    ks = S.ks_dense
    ωs = S.ωs_dense
    dk = ks[2] - ks[1]
    dω = ωs[2] - ωs[1]
    extent = [ks[1] - dk/2, ks[end] + dk/2, (ωs[1] - dω/2) * yscale_fac, (ωs[end] + dω/2) * yscale_fac]
    Σs = [Σ_itp(ω, k) for ω in ωs, k in ks]

    func = term == :real ? real : imag

    img = ax.imshow(func.(Σs); origin="lower", extent, aspect="auto", kwargs_plot...)

    if bare_band
        ax.plot(S.ks_dense, get_εk.(S.ks_dense, S.model) .* yscale_fac, c="grey", ls="--", lw=1)
    end
    if chemical_potential
        ax.axhline(S.model.μ * yscale_fac, c="r", ls="--", lw=1)
    end

    return img
end

function update_chemical_potential!(S)

    if S.occupation == 0
        μ_new = -Inf
    else
        μ_new = find_zero(μ -> compute_occupation(S, μ) - S.occupation, extrema(S.ωs_dense))
    end

    S.model = FrohlichModel(S.model.α, S.model.ω₀, S.model.m, μ_new, S.model.T)

    return S
end


function compute_self_energy!(S :: ElectronPhononSolver)
    (; ω₀, μ, T) = S.model
    dim = get_dimension(S.qpts)
    Σ_itp = get_Σ_itp_dense(S, S.η)

    Σs_all = tmap(S.ks) do k
        Σs_imag_k = tmapreduce(.+, chunks(1:length(S.qpts); n = 2 * Threads.nthreads()); chunking = false) do iqs
            Σs_imag_q = zeros(length(S.ωs))

            for iq in iqs
                q, weight = S.qpts[iq]
                εkq = get_εk(SVector(k, 0, 0) + q, S.model)
                kq = norm(SVector(k, 0, 0) + q)
                if norm(kq) > S.ks[end]
                    # Keep k' = k+q only inside the k range, skip k+q points outside.
                    continue
                end

                ωq = ω₀
                nq = occ_boson(ωq, T)
                gq = get_eph_g(q, S.model)

                factor = weight * abs2(gq)

                for (iω, ω) in enumerate(S.ωs)
                    if nq < sqrt(eps(ω₀))
                        if real(ω) > μ + ωq
                            Σs_imag_q[iω] += imag(1 / (ω - ωq - εkq - Σ_itp(ω - ωq, kq))) * factor
                        elseif real(ω) < μ - ωq
                            Σs_imag_q[iω] += imag(1 / (ω + ωq - εkq - Σ_itp(ω + ωq, kq))) * factor
                        end
                    else
                        # Finite-temperature case
                        fac1 = imag(1 / (ω + ωq - εkq - Σ_itp(ω + ωq, kq)))
                        fac2 = imag(1 / (ω - ωq - εkq - Σ_itp(ω - ωq, kq)))
                        Σs_imag_q[iω] += ( fac1 * (nq + occ_fermion(ω + ωq - μ, T))
                                         + fac2 * (nq + 1 - occ_fermion(ω - ωq - μ, T)) ) * factor
                    end
                end
            end

            Σs_imag_q
        end :: Vector{Float64}

        Σs_imag_k ./= (2π)^dim
        Σs_real_k = kramers_kronig(real.(S.ωs), Σs_imag_k)

        Σs_real_k .+ im .* Σs_imag_k
    end

    for ik in eachindex(S.ks)
        S.Σs[:, ik] .= Σs_all[ik] .+ S.Σs_rest[ik]
    end

    return S
end


function compute_self_energy!(S :: ElectronPhononSolver{FrohlichModel_2D})
    (; α, ω₀, m, μ) = S.model
    Σ_itp = linear_interpolation((S.ωs, S.ks), S.Σs .- im .* S.η; extrapolation_bc = Flat());

    for (ik, k) in enumerate(S.ks)
        Σs_imag = tmapreduce(.+, chunks(1:length(S.qpts); n = 2 * Threads.nthreads()); chunking = false) do iqs
            Σs_imag_q = zeros(length(S.ωs))

            for iq in iqs
                q, weight = S.qpts[iq]
                εkq = get_εk(SVector(k, 0) + q, S.model)
                kq = norm(SVector(k, 0) + q)

                factor = weight / norm(q)

                for (iω, ω) in enumerate(S.ωs)
                    if real(ω) > μ + ω₀
                        Σs_imag_q[iω] += imag(1 / (ω - ω₀ - εkq - Σ_itp(real(ω) - ω₀, kq))) * factor
                    elseif real(ω) < μ - ω₀
                        Σs_imag_q[iω] += imag(1 / (ω + ω₀ - εkq - Σ_itp(real(ω) + ω₀, kq))) * factor
                    end
                end
            end

            Σs_imag_q
        end :: Vector{Float64}

        Σs_imag .*= α * sqrt(ω₀^3 / 2m) * 4π / (2π)^2
        Σs_real = kramers_kronig(real.(S.ωs), Σs_imag)

        S.Σs[:, ik] .= Σs_real .+ im .* Σs_imag
        S.Σs[:, ik] .+= S.Σs_rest[ik]
    end

    return S
end
