
mutable struct AbInitioElectronPhononSolverX2X <: EPSpectral.AbstractSolver
    model :: ElectronPhonon.Model
    η :: Float64

    occ   :: ElectronOccupationParams
    kpts  :: ElectronPhonon.GridKpoints  # k   points, on the irreducible BZ
    kqpts :: ElectronPhonon.GridKpoints  # k+q points, on the full BZ
    ikq_to_ik :: Vector{Int}         # Mapping from the k+q index to the k index by symmetry

    εks :: Matrix{Float64}
    εks_dense :: Matrix{Float64}

    ωs :: Vector{Float64}
    ωs_dense :: StepRangeLen{Float64, Base.TwicePrecision{Float64}, Base.TwicePrecision{Float64}, Int64}

    Σs :: Array{ComplexF64, 4}
    # As :: Array{Float64, 3}
    Σs_dense :: Array{ComplexF64, 4}
    # As_dense :: Array{Float64, 3}

    spectral_sum :: Array{Float64, 3}
    spectral_occ :: Array{Float64, 3}

    # Self-energy from the rest subspace, added to Σs after the calculation
    # Σs[ik, ib, iw] += Σs_rest[ik, ib]
    Σs_rest :: Array{ComplexF64, 3}

    # Active space window
    window_active :: Tuple{Float64, Float64}

    # Number of occupied states per unit cell that are outside the window. Used to update
    # the chemical potential.
    nstates_base :: Float64
end

function AbInitioElectronPhononSolverX2X(model, ωs, occ, kpts, kqpts, ikq_to_ik, Σs_rest, window_active
        ; η, ωs_dense = nothing, nstates_base = 0.0)
    nk = kpts.n
    nkq = kqpts.n
    nband = model.nw
    nT = length(occ)

    if size(Σs_rest) !== (nk, nband, nT)
        throw(ArgumentError("Σs_rest should have size ($nk, $nband, $nT), but got $(size(Σs_rest))"))
    end

    if ωs_dense === nothing
        ωs_dense = range(extrema(ωs)..., step = minimum(diff(ωs)) / 5)
    end

    Σs = zeros(ComplexF64, length(ωs), nband, nk, nT)
    Σs_dense = zeros(ComplexF64, length(ωs_dense), nband, nkq, nT)
    spectral_sum = zeros(nband, nkq, nT)
    spectral_occ = zeros(nband, nkq, nT)

    εks = compute_eigenvalues_el(model, kpts)
    εks_dense = compute_eigenvalues_el(model, kqpts)

    AbInitioElectronPhononSolverX2X(model, η, occ, kpts, kqpts, ikq_to_ik, εks, εks_dense, Vector(ωs), ωs_dense,
        Σs, Σs_dense, spectral_sum, spectral_occ, Σs_rest, window_active, nstates_base)
end

function compute_self_energy!(S :: AbInitioElectronPhononSolverX2X)
    calc = ElectronSpectralCalculatorX6X(;
        occ = S.occ,
        ωlist = S.ωs,
        smearing = S.η,
        ωs_input = S.ωs_dense,
        Σs_input = S.Σs_dense,
    )

    @time res = ElectronPhonon.run_eph_over_k_and_kq(model, S.kpts, S.kqpts, [calc]; window_kq = S.window_active)

    @views for iT in 1:length(S.occ)
        # Add the rest-space contribution and copy results from calc.Σ to S.Σs
        for ik in 1:S.kpts.n, ib in 1:S.model.nw
            S.Σs[:, ib, ik, iT] .= calc.Σ[ik][:, ib, iT] .+ S.Σs_rest[ik, ib, iT]
        end
    end

    unfold_and_interpolate_self_energy!(S)

    S
end

function unfold_and_interpolate_self_energy!(S :: AbInitioElectronPhononSolverX2X)
    # Unfold and interpolate self-energy from (k, ωs) to (k+q, ωs_dense)

    @views for iT in 1:length(S.occ)
        for ikq in 1:S.kqpts.n
            ik = S.ikq_to_ik[ikq]
            @views for ib in 1:S.model.nw
                itp = extrapolate(interpolate!((S.ωs,), S.Σs[:, ib, ik, iT], Gridded(Linear())), Flat())
                S.Σs_dense[:, ib, ikq, iT] .= itp.(S.ωs_dense)
            end
        end
    end
end

# function compute_spectral_function!(S :: AbInitioElectronPhononSolverX2X)
#     # Compute spectral function
#     @views for ikq in 1:S.kqpts.n, ib in 1:S.model.nw
#         # Old version: Add -im * η to all elements
#         # Σs_with_η = S.Σs_dense[:, ib, ikq] .- im .* S.η

#         # New version: make |Im Σ| at least η
#         Σs_with_η = copy(S.Σs_dense[:, ib, ikq])
#         inds = @. -imag(Σs_with_η) < S.η
#         Σs_with_η[inds] .= real.(Σs_with_η[inds]) .- im .* S.η

#         εk = S.εks_dense[ib, ikq]
#         @. S.As_dense[:, ib, ikq] = -imag(1 / (S.ωs_dense - εk - Σs_with_η)) / π
#     end

#     return S
# end

"""
    spectral_window_for_semiconductors(S :: AbInitioElectronPhononSolverX2X, iT)

For one-shot FM calculations of doped semiconductors, we truncate the spectral function
to avoid unphysical long tails.
For hole doping, truncate energies above (e_vbm + e_vbm_renorm) / 2 + 4T.
For electron doping, truncate energies below (e_cbm + e_cbm_renorm) / 2 - 4T.
Truncate only if chemical potential is outside the truncation range.
"""
function spectral_window_for_semiconductors(S :: AbInitioElectronPhononSolverX2X, iT)
    (; εks, window_active, ωs, Σs, occ) = S
    spectral_window_for_semiconductors(ωs, Σs, occ, εks, window_active, iT)
end

function spectral_window_for_semiconductors(ωs, Σs :: Array{ComplexF64, 4}, occ, εks, window_active, iT)
    # Implementation for Σs[iω, ib, ik, iT]
    (; T, μ) = occ[iT]

    @views if occ.type === :Semiconductor
        inds_active = findall(x -> window_active[1] <= x <= window_active[2], εks)

        if occ.nlist[iT] < 0
            # Hole doping. Truncate at e_trunc = (e_vbm + e_vbm_renorm) / 2 + 4T.
            # Truncate only if μ > e_trunc hold
            ib, ik = inds_active[argmax(εks[inds_active])].I
            e_vbm = εks[ib, ik]
            As_vbm = @. -imag(1 / (ωs - e_vbm - Σs[:, ib, ik, iT])) / π
            e_vbm_renorm = ωs[argmax(As_vbm)]
            e_trunc = _spectral_window_for_semiconductors_inner(e_vbm, e_vbm_renorm, T, :Hole)

            if μ > e_trunc
                return (-Inf, e_trunc)
            else
                return (-Inf, Inf)
            end

        else
            # Electron doping, Truncate at e_trunc = (e_cbm + e_cbm_renorm) / 2 - 4T.
            # Truncate only if μ < e_trunc hold
            ib, ik = inds_active[argmin(εks[inds_active])].I
            e_cbm = εks[ib, ik]
            As_cbm = @. -imag(1 / (ωs - e_cbm - Σs[:, ib, ik, iT])) / π
            e_cbm_renorm = ωs[argmax(As_cbm)]
            e_trunc = _spectral_window_for_semiconductors_inner(e_cbm, e_cbm_renorm, T, :Electron)

            if μ < e_trunc
                return (e_trunc, Inf)
            else
                return (-Inf, Inf)
            end

        end
    else
        # Metallic case, do not perform truncation.
        return (-Inf, Inf)
    end
end

function spectral_window_for_semiconductors(ωs, Σs :: Array{ComplexF64, 3}, occ, el, iT)
    # Implementation for Σs[iω, ind_el_i, iT]
    (; T, μ) = occ[iT]

    @views if occ.type === :Semiconductor
        if occ.nlist[iT] < 0
            # Hole doping. Truncate at e_trunc = (e_vbm + e_vbm_renorm) / 2 + 4T.
            # Truncate only if μ > e_trunc hold
            ind = argmax(el.e)
            e_vbm = el.e[ind]
            As_vbm = @. -imag(1 / (ωs - e_vbm - Σs[:, ind, iT])) / π
            e_vbm_renorm = ωs[argmax(As_vbm)]
            e_trunc = _spectral_window_for_semiconductors_inner(e_vbm, e_vbm_renorm, T, :Hole)

            if μ > e_trunc
                return (-Inf, e_trunc)
            else
                return (-Inf, Inf)
            end

        else
            # Electron doping, Truncate at e_trunc = (e_cbm + e_cbm_renorm) / 2 - 4T.
            # Truncate only if μ < e_trunc hold
            ind = argmin(el.e)
            e_cbm = el.e[ind]
            As_cbm = @. -imag(1 / (ωs - e_cbm - Σs[:, ind, iT])) / π
            e_cbm_renorm = ωs[argmax(As_cbm)]
            e_trunc = _spectral_window_for_semiconductors_inner(e_cbm, e_cbm_renorm, T, :Electron)

            if μ < e_trunc
                return (e_trunc, Inf)
            else
                return (-Inf, Inf)
            end

        end
    else
        # Metallic case, do not perform truncation.
        return (-Inf, Inf)
    end
end

function _spectral_window_for_semiconductors_inner(e1, e2, T, doping)
    if doping == :Electron
        return (e1 + e2) / 2 - 2T - sqrt(4T^2 + (e1 - e2)^2 / 4)
    elseif doping == :Hole
        return (e1 + e2) / 2 + 2T + sqrt(4T^2 + (e1 - e2)^2 / 4)
    else
        throw(ArgumentError("doping should be :Hole or :Electron, but got $doping"))
    end
end


function compute_occupation!(S :: AbInitioElectronPhononSolverX2X, iT; μ = S.occ.μlist[iT], truncate_spectral = false)
    (; occ, ωs_dense) = S
    dω = step(ωs_dense)
    T = occ.Tlist[iT]

    if occ.type === :Semiconductor && occ.nlist[iT] < 0
        # Count holes instead of electrons
        count_holes = true
        fermi = [occ_fermion(μ - ω, T) for ω in ωs_dense]
    else
        count_holes = false
        fermi = [occ_fermion(ω - μ, T) for ω in ωs_dense]
    end

    @views if truncate_spectral
        # For doped semiconductors, if truncate_spectral is true, truncate the spectral function.
        emin, emax = spectral_window_for_semiconductors(S, iT)
        fermi[ωs_dense .< emin] .= 0
        fermi[ωs_dense .> emax] .= 0
    end

    nocc = 0.0
    nocc_base = 0.0
    As = zeros(length(ωs_dense))

    @views for ik in 1:S.kpts.n, ib in 1:S.model.nw
        ek = S.εks[ib, ik]
        S.window_active[1] <= ek <= S.window_active[2] || continue

        Σ_itp = extrapolate(interpolate!((S.ωs,), S.Σs[:, ib, ik, iT], Gridded(Linear())), Flat())
        @. As = -imag(1 / (ωs_dense - ek - Σ_itp.(ωs_dense))) / π

        # Compute integral of the spectral function (should be 1 when exact)
        S.spectral_sum[ib, ik, iT] = sum(As) * dω

        # Compute occupation
        S.spectral_occ[ib, ik, iT] = sum(As .* fermi) * dω

        if count_holes
            # The electron occupation is 1 - S.spectral_occ. Since S.spectral_occ is tiny,
            # sum the two terms separately to avoid numerical errors.
            nocc -= S.spectral_occ[ib, ik, iT] * S.kpts.weights[ik]
            nocc_base += S.kpts.weights[ik]
        else
            nocc += S.spectral_occ[ib, ik, iT] * S.kpts.weights[ik]
        end
    end

    # Add the number of states outside the window, multiply by spin degeneracy
    nocc = (nocc + nocc_base + S.nstates_base) * occ.spin_degeneracy

    nocc
end


# function EPSpectral.compute_occupation(S :: AbInitioElectronPhononSolverX2X, μ)
#     dω = S.ωs_dense[2] - S.ωs_dense[1]

#     if S.occ.type === :Semiconductor && S.occ.nlist[1] < 0
#         # Count holes instead of electrons
#         count_holes = true
#         fermi = [occ_fermion(μ - ω, S.occ.Tlist[1]) for ω in S.ωs_dense]
#     else
#         count_holes = false
#         fermi = [occ_fermion(ω - μ, S.occ.Tlist[1]) for ω in S.ωs_dense]
#     end

#     spectral_occ = copy(S.spectral_occ)

#     @views for ik in 1:S.kqpts.n, ib in 1:S.model.nw
#         # Compute occupation
#         spectral_occ[ib, ik] = sum(S.As_dense[:, ib, ik] .* fermi) * dω
#     end

#     # Integrate over k+q to get the total occupation
#     n = sum(dropdims(sum(spectral_occ, dims=1), dims=1) .* S.kqpts.weights)

#     return n
# end

import EPSpectral: update_chemical_potential!

function EPSpectral.update_chemical_potential!(S :: AbInitioElectronPhononSolverX2X; truncate_spectral = false)
    for iT in 1:length(S.occ)
        n = S.occ.nlist[iT] + S.occ.nelec
        μ_new = EPSpectral.find_zero(μ -> compute_occupation!(S, iT; μ, truncate_spectral) - n, extrema(S.ωs_dense) .+ (-0.1, +0.1))
        S.occ.μlist[iT] = μ_new
    end
    return S
end


function compute_dos(kpts, eks, ωs, Σs; elist = ωs, η = 0.0, window_active = nothing)
    nband = size(eks, 1)
    nT = size(Σs, 4)
    @assert size(eks) == (nband, kpts.n)
    @assert size(Σs) == (length(ωs), nband, kpts.n, nT)
    
    dos = zeros(length(elist), nT)
    for iT in 1:nT, ik in 1:kpts.n, ib in 1:nband
        e = eks[ib, ik]
        if window_active !== nothing
            window_active[1] <= e <= window_active[2] || continue
        end
        Σ_itp = extrapolate(interpolate!((ωs,), Σs[:, ib, ik, iT], Gridded(Linear())), Flat())
        @. dos[:, iT] += -imag(1 / (elist - e - Σ_itp(elist) + im * η)) / π * kpts.weights[ik]
    end
    dos
end
