
import ElectronPhonon.OffsetArrays: OffsetArray, OffsetMatrix, no_offset_view

Base.@kwdef struct ElectronSelfEnergyCalculatorX2X
    # --- Parameter fields ---
    # Electron occupation
    occ :: ElectronOccupationParams

    # smearing
    smearing :: Float64

    # --- Data fields ---
    # Frequency-dependent self-energy of electrons
    Σ :: Vector{OffsetArray{ComplexF64, 2, Array{ComplexF64, 2}}} = [OffsetArray(zeros(ComplexF64, 0, 0))]

    # --- Data buffer fields ---
    Σ_buffer :: Vector{OffsetArray{ComplexF64, 2, Array{ComplexF64, 2}}} = [OffsetArray(zeros(ComplexF64, 0, 0))]

    # DEBUG field
    method :: Symbol = :Original
end

ElectronPhonon.allow_eph_outer_k(:: ElectronSelfEnergyCalculatorX2X) = true


function ElectronPhonon.setup_calculator!(calc::ElectronSelfEnergyCalculatorX2X, kpts, qpts, el_states; rng_band, nchunks_threads, kwargs...)
    mpi_isroot() && println("Setting up ElectronSelfEnergyCalculatorX2X")

    if !ElectronPhonon.chemical_potential_is_computed(calc.occ)
        el, _ = ElectronPhonon.electron_states_to_BTStates(el_states, kpts)
        ElectronPhonon.bte_compute_μ!(calc.occ, el; do_print=true)
    end

    nband = length(rng_band)
    resize!(calc.Σ, kpts.n)
    for i in eachindex(calc.Σ)
        calc.Σ[i] = OffsetArray(zeros(ComplexF64, nband, length(calc.occ)), rng_band, :)
    end

    resize!(calc.Σ_buffer, nchunks_threads)
    for i in eachindex(calc.Σ_buffer)
        calc.Σ_buffer[i] = OffsetArray(zeros(ComplexF64, nband, length(calc.occ)), rng_band, :)
    end

    calc
end

function ElectronPhonon.setup_calculator_inner!(calc::ElectronSelfEnergyCalculatorX2X; ik, kwargs...)
    for buf in calc.Σ_buffer
        buf .= 0
    end

    calc
end

function ElectronPhonon.postprocess_calculator_inner!(calc::ElectronSelfEnergyCalculatorX2X; ik, kwargs...)
    for buf in calc.Σ_buffer
        calc.Σ[ik] .+= buf
    end
    calc
end

function ElectronPhonon.postprocess_calculator!(calc::ElectronSelfEnergyCalculatorX2X; kwargs...)
    calc
end

function ElectronPhonon.run_calculator!(calc::ElectronSelfEnergyCalculatorX2X, epdata, ik, iq, ikq ; id_chunk, xq, kwargs...)

    Σ = calc.Σ_buffer[id_chunk]

    η = calc.smearing
    (; el_k, el_kq, ph, wtq) = epdata

    for (iocc, (; μ, T)) in enumerate(calc.occ)

        for imode in 1:ph.nmodes
            ωq = ph.e[imode]
            ωq < ElectronPhonon.omega_acoustic && continue
            nq = occ_boson(ωq, T)

            for ib in el_k.rng, jb in el_kq.rng
                ek = el_k.e[ib]
                ekq = el_kq.e[jb]
                ef_1 = ekq - ωq
                ef_2 = ekq + ωq

                # Lifetime applied to phonon
                fkq = occ_fermion(ekq - μ, T)
                f1 = nq + fkq
                f2 = nq + 1 - fkq

                # Lifetime applied to electron
                # To use this, the real part needs to be modified using digamma.
                # f1 = nq + occ_fermion(ek + ωq - μ, T)
                # f2 = nq + 1 - occ_fermion(ek - ωq - μ, T)

                g2 = epdata.g2[jb, ib, imode]

                # Skip coupling with oneself or between degenerate states at the same k point
                if all(abs.(xq) .< 1e-8) && abs(ek - ekq) < 2e-5
                    continue
                end

                Σ_coeff = f1 / (ek - ef_1 + im * η) + f2 / (ek - ef_2 + im * η)
                Σ[ib, iocc] += Σ_coeff * g2 * wtq
            end
        end
    end
    calc
end



function compute_self_energy_rest_space(
    kpts, occupation_params, model, folder, outdir, prefix,
    ;
    η_rest,
    window_active,
    window_wfpt,
    nqs_rest,
    )

    @time wfpt_objs = ElectronPhonon.load_wfpt(folder, outdir, prefix, model.nw, model.nmodes, model.recip_lattice, model.alat);

    # 1. Outside frozen window: use WFPT
    Σ_wfpt = ElectronPhonon.run_wfpt(folder, outdir, prefix, model, window_wfpt, kpts, occupation_params, wfpt_objs)

    # 2. Fan, inside frozen window, but outside active space: use Wannier interpolation, dense grid
    kqpts_full = kpoints_grid(nqs_rest)
    @info "kpts.n = $(kpts.n)"
    @info "kqpts_full.n = $(kqpts_full.n)"

    calc_full = ElectronSelfEnergyCalculatorX2X(; occ = occupation_params, smearing = η_rest)
    calc_acti = ElectronSelfEnergyCalculatorX2X(; occ = occupation_params, smearing = η_rest)

    # Compute Σ from states inside window_wfpt, and subtract contributions from states inside window_active
    @time res = ElectronPhonon.run_eph_over_k_and_kq(model, kpts, kqpts_full, [calc_full]; window_kq = window_wfpt)
    Σ_rest_fan = copy(calc_full.Σ)

    if window_active !== nothing
        @time res = ElectronPhonon.run_eph_over_k_and_kq(model, kpts, kqpts_full, [calc_acti]; window_kq = window_active)
        Σ_rest_fan .-= calc_acti.Σ
    end

    # 3. Debye-Waller, inside frozen window
    # We do not subtract active space contribution because it is static so we don't compute
    # it self-consistently.
    qpts_full = kqpts_full
    @time Σ_rest_dw = ElectronPhonon.compute_self_energy_active_DW(folder, outdir, prefix, model, window_wfpt, kpts, qpts_full, occupation_params, wfpt_objs)

    Σ_rest = Σ_wfpt.Σ .+ Σ_rest_fan .+ Σ_rest_dw

    (; Σ_rest, Σ_wfpt, Σ_rest_fan, Σ_rest_dw)
end
