
import ElectronPhonon.OffsetArrays: OffsetArray, OffsetMatrix, no_offset_view

Base.@kwdef struct ElectronSpectralCalculatorX6X{IT}
    # --- Parameter fields ---
    # Electron occupation
    occ :: ElectronOccupationParams

    # Frequencies to compute the self-energy
    ωlist :: Vector{Float64}

    # smearing
    smearing :: Float64

    # Input self-energy for a self-consistent calculation
    ωs_input :: StepRangeLen{Float64, Base.TwicePrecision{Float64}, Base.TwicePrecision{Float64}, Int64} = 0.:-1.
    Σs_input :: Array{ComplexF64, 4} = zeros(ComplexF64, 0, 0, 1, 1)

    Σs_input_itp :: IT = [[linear_interpolation((ωs_input,),
        zeros(ComplexF64, length(ωs_input)); extrapolation_bc = Flat()) for _ in 0:0]]

    # --- Data fields ---
    # Frequency-dependent self-energy of electrons
    Σ :: Vector{OffsetArray{ComplexF64, 3, Array{ComplexF64, 3}}} = [OffsetArray(zeros(ComplexF64, 0, 0, 0))]

    # --- Data buffer fields ---
    Σ_buffer :: Vector{OffsetArray{ComplexF64, 3, Array{ComplexF64, 3}}} = [OffsetArray(zeros(ComplexF64, 0, 0, 0))]

    # --- Preallocated buffer fields ---
    Akq_1 :: Vector{Vector{ComplexF64}} = [zeros(ComplexF64, 0)]
    Akq_2 :: Vector{Vector{ComplexF64}} = [zeros(ComplexF64, 0)]
    focc_1 :: Vector{Vector{ComplexF64}} = [zeros(ComplexF64, 0)]
    focc_2 :: Vector{Vector{ComplexF64}} = [zeros(ComplexF64, 0)]

    # DEBUG field
    method :: Symbol = :Original
end

ElectronPhonon.allow_eph_outer_k(:: ElectronSpectralCalculatorX6X) = true


function ElectronPhonon.setup_calculator!(calc::ElectronSpectralCalculatorX6X, kpts, qpts, el_states; rng_band, nchunks_threads, kwargs...)
    mpi_isroot() && println("Setting up ElectronSpectralCalculatorX6X")
    (; ωs_input) = calc

    if !ElectronPhonon.chemical_potential_is_computed(calc.occ)
        el, _ = ElectronPhonon.electron_states_to_BTStates(el_states, kpts)
        ElectronPhonon.bte_compute_μ!(calc.occ, el; do_print=true)
    end

    nband = length(rng_band)
    resize!(calc.Σ, kpts.n)
    for i in eachindex(calc.Σ)
        calc.Σ[i] = OffsetArray(zeros(ComplexF64, length(calc.ωlist), nband, length(calc.occ)), :, rng_band, :)
    end

    resize!(calc.Σ_buffer, nchunks_threads)
    for i in eachindex(calc.Σ_buffer)
        calc.Σ_buffer[i] = OffsetArray(zeros(ComplexF64, length(calc.ωlist), nband, length(calc.occ)), :, rng_band, :)
    end

    # Preallocate buffers
    resize!(calc.Akq_1, nchunks_threads)
    resize!(calc.Akq_2, nchunks_threads)
    resize!(calc.focc_1, nchunks_threads)
    resize!(calc.focc_2, nchunks_threads)
    for i in eachindex(calc.Akq_1)
        calc.Akq_1[i] = zeros(ComplexF64, length(calc.ωlist))
        calc.Akq_2[i] = zeros(ComplexF64, length(calc.ωlist))
        calc.focc_1[i] = zeros(ComplexF64, length(calc.ωlist))
        calc.focc_2[i] = zeros(ComplexF64, length(calc.ωlist))
    end

    # Preallocate Σs_input_itp : OffsetVector of 1d interpolators for frequency mesh ωs_input
    empty!(calc.Σs_input_itp)

    η = calc.smearing
    Σs_dummy = [-im * η, -im * η]
    for _ in 1:nchunks_threads
        push!(calc.Σs_input_itp, eltype(calc.Σs_input_itp)[])

        for _ in 1 : (rng_band[1] - 1)
            # Dummy interpolator (not to use OffsetArray)
            push!(calc.Σs_input_itp[end], linear_interpolation(0.:1.:1., Σs_dummy; extrapolation_bc = Flat()))
        end

        for _ in rng_band
            if length(ωs_input) > 0
                push!(calc.Σs_input_itp[end], linear_interpolation((ωs_input,),
                    zeros(ComplexF64, length(ωs_input)); extrapolation_bc = Flat()))
            else
                # No input self-energy. Set dummy Σ_itp_list which always gives -im * η
                push!(calc.Σs_input_itp[end], linear_interpolation(0.:1.:1., Σs_dummy; extrapolation_bc = Flat()))
            end
        end
    end

    calc
end

function ElectronPhonon.setup_calculator_inner!(calc::ElectronSpectralCalculatorX6X; ik, kwargs...)
    for buf in calc.Σ_buffer
        buf .= 0
    end

    calc
end

function ElectronPhonon.postprocess_calculator_inner!(calc::ElectronSpectralCalculatorX6X; ik, kwargs...)
    for buf in calc.Σ_buffer
        calc.Σ[ik] .+= buf
    end
    calc
end

function ElectronPhonon.postprocess_calculator!(calc::ElectronSpectralCalculatorX6X; kwargs...)
    # Compute the real part using Kramers-Kronig relation
    Threads.@threads for Σ in calc.Σ
        @views for iocc in axes(Σ, 3), n in axes(Σ, 2)
            Σ[:, n, iocc] .+= kramers_kronig(calc.ωlist, imag.(Σ[:, n, iocc]))
        end
    end
    calc
end

function ElectronPhonon.run_calculator!(calc::ElectronSpectralCalculatorX6X, epdata, ik, iq, ikq ; id_chunk, kwargs...)

    Σs_input_itp = calc.Σs_input_itp[id_chunk]
    Σ = calc.Σ_buffer[id_chunk]

    η = calc.smearing
    ωlist = calc.ωlist

    (; el_k, el_kq, ph, wtq) = epdata

    # Preallocate buffers
    Akq_1 = calc.Akq_1[id_chunk]
    Akq_2 = calc.Akq_2[id_chunk]
    focc_1 = calc.focc_1[id_chunk]
    focc_2 = calc.focc_2[id_chunk]

    for (iocc, (; μ, T)) in enumerate(calc.occ)

        # Setup interpolator for input self-energy
        if length(calc.ωs_input) > 0
            # Make |Im Σ| at least η
            for m in el_kq.rng
                Σs_mkq = view(calc.Σs_input, :, m, ikq, iocc)
                parent(parent(parent(Σs_input_itp[m]))) .= @. complex(real(Σs_mkq), min(imag(Σs_mkq), -η))
            end
        end

        for imode in 1:ph.nmodes
            ωq = ph.e[imode]
            ωq < ElectronPhonon.omega_acoustic && continue
            nq = occ_boson(ωq, T)

            # ωq < 15 * unit_to_aru(:meV) && continue

            # Precompute occupation f(ω + ωq) and f(ω - ωq)
            @. focc_1 = occ_fermion(ωlist + ωq - μ, T)
            @. focc_2 = occ_fermion(ωlist - ωq - μ, T)

            for m in el_kq.rng
                ekq = el_kq.e[m]
                Σ_itp = Σs_input_itp[m]

                # Precompute spectral function at (m, k+q) multipled by occupation factors
                @. Akq_1 = imag(1 / (ωlist + ωq - ekq - Σ_itp(ωlist + ωq))) * (nq + focc_1)
                @. Akq_2 = imag(1 / (ωlist - ωq - ekq - Σ_itp(ωlist - ωq))) * (nq + 1 - focc_2)

                @views for n in el_k.rng
                    g2 = epdata.g2[m, n, imode]
                    fac = g2 * wtq * im
                    @. Σ[:, n, iocc] += (Akq_1 + Akq_2) * fac
                end
            end
        end
    end
    calc
end
