using Base.Threads
using Interpolations

struct SelfEnergyInterpolatorX3X{IT}
    Σs :: Array{ComplexF64, 4}
    ωs :: Vector{Float64}
    iks_to_ik_irr :: Array{Int64, 3}
    Σs_all :: Vector{Array{ComplexF64, 4}}
    Σs_all_itp :: IT
end

"""
    SelfEnergyInterpolatorX5X{IT}
Perform 3-dimensional interpolation of the self-energy on the full BZ.

- `Σs` : temperature- and band-index-dependent self-energy, only for irreducible k-points
inside the active-space window.
- `Σs_all` : `nω * nk * nk * nk` array used for interpolation. To save memory, we store only
a single instance of it.
To perform interpolation, first call `_set_Σs_all_itp!(Σs_itp, iT, ib)` to set `Σs_all`.
Then, call `Σs_itp(ω :: Float64, xk :: Vec3{Float64})` to get the interpolated value.
"""
mutable struct SelfEnergyInterpolatorX5X{IT}
    iband :: Int
    const Σs :: Array{ComplexF64, 4}
    const ωs :: Vector{Float64}
    const iks_to_ik_irr :: Array{Int64, 3}
    const Σs_all :: Array{ComplexF64, 4}
    const Σs_all_itp :: IT
end

function compute_Σs_itp(kpts, ωs, Σs, symmetry)
    # Interpolator for the self-energy on the full BZ.
    # Interpolator for the self-energy: Σs_all_itp[ib](ω, kx, ky, kz)
    # The parent array Σs_all[ib] is updated for each T.
    @assert size(Σs, 1) == length(ωs)

    # Setup the unfolding map. Size nk+1 to include both 0 and 1.
    nks = kpts.ngrid
    iks_to_ik_irr = zeros(Int, kpts.ngrid .+ 1)
    for (ik, xk) in enumerate(kpts.vectors)
        for symop in symmetry
            sxk = symop.is_tr ? -symop.S * xk : symop.S * xk
            sxk = ElectronPhonon.normalize_kpoint_coordinate.(sxk)
            iks = round.(Int, sxk .* nks) .+ 1
            iks_to_ik_irr[iks...] = ik
        end
    end
    iks_to_ik_irr[end, :, :] .= iks_to_ik_irr[1, :, :]
    iks_to_ik_irr[:, end, :] .= iks_to_ik_irr[:, 1, :]
    iks_to_ik_irr[:, :, end] .= iks_to_ik_irr[:, :, 1]

    # nband = size(Σs, 2)
    # Σs_all = [zeros(ComplexF64, length(ωs), (nks.+1)...) for _ in 1:nband]  # Bottleneck

    # # Interpolator for the self-energy: Σs_all_itp[ib](ω, kx, ky, kz)
    # # The parent array Σs_all[ib] is updated for each T.
    # Σs_all_itp = [interpolate!((ωs, range(0, 1, length=nks[1]+1),
    #                                 range(0, 1, length=nks[2]+1),
    #                                 range(0, 1, length=nks[3]+1)),
    #                             Σs_all[ib], Gridded(Linear())) for ib in 1:nband]

    # SelfEnergyInterpolatorX3X(Σs, ωs, iks_to_ik_irr, Σs_all, Σs_all_itp)

    Σs_all = zeros(ComplexF64, length(ωs), (nks.+1)...)

    # Interpolator for the self-energy: Σs_all_itp[ib](ω, kx, ky, kz)
    # The parent array Σs_all[ib] is updated for each T.
    Σs_all_itp = interpolate!((ωs, range(0, 1, length=nks[1]+1),
                                   range(0, 1, length=nks[2]+1),
                                   range(0, 1, length=nks[3]+1)),
                               Σs_all, Gridded(Linear()))

    SelfEnergyInterpolatorX5X(-1, Σs, ωs, iks_to_ik_irr, Σs_all, Σs_all_itp)
end

function _set_Σs_all_itp!(Σs_itp, iT)
    # Unfold self-energy to the full BZ, update Σs_all.
    (; Σs, iks_to_ik_irr, Σs_all) = Σs_itp
    nband = size(Σs, 2)
    for iks in CartesianIndices(size(Σs_all[1])[2:4])
        ik = iks_to_ik_irr[iks]
        ik == 0 && continue
        @views for ib in 1:nband
            Σs_all[ib][:, iks] .= Σs[:, ib, ik, iT]
        end
    end
    Σs_itp
end

function _get_Σs_itp_1d(Σs_itp, ib, xk)
    # Return a 1d interpolator for frequency interpolation
    ωs = Σs_itp.ωs
    Σs_tmp = Σs_itp.Σs_all_itp[ib].(ωs, xk...)
    extrapolate(interpolate!((ωs,), Σs_tmp, Gridded(Linear())), 0)
end

function (Σs_itp::SelfEnergyInterpolatorX3X)(ib :: Int, ω :: Float64, xk :: Vec3{Float64})
    if Σs_itp.ωs[1] <= ω <= Σs_itp.ωs[end]
        Σs_itp.Σs_all_itp[ib](ω, xk...)
    else
        eltype(Σs_itp.Σs)(0)
    end
end


function _set_Σs_all_itp!(Σs_itp :: SelfEnergyInterpolatorX5X, iT, ib)
    # Unfold self-energy to the full BZ, update Σs_all.
    (; Σs, iks_to_ik_irr, Σs_all) = Σs_itp
    Σs_itp.iband = ib
    @views for iks in CartesianIndices(size(Σs_all)[2:4])
        ik = iks_to_ik_irr[iks]
        ik == 0 && continue
        Σs_all[:, iks] .= Σs[:, ib, ik, iT]
    end
    Σs_itp
end

function _get_Σs_itp_1d(Σs_itp :: SelfEnergyInterpolatorX5X, ib, xk)
    throw(ArgumentError("Not implemented"))
    # # Return a 1d interpolator for frequency interpolation
    # ωs = Σs_itp.ωs
    # Σs_tmp = Σs_itp.Σs_all_itp[ib].(ωs, xk...)
    # extrapolate(interpolate!((ωs,), Σs_tmp, Gridded(Linear())), 0)
end

function (Σs_itp::SelfEnergyInterpolatorX5X)(ib :: Int, ω :: Float64, xk :: Vec3{Float64})
    @assert ib == Σs_itp.iband
    if Σs_itp.ωs[1] <= ω <= Σs_itp.ωs[end]
        Σs_itp.Σs_all_itp(ω, xk...)
    else
        eltype(Σs_itp.Σs)(0)
    end
end

function (Σs_itp::SelfEnergyInterpolatorX5X)(ω :: Float64, xk :: Vec3{Float64})
    if Σs_itp.ωs[1] <= ω <= Σs_itp.ωs[end]
        Σs_itp.Σs_all_itp(ω, xk...)
    else
        eltype(Σs_itp.Σs)(0)
    end
end


"""
- `truncate_spectral` : Truncate the spectral function outside the window. Used for one-shot FM
calculations for lightly doped semiconductors. Can be computed using `spectral_window_for_semiconductors`.
"""
# function _compute_conductivity(Ω, occ, ωs, Σs, nk, iks_to_ik_irr, Σs_all, kpts, eks, vks, basis, window_active)
function compute_conductivity_bubble(Ω, occ, Σs_itp :: SelfEnergyInterpolatorX5X, kpts, eks, vks, basis, window_active, model; truncate_spectral = false)
    # TODO: Remove. Replace with one in scKubo_bubble (using BTStates)
    nband = size(eks, 1)
    ωgrid = grid_points(basis)

    # Basis factor for ∫ dν bᵢ(ν)
    basis_sum = @. (basis.xs[3:end] .- basis.xs[1:end-2]) / 2

    dfermi = zeros(length(ωgrid))
    σks = [zeros(3, 3, length(occ)) for _ in 1:nband, _ in 1:kpts.n]
    jdosks = [zeros(length(occ)) for _ in 1:nband, _ in 1:kpts.n]

    for (iT, (; μ, T)) in enumerate(occ)

        if Ω == 0
            @. dfermi = -docc_fermion(ωgrid - μ, T)
        else
            @. dfermi = (occ_fermion(ωgrid - Ω/2 - μ, T) - occ_fermion(ωgrid + Ω/2 - μ, T)) / Ω
        end

        if truncate_spectral
            emin, emax = spectral_window_for_semiconductors(Σs_itp.ωs, Σs_itp.Σs, occ, eks, window_active, iT)
            dfermi[ωgrid .< emin] .= 0
            dfermi[ωgrid .> emax] .= 0
        end

        for ib in 1:nband
            _set_Σs_all_itp!(Σs_itp, iT, ib)

            Threads.@threads for ik in 1:kpts.n
                xk = kpts.vectors[ik]
                ek = eks[ib, ik]
                vk = vks[ib, ik]

                # Skip states outside the activate space window
                window_active[1] <= ek <= window_active[2] || continue

                σk_int = 0.0
                @views for (iω, ω) in enumerate(ωgrid)
                    ωm = ω - Ω/2
                    ωp = ω + Ω/2
                    Am = -imag(1 / (ωm - ek - Σs_itp(ωm, xk))) / π
                    Ap = -imag(1 / (ωp - ek - Σs_itp(ωp, xk))) / π
                    σk_int += Am * Ap * dfermi[iω] * basis_sum[iω] * π
                end

                # Get a 3*3 conductivity tensor
                σk = real.(vk * vk') * σk_int

                σks[ib, ik][:, :, iT] .= σk * kpts.weights[ik]
                jdosks[ib, ik][iT] = σk_int * kpts.weights[ik]
            end
        end
    end
    σ = sum(σks)
    jdos = sum(jdosks)

    # Symmetrize conductivity tensor
    σ = symmetrize_array(σ, model.symmetry; order = 2)

    σ .*= occ.spin_degeneracy ./ model.volume
    jdos .*= occ.spin_degeneracy ./ model.volume

    (; σ, jdos, σks, jdosks)
end

function compute_conductivity_bubble_using_A(Ω, occ, ωs, As, kpts, eks, vks, basis, window_active, model)
    nband = size(eks, 1)
    ωgrid = grid_points(basis)
    @assert size(As, 1) == length(ωs)
    @assert size(As, 2) == nband
    @assert size(As, 3) == kpts.n
    @assert size(As, 4) == length(occ)

    # Basis factor for ∫ dν bᵢ(ν)
    basis_sum = @. (basis.xs[3:end] .- basis.xs[1:end-2]) / 2

    dfermi = zeros(length(ωgrid))
    σks = [zeros(3, 3, length(occ)) for _ in 1:nband, _ in 1:kpts.n]
    jdosks = [zeros(length(occ)) for _ in 1:nband, _ in 1:kpts.n]

    for (iT, (; μ, T)) in enumerate(occ)

        if Ω == 0
            @. dfermi = -docc_fermion(ωgrid - μ, T)
        else
            @. dfermi = (occ_fermion(ωgrid - Ω/2 - μ, T) - occ_fermion(ωgrid + Ω/2 - μ, T)) / Ω
        end

        Threads.@threads for ik in 1:kpts.n
            for ib in 1:nband
                xk = kpts.vectors[ik]
                ek = eks[ib, ik]
                vk = vks[ib, ik]

                As_itp = extrapolate(interpolate!((ωs,), As[:, ib, ik, iT], Gridded(Linear())), 0)

                # Skip states outside the activate space window
                window_active[1] <= ek <= window_active[2] || continue

                σk_int = 0.0
                @views for (iω, ω) in enumerate(ωgrid)
                    ωm = ω - Ω/2
                    ωp = ω + Ω/2
                    Am = As_itp(ωm)
                    Ap = As_itp(ωp)
                    σk_int += Am * Ap * dfermi[iω] * basis_sum[iω] * π
                end

                # Get a 3*3 conductivity tensor
                σk = real.(vk * vk') * σk_int

                σks[ib, ik][:, :, iT] .= σk * kpts.weights[ik]
                jdosks[ib, ik][iT] = σk_int * kpts.weights[ik]
            end
        end
    end
    σ = sum(σks)
    jdos = sum(jdosks)

    # Symmetrize conductivity tensor
    σ = symmetrize_array(σ, model.symmetry; order = 2)

    σ .*= occ.spin_degeneracy ./ model.volume
    jdos .*= occ.spin_degeneracy ./ model.volume

    (; σ, jdos, σks)
end


function _compute_conductivity_dilute(Ω, occ, ωs, Σs, nk, iks_to_ik_irr, Σs_all, kpts, eks, vks, basis, window_active)
    # Compute conductivity for the dilute limit
    # Scan μ in [emin - 10T, emin] and compute conductivity for each μ.
    # FIXME: Hole doping is not supported yet.
    @assert all(occ.μlist .== -Inf)
    e0 = minimum(eks)

    num_μ = 7

    nband = size(eks, 1)
    ωgrid = grid_points(basis)

    # Basis factor for ∫ dν bᵢ(ν)
    basis_sum = @. (basis.xs[3:end] .- basis.xs[1:end-2]) / 2

    # Interpolator for the self-energy: Σs_all_itp[ib](ω, kx, ky, kz)
    # The parent array Σs_all[ib] is updated for each T.
    Σs_all_itp = [interpolate!((ωs, range(0, 1, length=nk+1), range(0, 1, length=nk+1), range(0, 1, length=nk+1)), Σs_all[ib], Gridded(Linear())) for ib in 1:nband]

    σks = [zeros(3, 3, length(occ), num_μ) for _ in 1:nband, _ in 1:kpts.n]
    nks = [zeros(length(occ), num_μ) for _ in 1:nband, _ in 1:kpts.n]
    μ_scan_list = []

    for (iT, (; T)) in enumerate(occ)

        # Unfold self-energy to the full BZ, update Σs_all.
        for iks in CartesianIndices((nk+1, nk+1, nk+1))
            ik = iks_to_ik_irr[iks]
            ik == 0 && continue
            for ib in 1:nband
                Σs_all[ib][:, iks] .= Σs[:, ib, ik, iT]
            end
        end

        μ_scan = range(minimum(ωs), minimum(eks), length=num_μ)
        push!(μ_scan_list, μ_scan)

        for (iμ, μ) in enumerate(μ_scan)

            # if Ω == 0
            #     dfermi = @. -docc_fermion(ωgrid - μ, T)
            # else
            #     dfermi = @. (occ_fermion(ωgrid - Ω/2 - μ, T) - occ_fermion(ωgrid + Ω/2 - μ, T)) / Ω
            # end

            # fermi = @. occ_fermion(ωgrid - μ, T)

            # Occupation function f(ω) = df_FD(ω, μ) / dμ * exp(-μ/T) in the μ -> -Inf limit
            fermi = @. exp(-(ωgrid .- e0) / T) / T
            dfermi = @. exp(-(ωgrid .- e0) / T) / T^2

            fermi[ωgrid .< μ] .= 0
            dfermi[ωgrid .< μ] .= 0

            Threads.@threads for ik in 1:kpts.n
                for ib in 1:nband
                    xk = kpts.vectors[ik]
                    ek = eks[ib, ik]
                    vk = vks[ib, ik]

                    # Skip states outside the activate space window
                    window_active[1] <= ek <= window_active[2] || continue

                    Σs_tmp = Σs_all_itp[ib].(ωs, xk...)
                    Σs_itp_1d = extrapolate(interpolate!((ωs,), Σs_tmp, Gridded(Linear())), 0)

                    σk_int = 0.0
                    @views for (iω, ω) in enumerate(ωgrid)
                        ωm = ω - Ω/2
                        ωp = ω + Ω/2
                        Am = -imag(1 / (ωm - ek - Σs_itp_1d(ωm))) / π
                        Ap = -imag(1 / (ωp - ek - Σs_itp_1d(ωp))) / π
                        σk_int += Am * Ap * dfermi[iω] * basis_sum[iω] * π
                    end

                    nk_int = 0.0
                    @views for (iω, ω) in enumerate(ωgrid)
                        A = -imag(1 / (ω - ek - Σs_itp_1d(ω))) / π
                        nk_int += A * fermi[iω] * basis_sum[iω]
                    end

                    # Get a 3*3 conductivity tensor
                    σk = real.(vk * vk') * σk_int

                    σks[ib, ik][:, :, iT, iμ] .= σk * kpts.weights[ik]

                    nks[ib, ik][iT, iμ] = nk_int * kpts.weights[ik]
                end
            end
        end
    end

    (; σks, nks, μ_scan_list)
end



function iterate_current_Dyson(J_c, filename_btedata, ik_and_isym_map, el_i, el_f, ph, basis, Gm_c, Gp_c, model, T)
    fid = h5open(filename_btedata, "r")

    tmp_DJ = zero.(J_c)
    J_new_c = zero.(J_c)

    group_scattering = open_group(fid, "scattering")

    @threads for key in keys(group_scattering)
        if threadid() == 1
            @info "$(now()) key = $key"
            flush(stdout)
            flush(stderr)
        end
        g = open_group(group_scattering, key)
        scat = load_BTData(g, ElectronPhonon.ElPhScatteringData{Float64})

        D_c = [zeros(ComplexF64, basis.N, basis.N) for _ in 1:2, _ in 1:2]
        D_map_buffer = zeros(length(basis.xs))
        J_c_f = zeros(Vec3{ComplexF64}, basis.N, 2, 2, 2)

        for (iscat, s) in enumerate(scat)
            s.sign_ph == 1 || continue

            (; ind_el_i, ind_el_f, ind_ph) = s

            e_i = el_i[ind_el_i]
            e_f = el_f[ind_el_f]
            ω_ph = ph.e[ind_ph]
            g2 = s.mel

            ω_ph < ElectronPhonon.omega_acoustic && continue

            EPSpectral.get_D_map_opt!(D_c, basis, ω_ph, T, D_map_buffer)

            fac = e_f.k_weight * g2

            i, isym = ik_and_isym_map[ind_el_f]
            symop = model.symmetry[isym]
            if symop.is_tr
                J_c_f .= Ref(-symop.Scart) .* J_c[i]
            else
                J_c_f .= Ref(+symop.Scart) .* J_c[i]
            end

            @views for c0 in 1:2, c2 in 1:2, c1 in 1:2
                mul!(tmp_DJ[ind_el_i][:, c1, c2, c0], D_c[c2, c1], J_c_f[:, c1, c2, c0], fac, 1)
            end
        end
    end
    close(fid)

    for i in 1:el_i.n
        # (-1)^{δ_{c,+}} sign factor when attaching Green function to vertex
        tmp_DJ[i][:, 2, :, :] .*= -1
        tmp_DJ[i][:, :, 2, :] .*= -1

        @views for c0 in 1:2, c1 in 1:2, c2 in 1:2, cp1 in 1:2, cp2 in 1:2
            @. J_new_c[i][:, c1, c2, c0] += Gm_c[i][:, c1, cp1] * tmp_DJ[i][:, cp1, cp2, c0] * Gp_c[i][:, cp2, c2]
        end
    end

    J_new_c .*= -1

    J_new_c
end;



function amputate_current_correlator(J_c, Gm_c, Gp_c)
    v_c = zero.(J_c)
    for i in eachindex(J_c)
        for c0 in 1:2, iω in axes(J_c[i], 1)
            Gm = SMatrix{2, 2}(Gm_c[i][iω, :, :])
            Gp = SMatrix{2, 2}(Gp_c[i][iω, :, :])
            J = SMatrix{2, 2}(J_c[i][iω, :, :, c0])
            v_c[i][iω, :, :, c0] = inv(Gm) * J * inv(Gp)
        end
    end
    v_c
end

function compute_conductivity(Ω, el, basis, v0_c, J_c, T)
    # Mobility using the current correlator

    # Basis factor for ∫ dν bᵢ(ν) / 2πi
    basis_sum = @. (basis.xs[3:end] .- basis.xs[1:end-2]) / 2 / (2π * im)

    χ_bubble = zeros(ComplexF64, 3, 3, 2, 2, el.n)

    for i in 1:el.n
        weight = el[i].k_weight

        for iω in eachindex(grid_points(basis))
            # Use bare current vertex v0_c, renormalized correlator J_c
            for c2 in 1:2, c1 in 1:2
                vL = v0_c[i][iω]
                JR = J_c[i][iω, c1, c1, c2]
                χ_bubble[:, :, c1, c2, i] += (conj.(vL) * transpose(JR)) * basis_sum[iω] * weight
            end
        end
    end

    if Ω == 0.
        freq_fac = im / 4T
    else
        freq_fac = -1 / (2im * Ω * (1 + 2 * occ_boson(Ω, T)))
    end
    σk = @. real((χ_bubble[:, :, 1, 2, :] + χ_bubble[:, :, 2, 1, :]) * freq_fac)

    σ = dropdims(sum(σk, dims=3), dims=3)

    (; σ, σk)
end


"""
Build a map that unfolds a velocity (polar, time-reversal odd) on `el_i` to `el_f` by `symmetry`.
``S(ik_and_isym_map[i][2]) * el_i[ik_and_isym_map[i][1]] -> el_f[i]``
"""
function find_unfolding_indices(el_i :: ElectronPhonon.BTStates, el_f :: ElectronPhonon.BTStates, symmetry)
    ik_and_isym_map = fill((0, 0), el_f.n)

    for i in 1:el_f.n
        xk_f = el_f[i].xks
        ib = el_f[i].iband

        found = false

        for (isym, (; S, is_tr)) in enumerate(symmetry)
            for j in 1:el_i.n
                if el_i[j].iband == ib
                    k_i = el_i[j].xks
                    Sk_i = is_tr ? -S * k_i : S * k_i
                    dk = Sk_i - xk_f
                    if all(abs.(dk .- round.(dk)) .< 1e-6)
                        ik_and_isym_map[i] = (j, isym)
                        found = true
                        break
                    end
                end
            end
        end

        found || error("Cannot find the corresponding state in el_i for i = $i, e_f = $(el_f[i])")
    end

    ik_and_isym_map
end


function get_dG_contour_dilute(ω, ε, Σ, T, ω_cutoff, dΣ = 0; ω0 = 0)
    # dG/dμ * exp(-μ/T) in the μ -> -Inf limit

    f = 0.0
    Gᴿ = 1 / (ω - ε - Σ)

    # Contribution of dΣ
    dGᴿ = dΣ * Gᴿ^2
    dGᴬ = conj(dGᴿ)
    dGᴷ = 2im * imag(dGᴿ) * (1 - 2f)

    # Contribution of df. Nonzero only for dGᴷ
    df = ω < ω_cutoff ? 0.0 : exp(-(ω - ω0) / T) / T
    dGᴷ += 2im * imag(Gᴿ) * -2df
    SMatrix{2, 2}(dGᴿ + dGᴬ + dGᴷ, dGᴿ - dGᴬ + dGᴷ, -dGᴿ + dGᴬ + dGᴷ, -dGᴿ - dGᴬ + dGᴷ) / 2
end

function compute_conductivity_dilute(Ω, Σs_itp_1d, basis, el, μ, T, symmetry, vL_c, vR_c)
    # Mobility in the dilute limit
    @assert μ == -Inf
    ω_cutoff = minimum(basis.xs)
    ω0 = minimum(el.e)

    # Basis factor for ∫ dν bᵢ(ν) / 2πi
    basis_sum = @. (basis.xs[3:end] .- basis.xs[1:end-2]) / 2 / 2π / im

    χ_ks = map(1:el.n) do i
        εk = el_i[i].e
        vk = el_i[i].vdiag

        # ib = el_i[i].iband
        # ik = xk_to_ik(el_i[i].xks, kpts)
        # ek = el_i[i].e


        ΔΠ_k = zeros(ComplexF64, 2, 2, 2, 2)
        χ_k = zeros(ComplexF64, 3, 3, 2, 2)

        @views for (iω, ω) in enumerate(EPSpectral.grid_points(basis))
            ωm = ω - Ω/2
            ωp = ω + Ω/2
            Σm = Σs_itp_1d[i](ωm)
            Σp = Σs_itp_1d[i](ωp)
            Gm = EPSpectral.get_G_contour(ωm, εk, Σm, μ, T)
            Gp = EPSpectral.get_G_contour(ωp, εk, Σp, μ, T)
            dGm = get_dG_contour_dilute(ωm, εk, Σm, T, ω_cutoff; ω0)
            dGp = get_dG_contour_dilute(ωp, εk, Σp, T, ω_cutoff; ω0)
            for cs in CartesianIndices((2, 2, 2, 2))
                c1, c2, c3, c4 = cs.I
                ΔΠ_k[cs] = Gp[c4, c1] * dGm[c2, c3] + dGp[c4, c1] * Gm[c2, c3]
            end

            if vL_c === nothing
                # Bubble diagram, use bare vertex vk
                for c2 in 1:2, c1 in 1:2
                    χ_k[:, :, c1, c2] .+= basis_sum[iω] * ΔΠ_k[c1, c1, c2, c2] * (vk * vk')
                end
            else
                # Use dressed current vertex vL_c, vR_c
                for c2 in 1:2, c1 in 1:2
                    for c12 in 1:2, c11 in 1:2, c21 in 1:2, c22 in 1:2
                        vL = vL_c[i][iω, c11, c12, c1]
                        vR = vR_c[i][iω, c21, c22, c2]
                        χ_k[:, :, c1, c2] .+= basis_sum[iω] * ΔΠ_k[c11, c12, c21, c22] * (vL * vR')
                    end
                end
            end
        end
        symmetrize_array(χ_k, symmetry; order = 2)
    end :: Vector{Array{ComplexF64, 4}}

    χ = sum(χ_ks .* el.k_weight)

    if Ω == 0.
        freq_fac = im / 4T
    else
        freq_fac = -1 / (2im * Ω * (1 + 2 * occ_boson(Ω, T)))
    end
    σ = real((χ[:, :, 1, 2] + χ[:, :, 2, 1]) * freq_fac)

    (; σ, χ, χ_ks)
end


function _compute_occupation_dilute(el, Σs_itp_1d, μ, T, ω_cutoff, basis)
    # Compute conductivity for the dilute limit
    # FIXME: Hole doping is not supported yet.
    @assert μ == -Inf
    e0 = minimum(el.e)

    ωgrid = grid_points(basis)

    # Basis factor for ∫ dν bᵢ(ν)
    basis_sum = @. (basis.xs[3:end] .- basis.xs[1:end-2]) / 2

    # Occupation function f(ω) = df_FD(ω, μ) / dμ * exp(-μ/T) in the μ -> -Inf limit
    fermi = @. exp(-(ωgrid .- e0) / T) / T
    fermi[ωgrid .< ω_cutoff] .= 0

    nks = map(1:el.n) do i
        ek = el.e[i]

        nk_int = 0.0
        @views for (iω, ω) in enumerate(ωgrid)
            A = -imag(1 / (ω - ek - Σs_itp_1d[i](ω))) / π
            nk_int += A * fermi[iω] * basis_sum[iω]
        end

        nk_int
    end

    n = sum(nks .* el.k_weight)

    (; n, nks)
end



function dilute_mode(occ)
    if all(occ.μlist .== -Inf)
        return :Electron
    elseif all(occ.μlist .== +Inf)
        return :Hole
    else
        error("occ.μlist must be -Inf or Inf to use dilute conductivity")
    end
end

function dilute_occupation_function(ωgrid, ω_cutoff, e0, T, mode, Ω = 0.0; doping_mode = :Dilute)
    if doping_mode == :Dilute
        if mode == :Electron
            # Occupation function f(ω) = df_FD(ω, μ) / dμ * exp(-μ/T) in the μ -> -Inf limit
            fermi = @. exp(-(ωgrid .- e0) / T) / T
            fermi[ωgrid .< ω_cutoff] .= 0
            if Ω == 0
                # -df/dω
                dfermi = @. exp(-(ωgrid .- e0) / T) / T^2
                dfermi[ωgrid .< ω_cutoff] .= 0
            else
                # (f(ε - Ω/2) - f(ε + Ω/2)) / Ω
                f1 = @. exp(-(ωgrid - Ω/2 - e0) / T) / T
                f2 = @. exp(-(ωgrid + Ω/2 - e0) / T) / T
                f1[ωgrid .- Ω/2 .< ω_cutoff] .= 0
                f2[ωgrid .+ Ω/2 .< ω_cutoff] .= 0
                dfermi = @. (f1 - f2) / Ω
            end
        elseif mode == :Hole
            # Similarly for holes
            fermi = @. exp((ωgrid .- e0) / T) / T
            fermi[ωgrid .> ω_cutoff] .= 0
            if Ω == 0
                # -df/dω * -1
                dfermi = @. exp((ωgrid .- e0) / T) / T^2
                dfermi[ωgrid .> ω_cutoff] .= 0
            else
                # (f(ε - Ω/2) - f(ε + Ω/2)) / Ω * -1
                f1 = @. exp((ωgrid - Ω/2 - e0) / T) / T
                f2 = @. exp((ωgrid + Ω/2 - e0) / T) / T
                f1[ωgrid .- Ω/2 .> ω_cutoff] .= 0
                f2[ωgrid .+ Ω/2 .> ω_cutoff] .= 0
                dfermi = @. (f1 - f2) / Ω * -1
            end
        else
            error("Wrong mode $mode")
        end

    elseif doping_mode === :FD
        if mode == :Electrion
            fermi = @. occ_fermion(ωgrid .- ω_cutoff, T)
        elseif mode == :Hole
            fermi = @. 1 - occ_fermion(ωgrid .- ω_cutoff, T)
        else
            error("Wrong mode $mode")
        end

        if Ω == 0
            dfermi = @. -docc_fermion(ωgrid .- ω_cutoff, T)
        else
            f1 = @. occ_fermion(ωgrid - Ω/2 - ω_cutoff, T)
            f2 = @. occ_fermion(ωgrid + Ω/2 - ω_cutoff, T)
            dfermi = @. (f1 - f2) / Ω
        end

    else
        error("Unknown doping_mode $doping_mode")
    end
    (; fermi, dfermi)
end

function dilute_compute_occupation(kpts, eks, occ, Σs_itp :: SelfEnergyInterpolatorX3X, basis, window_active, ω_cutoff)
    mode = dilute_mode(occ)
    e0 = mode == :Electron ? minimum(eks) : maximum(eks)

    nband = size(eks, 1)
    ωgrid = grid_points(basis)

    # Basis factor for ∫ dν bᵢ(ν)
    basis_sum = @. (basis.xs[3:end] .- basis.xs[1:end-2]) / 2

    n = zeros(length(occ))

    for (iT, (; T)) in enumerate(occ)

        # Unfold self-energy to the full BZ, update Σs_all.
        _set_Σs_all_itp!(Σs_itp, iT)

        # Occupation function f(ω) = df_FD(ω, μ) / dμ * exp(-μ/T) in the μ -> -Inf limit
        fermi, dfermi = dilute_occupation_function(ωgrid, ω_cutoff, e0, T, mode)

        for ik in 1:kpts.n
            for ib in 1:nband
                xk = kpts.vectors[ik]
                ek = eks[ib, ik]

                # Skip states outside the activate space window
                window_active[1] <= ek <= window_active[2] || continue

                nk_int = 0.0
                @views for (iω, ω) in enumerate(ωgrid)
                    A = -imag(1 / (ω - ek - Σs_itp(ib, ω, xk))) / π
                    nk_int += A * fermi[iω] * basis_sum[iω]
                end

                n[iT] += nk_int * kpts.weights[ik]
            end
        end
    end

    n .*= occ.spin_degeneracy

    n
end;

function dilute_find_ω_cutoff(kpts, eks, occ, Σs_itp :: SelfEnergyInterpolatorX3X, basis, window_active; do_plot = false)
    # Compute the cutoff energy for the occupation function
    mode = dilute_mode(occ)
    e0 = mode == :Electron ? minimum(eks) : maximum(eks)
    num_ω_cutoff = 15

    ωs = Σs_itp.ωs
    if mode == :Electron
        ω_cutoff_scan = range(minimum(ωs), e0, length = num_ω_cutoff)
    else
        ω_cutoff_scan = range(e0, maximum(ωs), length = num_ω_cutoff)
    end

    n = zeros(length(occ), num_ω_cutoff)
    for (icutoff, ω_cutoff) in enumerate(ω_cutoff_scan)
        n[:, icutoff] .= dilute_compute_occupation(kpts, eks, occ, Σs_itp, basis, window_active, ω_cutoff)
    end

    # 2. Find a plateau in ω_cutoff vs occ has smallest slope in logarithmic scale
    ω_cutoff_list = zeros(length(occ))
    for iT in 1:length(occ)
        icutoff = argmin(abs.(log.(n[iT, 2:end]) .- log.(n[iT, 1:end-1])))
        ω_cutoff_list[iT] = (ω_cutoff_scan[icutoff] + ω_cutoff_scan[icutoff+1]) / 2
    end

    if do_plot
        for iT in 1:length(occ)
            plot(ω_cutoff_scan, n[iT, :], "o-")
            axvline(ω_cutoff_list[iT]; c = "C$(iT-1)")
        end
        yscale("log")
        fig = gcf(); display(fig); close(fig)
    end

    ω_cutoff_list
end;

function dilute_compute_conductivity_bubble(Ω, occ, Σs_itp :: SelfEnergyInterpolatorX3X, kpts, eks, vks, basis, window_active,
                                            ω_cutoff_list, model; doping_mode = :Dilute)
    # doping == :Dilute : Maxwell occupation, ω_cutoff_list is high energy cutoff
    # doping == :FD : Fermi-Dirac occupation. ω_cutoff_list is chemical potential
    mode = dilute_mode(occ)
    e0 = mode == :Electron ? minimum(eks) : maximum(eks)

    nband = size(eks, 1)
    ωgrid = grid_points(basis)

    # Basis factor for ∫ dν bᵢ(ν)
    basis_sum = @. (basis.xs[3:end] .- basis.xs[1:end-2]) / 2

    σ = zeros(3, 3, length(occ))
    n = zeros(length(occ))

    for (iT, (; T)) in enumerate(occ)
        ω_cutoff = ω_cutoff_list[iT]

        # Unfold self-energy to the full BZ, update Σs_all.
        _set_Σs_all_itp!(Σs_itp, iT)

        # Occupation function f(ω) = df_FD(ω, μ) / dμ * exp(-μ/T) in the μ -> -Inf limit
        fermi, dfermi = dilute_occupation_function(ωgrid, ω_cutoff, e0, T, mode, Ω; doping_mode)

        σ_iT, n_iT = tmapreduce(.+, 1:kpts.n) do ik
            σk = zeros(Mat3)
            nk = 0.0

            @views for ib in 1:nband
                xk = kpts.vectors[ik]
                ek = eks[ib, ik]
                vk = vks[ib, ik]

                # Skip states outside the activate space window
                window_active[1] <= ek <= window_active[2] || continue

                σk_int = 0.0
                for (iω, ω) in enumerate(ωgrid)
                    ωm = ω - Ω/2
                    ωp = ω + Ω/2
                    Am = -imag(1 / (ωm - ek - Σs_itp(ib, ωm, xk))) / π
                    Ap = -imag(1 / (ωp - ek - Σs_itp(ib, ωp, xk))) / π
                    σk_int += Am * Ap * dfermi[iω] * basis_sum[iω] * π
                end

                # Get a 3*3 conductivity tensor
                σk += real.(vk * vk') * σk_int * kpts.weights[ik]

                nk_int = 0.0
                for (iω, ω) in enumerate(ωgrid)
                    A = -imag(1 / (ω - ek - Σs_itp(ib, ω, xk))) / π
                    nk_int += A * fermi[iω] * basis_sum[iω]
                end
                nk += nk_int * kpts.weights[ik]
            end

            (σk, nk)
        end

        σ[:, :, iT] .= σ_iT
        n[iT] = n_iT
    end

    # Symmetrize conductivity tensor
    σ = symmetrize_array(σ, model.symmetry; order = 2)
    n = symmetrize_array(n, model.symmetry; order = 0)

    σ .*= occ.spin_degeneracy ./ model.volume
    n .*= occ.spin_degeneracy ./ model.volume

    (; σ, n)
end
