
import ElectronPhonon.OffsetArrays: OffsetArray, OffsetMatrix, no_offset_view

Base.@kwdef mutable struct MatrixElementCalculatorX4X
    # --- Data fields ---
    # Electron and phonon eigenvalues
    e_el_k  :: Matrix{Float64} = zeros(0, 0)
    e_el_kq :: Matrix{Float64} = zeros(0, 0)
    e_ph    :: Matrix{Float64} = zeros(0, 0)

    # Electron-phonon coupling
    ep :: Vector{Array{ComplexF64, 3}} = []

    ep_sym :: Vector{Array{ComplexF64, 3}} = []

    # # --- Data buffer fields ---
    # Σ_buffer :: Vector{OffsetArray{ComplexF64, 2, Array{ComplexF64, 2}}} = [OffsetArray(zeros(ComplexF64, 0, 0))]

    # # DEBUG field
    # method :: Symbol = :Original
end

ElectronPhonon.allow_eph_outer_k(:: MatrixElementCalculatorX4X) = true
ElectronPhonon.allow_eph_outer_q(:: MatrixElementCalculatorX4X) = true


function ElectronPhonon.setup_calculator!(calc::MatrixElementCalculatorX4X, kpts, qpts, el_states;  
                                          nw, nmodes, nchunks_threads, kwargs...)
    mpi_isroot() && println("Setting up MatrixElementCalculator")

    if nchunks_threads > 1
        error("nchunks_threads > 1 for MatrixElementCalculator is not supported yet.")
    end

    if kpts.n > 1
        error("kpts.n must be 1 for MatrixElementCalculator.")
    end

    resize!(calc.ep, qpts.n)
    for i in eachindex(calc.ep)
        calc.ep[i] = zeros(ComplexF64, nw, nw, nmodes)
    end

    calc.e_el_k = zeros(nw, qpts.n)
    calc.e_el_kq = zeros(nw, qpts.n)
    calc.e_ph = zeros(nmodes, qpts.n)

    # resize!(calc.Σ_buffer, nchunks_threads)
    # for i in eachindex(calc.Σ_buffer)
    #     calc.Σ_buffer[i] = OffsetArray(zeros(ComplexF64, nband, length(calc.occ)), rng_band, :)
    # end

    calc
end

function ElectronPhonon.setup_calculator_inner!(calc::MatrixElementCalculatorX4X, ik; kwargs...)
    # for buf in calc.Σ_buffer
    #     buf .= 0
    # end
    calc
end

function ElectronPhonon.postprocess_calculator_inner!(calc::MatrixElementCalculatorX4X; ik, kwargs...)
    # for buf in calc.Σ_buffer
    #     calc.Σ[ik] .+= buf
    # end
    calc
end

function ElectronPhonon.postprocess_calculator!(calc::MatrixElementCalculatorX4X; kwargs...)
    # Average over degenerate states

    nw, _, nmodes = size(calc.ep[1])

    calc.ep_sym = [zeros(ComplexF64, nw, nw, nmodes) for _ in 1:length(calc.ep)]
    
    ep_avg = zeros(nw, nw)
    for iq in 1:length(calc.ep)
        # average over bands at k
        @views for i in 1:nmodes
            ep_avg .= 0
            ndegen = 0
            for j in 1:nmodes
                if abs(calc.e_ph[i, iq] - calc.e_ph[j, iq]) <= 0.01 * unit_to_aru(:meV)
                    ep_avg .+= abs2.(calc.ep[iq][:, :, j])
                    ndegen += 1
                end
            end
            calc.ep_sym[iq][:, :, i] .= sqrt.(ep_avg ./ ndegen)
        end
    end

    calc
end

function ElectronPhonon.run_calculator!(calc::MatrixElementCalculatorX4X, epdata, ik, iq, ikq;
                                        id_chunk, kwargs...)
    (; el_k, el_kq, ph) = epdata

    for n in el_k.rng
        calc.e_el_k[n, iq]  = el_k.e[n]
    end
    for m in el_kq.rng
        calc.e_el_kq[m, iq] = el_kq.e[m]
    end
    calc.e_ph[:, iq] = ph.e

    @views for n in el_k.rng, m in el_kq.rng
        calc.ep[iq][m, n, :] = epdata.ep[m, n, :]
    end

    calc
end



function parse_qe_eph(filename, nband, nmodes)
    xqs_tpiba = Vec3[]
    e_el_k = Vector{Float64}[]
    e_el_kq = Vector{Float64}[]
    e_ph = Vector{Float64}[]
    g_abs = Array{Float64, 3}[]

    f = open(filename, "r")
    while !eof(f)
        line = readline(f)
        if occursin("Calculation of q =", line)
            xq = Vec3([parse(Float64, x) for x in split(line)[end-2:end]])
            push!(xqs_tpiba, xq)
        end
        if occursin("Electron-phonon vertex |g| (meV)", line)
            for _ in 1:5
                readline(f)
            end

            push!(e_el_k, zeros(nband))
            push!(e_el_kq, zeros(nband))
            push!(e_ph, zeros(nmodes))
            push!(g_abs, zeros(nband, nband, nmodes))

            for ib in 1:nband
                for jb in 1:nband
                    for imode in 1:nmodes
                        data = split(readline(f))
                        e_el_k[end][ib] = parse(Float64, data[4])
                        e_el_kq[end][jb] = parse(Float64, data[5])
                        e_ph[end][imode] = parse(Float64, data[6])
                        g_abs[end][jb, ib, imode] = parse(Float64, data[7])
                    end
                end
            end
        end
    end
    close(f)

    println("nq = $(length(e_el_k))")

    # Convert to Rydberg units
    e_el_k .*= unit_to_aru(:eV)
    e_el_kq .*= unit_to_aru(:eV)
    e_ph .*= unit_to_aru(:meV)
    g_abs .*= unit_to_aru(:meV)

    (; e_el_k = stack(e_el_k),
       e_el_kq = stack(e_el_kq),
       e_ph = stack(e_ph),
       g_abs = stack(g_abs),
       xqs_tpiba,
    )
end
