# Run one-shot Dyson calculation

"""
- `truncate_spectral`: Truncate DOS for chemical potential calculation in doped semiconductors.
Needed for FM calculation of intrinsic doped semiconductors due to long Lorentzian tails.
"""
function run_Dyson(nk, wfpt, ind_doping;
        η_tag = "smearing_5meV",
        η_list = fill(5.0, length(setup_occupation_params(1))) .* unit_to_aru(:meV),
        maxiter = 20,
        nk_tuple = (nk, nk, nk),
        truncate_spectral = true,
        update_μ = true,
        μlist_input = nothing
    )
    postfix = "doping$ind_doping.wfpt$wfpt.nk$nk"

    n_doping = n_doping_list[ind_doping]
    occ = setup_occupation_params(n_doping)

    kpts, _, _, nstates_base = filter_kpoints(nk_tuple, model.nw, model.el_ham, window_active; model.symmetry)

    # Input chemical potential
    if μlist_input !== nothing
        length(μlist_input) == length(occ) || error("Length of μlist_input should be equal to length of occ")
        occ.μlist .= μlist_input
    end

    # Compute Fermi level
    if !ElectronPhonon.chemical_potential_is_computed(occ)
        el_states = compute_electron_states(model, kpts, ["eigenvalue", "velocity_diagonal"], window_active; fourier_mode = "gridopt")
        el, _ = ElectronPhonon.electron_states_to_BTStates(el_states, kpts, nstates_base)
        ElectronPhonon.bte_compute_μ!(occ, el; do_print=true)
    end

    kpts, _, _, nstates_base = filter_kpoints(nk_tuple, model.nw, model.el_ham, window_active; model.symmetry)
    kpts = GridKpoints(kpts)
    kqpts, ik_to_ikirr_isym = ElectronPhonon.unfold_kpoints(kpts, model.symmetry)
    ikq_to_ik = getindex.(ik_to_ikirr_isym, 1)
    @info "kpts.n = $(kpts.n)"
    @info "kqpts.n = $(kqpts.n)"

    el_k_save = compute_electron_states(model, kpts, ["eigenvalue", "velocity_diagonal"]; fourier_mode = "gridopt")
    eks = hcat(getproperty.(el_k_save, :e)...)
    vks = hcat(getproperty.(el_k_save, :vdiag)...)

    if wfpt
        # Load rest space contribution computed using WFPT
        filename_wfpt = "$folder_output/data_wfpt.nk$nk.jld2"
        isfile(filename_wfpt) || error("WFPT data file not found : $filename_wfpt")
        Σs_rest = load(filename_wfpt, "Σs_rest")
        @assert size(Σs_rest) == (kpts.n, model.nw, length(occ))
    else
        Σs_rest = zeros(ComplexF64, kpts.n, model.nw, length(occ))
    end

    Σs = zeros(ComplexF64, length(ωs), model.nw, kpts.n, length(occ))

    if all(η_list .== η_list[1])
        # All same η: run together.

        println("==== nk = $nk, wfpt = $wfpt, Dyson ====")

        η = η_list[1]
        S = AbInitioElectronPhononSolverX2X(model, ωs, occ, kpts, kqpts, ikq_to_ik, Σs_rest, window_active; ωs_dense, η, nstates_base);

        if n_doping == 0 || (! update_μ)
            # Reset self-energy to 0 to do Dyson
            S.Σs .= 0
            S.Σs_dense .= 0
            compute_self_energy!(S);

        else

            μlist_prev = copy(S.occ.μlist)
            for iter in 1:maxiter
                flush(stdout)
                flush(stderr)

                # Reset self-energy to 0 to do Dyson
                S.Σs .= 0
                S.Σs_dense .= 0
                compute_self_energy!(S);

                # Update chemical potential
                println("Occupation before chemical potential update : $([compute_occupation!(S, iT; truncate_spectral) - occ.nelec for iT in 1:length(S.occ)])")
                update_chemical_potential!(S; truncate_spectral)
                println("Updated chemical potential (meV) : $(_e_to_eV.(S.occ.μlist) .* 1e3)")

                if all(abs.(S.occ.μlist .- μlist_prev) .< 0.01 * unit_to_aru(:meV))
                    break
                end

                @. μlist_prev = S.occ.μlist
            end

        end

        Σs .= S.Σs

        jldsave("$folder_output/data_Dyson.$η_tag.$postfix.jld2";
            S.kpts, S.kqpts, S.εks, S.εks_dense, S.window_active, S.Σs, S.ωs, S.ωs_dense, S.occ,
            η_list, S.Σs_rest, S.ikq_to_ik, eks, vks)

    else

        # Different η for different T: run each T separately

        for iT in 1:length(occ)
            println("==== nk = $nk, wfpt = $wfpt, Dyson, T = $(occ.Tlist[iT] / unit_to_aru(:K)) K ====")
            flush(stdout)
            flush(stderr)

            occ_iT = ElectronOccupationParams(;
                Tlist = [occ.Tlist[iT]],
                nlist = [occ.nlist[iT]],
                μlist = [occ.μlist[iT]],
                occ.volume,
                occ.nelec,
                occ.spin_degeneracy,
            )

            Σs_rest_iT = reshape(Σs_rest[:, :, iT], kpts.n, model.nw, 1)

            S = AbInitioElectronPhononSolverX2X(model, ωs, occ_iT, kpts, kqpts, ikq_to_ik, Σs_rest_iT, window_active; ωs_dense, η = η_list[iT], nstates_base);

            if n_doping == 0 || (! update_μ)
                # Reset self-energy to 0 to do Dyson
                S.Σs .= 0
                S.Σs_dense .= 0
                compute_self_energy!(S);

            else
                # Update chemical potential

                μlist_prev = copy(S.occ.μlist)

                for iter in 1:maxiter
                    # Reset self-energy to 0 to do Dyson
                    S.Σs .= 0
                    S.Σs_dense .= 0
                    compute_self_energy!(S);

                    # Update chemical potential
                    println("Occupation before chemical potential update : $(compute_occupation!(S, 1; truncate_spectral) - occ.nelec)")
                    update_chemical_potential!(S; truncate_spectral)
                    println("Updated chemical potential (meV) : $(_e_to_eV.(S.occ.μlist) .* 1e3)")

                    if all(abs.(S.occ.μlist .- μlist_prev) .< 0.01 * unit_to_aru(:meV))
                        break
                    end

                    @. μlist_prev = S.occ.μlist
                end
            end

            Σs[:, :, :, iT] .= S.Σs[:, :, :, 1]
            occ.μlist[iT] = S.occ.μlist[1]

            jldsave("$folder_output/data_Dyson.$η_tag.$postfix.jld2";
                S.kpts, S.kqpts, S.εks, S.εks_dense, S.window_active, Σs, S.ωs, S.ωs_dense, occ,
                η_list, Σs_rest, S.ikq_to_ik, eks, vks)
        end
    end

    S = AbInitioElectronPhononSolverX2X(model, ωs, occ, kpts, kqpts, ikq_to_ik, Σs_rest, window_active; ωs_dense, η=η_list[1], nstates_base);
    S.Σs .= Σs

    return S
end;
