# Compute conductivity and mobility using bubble approximation
include("/home/ucl/modl/jmlihm/eph_self_consistent/dev/scKubo.jl")
include("/home/ucl/modl/jmlihm/eph_self_consistent/dev/scKubo_bubble.jl")

"""
- `truncate_spectral`: if `true`, the spectral function is truncated to the window. Used for
    one-shot FM calculations on lightly doped semiconductors.
- `phonon_assisted` : if `true`, compute the phonon-assisted conductivity as well.
"""
function run_bubble(nk, wfpt, ind_doping, prefix;
        truncate_spectral = false,
        phonon_assisted = false,
    )
    postfix = "doping$ind_doping.wfpt$wfpt.nk$nk"
    filename = "$folder_output/$prefix.$postfix.jld2"
    n_doping = n_doping_list[ind_doping]

    if !isfile(filename)
        println("File not found: $filename")
        return nothing
    end

    filename_BTStates = "$folder_output/data_scKubo_BTStates_nk$nk.jld2"
    if !isfile(filename_BTStates)
        println("BTStates file not found: $filename_BTStates")
        return nothing
    end
    el_i = load("$folder_output/data_scKubo_BTStates_nk$nk.jld2", "el_i")

    kpts, occ, ωs_dense = load(filename, "kpts", "occ", "ωs_dense")
    dω = ωs_dense[2] - ωs_dense[1]
    basis = LinearSplineBasis(vcat(ωs_dense[1] - dω, ωs_dense, ωs_dense[end] + dω))

    # Load self-energy
    Σs, ωs = load(filename, "Σs", "ωs")
    Σs = scKubo_trim_Σs_to_BTStates(Σs, el_i, kpts);

    fac_mobility_SI = ElectronPhonon.e2 / (unit_to_aru(:A) / unit_to_aru(:V) * unit_to_aru(:cm)^2) ./ ElectronPhonon.units.e_SI

    Ω = 0.0
    (; σ, jdos) = compute_conductivity_bubble(Ω, occ, basis, model, el_i, ωs, Σs; truncate_spectral)
    mobility = zeros(3, 3, length(occ))
    for iT in 1:length(occ)
        mobility[:, :, iT] .= σ[:, :, iT] ./ (abs(occ.nlist[iT]) / model.volume) .* fac_mobility_SI
    end

    if phonon_assisted
        filename_mel = "$folder_output/data_scKubo_matrix_elements_nk$nk.h5"
        if !isfile(filename_mel)
            println("phonon_assisted is true, but matrix elements file not found: $filename_mel")
            return nothing
        end

        el_i, el_f, ph = load("$folder_output/data_scKubo_BTStates_nk$nk.jld2", "el_i", "el_f", "ph")
        ik_and_isym_map = find_unfolding_indices(el_i, el_f, model.symmetry)
        σ_ph_ass = compute_phonon_assisted_bubble(Ω, occ, basis, model, el_i, el_f, ph, filename_mel, ωs, Σs, ik_and_isym_map; truncate_spectral);

        mobility_ph_ass = zeros(3, 3, length(occ))
        for iT in 1:length(occ)
            mobility_ph_ass[:, :, iT] .= σ[:, :, iT] ./ (abs(occ.nlist[iT]) / model.volume) .* fac_mobility_SI
        end
    else
        σ_ph_ass = nothing
        mobility_ph_ass = nothing
    end

    n_calc = compute_occupation(occ, basis, el_i, ωs, Σs; truncate_spectral)
    for iT in 1:length(occ)
        if abs((n_calc[iT] - occ.nlist[iT]) / occ.nlist[iT]) > 1e-2
            println("WARNING: Occupation not converged for iT = $iT")
            println("T = $(occ.Tlist[iT] / unit_to_aru(:K)) K")
            println("n_target = $(occ.nlist[iT])")
            println("n_calc   = $(n_calc[iT])")
            println("Setting mobility to NaN")
            mobility[:, :, iT] .= NaN
            if mobility_ph_ass !== nothing
                mobility_ph_ass[:, :, iT] .= NaN
            end
        end
    end

    return (; mobility, σ, jdos, mobility_ph_ass, σ_ph_ass, occ)
end;

function run_bubble_AC(nk, wfpt, ind_doping, prefix, Ωs;
        truncate_spectral = false,
        phonon_assisted = false,
    )
    postfix = "doping$ind_doping.wfpt$wfpt.nk$nk"
    filename = "$folder_output/$prefix.$postfix.jld2"
    n_doping = n_doping_list[ind_doping]

    if !isfile(filename)
        println("File not found: $filename")
        return nothing
    end

    filename_BTStates = "$folder_output/data_scKubo_BTStates_nk$nk.jld2"
    if !isfile(filename_BTStates)
        println("BTStates file not found: $filename_BTStates")
        return nothing
    end
    el_i = load(filename_BTStates, "el_i")

    kpts, occ, ωs_dense = load(filename, "kpts",  "occ", "ωs_dense")
    basis = LinearSplineBasis(ωs_dense)

    fac_mobility_SI = ElectronPhonon.e2 / (unit_to_aru(:A) / unit_to_aru(:V) * unit_to_aru(:cm)^2) ./ ElectronPhonon.units.e_SI

    # Load self-energy
    Σs, ωs = load(filename, "Σs", "ωs")
    Σs = scKubo_trim_Σs_to_BTStates(Σs, el_i, kpts);

    σ = zeros(3, 3, length(occ), length(Ωs))
    jdos = zeros(length(occ), length(Ωs))
    mobility = zeros(3, 3, length(occ), length(Ωs))

    for (iΩ, Ω) in enumerate(Ωs)
        if mod(iΩ, 10) == 0
            println("Ω = $Ω (iΩ = $iΩ / $(length(Ωs)))")
        end
        res = compute_conductivity_bubble(Ω, occ, basis, model, el_i, ωs, Σs; truncate_spectral)
        σ[:, :, :, iΩ] .= res.σ
        jdos[:, iΩ] .= res.jdos
    end

    mobility = σ ./ (abs(n_doping) / model.volume) .* fac_mobility_SI

    if phonon_assisted
        filename_mel = "$folder_output/data_scKubo_matrix_elements_nk$nk.h5"
        if !isfile(filename_mel)
            println("phonon_assisted is true, but matrix elements file not found: $filename_mel")
            return nothing
        end

        el_i, el_f, ph = load("$folder_output/data_scKubo_BTStates_nk$nk.jld2", "el_i", "el_f", "ph")
        ik_and_isym_map = find_unfolding_indices(el_i, el_f, model.symmetry)

        σ_ph_ass = zeros(3, 3, length(occ), length(Ωs))
        mobility_ph_ass = zeros(3, 3, length(occ), length(Ωs))
        for (iΩ, Ω) in enumerate(Ωs)
            println("Phonon-assisted bubble, Ω = $Ω (iΩ = $iΩ / $(length(Ωs)))")
            σ_ph_ass[:, :, :, iΩ] .= compute_phonon_assisted_bubble(Ω, occ, basis, model, el_i, el_f, ph, filename_mel, ωs, Σs, ik_and_isym_map; truncate_spectral);
        end
    
        mobility_ph_ass = σ_ph_ass ./ (abs(n_doping) / model.volume) .* fac_mobility_SI
    else
        σ_ph_ass = nothing
        mobility_ph_ass = nothing
    end

    n_calc = compute_occupation(occ, basis, el_i, ωs, Σs; truncate_spectral)
    for iT in 1:length(occ)
        if abs((n_calc[iT] - occ.nlist[iT]) / occ.nlist[iT]) > 1e-2
            println("WARNING: Occupation not converged for iT = $iT")
            println("T = $(occ.Tlist[iT] / unit_to_aru(:K)) K")
            println("n_target = $(occ.nlist[iT])")
            println("n_calc   = $(n_calc[iT])")
            println("Setting mobility to NaN")
            mobility[:, :, iT, :] .= NaN
            if mobility_ph_ass !== nothing
                mobility_ph_ass[:, :, iT, :] .= NaN
            end
        end
    end

    return (; mobility, σ, jdos, mobility_ph_ass, σ_ph_ass)
end;

function _kramers_kronig_for_optical_conductivity(Ωs, σs)
    @assert length(σs) == length(Ωs)
    N = length(Ωs)
    basis_Ω = LinearSplineBasis(vcat(reverse(.-Ωs), Ωs[2:end]))
    σs_im = EPSpectral.get_Kramers_Kronig_map(basis_Ω) * vcat(reverse(σs), σs[2:end])[2:end-1] ./ π
    return vcat(0, σs_im, 0)[N:end]
end

# function run_bubble_AC(nk, wfpt, ind_doping, prefix, Ωs;
#         truncate_spectral = false
#     )
#     postfix = "doping$ind_doping.wfpt$wfpt.nk$nk"
#     filename = "$folder_output/$prefix.$postfix.jld2"
#     n_doping = n_doping_list[ind_doping]

#     if !isfile(filename)
#         println("File not found: $filename")
#         return nothing
#     end

#     kpts, window_active, occ, eks, vks, ωs_dense = load(filename,
#         "kpts", "window_active", "occ", "eks", "vks", "ωs_dense")

#     basis = LinearSplineBasis(ωs_dense)

#     fac_mobility_SI = ElectronPhonon.e2 / (unit_to_aru(:A) / unit_to_aru(:V) * unit_to_aru(:cm)^2) ./ ElectronPhonon.units.e_SI

#     # Use self-energy
#     Σs, ωs = load(filename, "Σs", "ωs")
#     Σs_itp = compute_Σs_itp(kpts, ωs, Σs, model.symmetry)  # Bottleneck

#     σ = zeros(3, 3, length(occ), length(Ωs))
#     mobility = zeros(3, 3, length(occ), length(Ωs))

#     for (iΩ, Ω) in enumerate(Ωs)
#         if mod(iΩ, 10) == 0
#             println("Ω = $Ω (iΩ = $iΩ / $(length(Ωs)))")
#         end
#         res = compute_conductivity_bubble(Ω, occ, Σs_itp, kpts, eks, vks, basis, window_active, model; truncate_spectral)
#         σ[:, :, :, iΩ] .= res.σ
#     end

#     mobility = σ ./ (abs(n_doping) / model.volume) .* fac_mobility_SI

#     return (; mobility, σ)
# end;



# function run_bubble(nk, wfpt, ind_doping, prefix;
#         truncate_spectral = false
#     )
#     postfix = "doping$ind_doping.wfpt$wfpt.nk$nk"
#     filename = "$folder_output/$prefix.$postfix.jld2"
#     n_doping = n_doping_list[ind_doping]

#     if !isfile(filename)
#         println("File not found: $filename")
#         return nothing
#     end

#     kpts, window_active, occ, eks, vks, ωs_dense = load(filename,
#         "kpts", "window_active", "occ", "eks", "vks", "ωs_dense")

#     basis = LinearSplineBasis(ωs_dense)

#     fac_mobility_SI = ElectronPhonon.e2 / (unit_to_aru(:A) / unit_to_aru(:V) * unit_to_aru(:cm)^2) ./ ElectronPhonon.units.e_SI

#     # Use self-energy
#     Σs, ωs = load(filename, "Σs", "ωs")
#     Σs_itp = compute_Σs_itp(kpts, ωs, Σs, model.symmetry)  # Bottleneck

#     Ω = 0.0
#     σ, = compute_conductivity_bubble(Ω, occ, Σs_itp, kpts, eks, vks, basis, window_active, model; truncate_spectral)
#     mobility = σ ./ (abs(n_doping) / model.volume) .* fac_mobility_SI

#     return (; mobility, σ)
# end;
