function run_cumulant_spectral(ωs, Σs, Σs_rest, el_i, occ, e_Fermi, ωs_dense)
    @assert size(Σs) == (length(ωs), el_i.n, length(occ))
    @assert size(Σs_rest) == (el_i.n, length(occ))

    weights_ωs_dense = vcat(ωs_dense[2] - ωs_dense[1], (ωs_dense[3:end] .- ωs_dense[1:end-2]) ./ 2, ωs_dense[end] - ωs_dense[end-1])

    ts = range(0, 2e4, step=5.)
    ωs_cum = EPSpectral.fftfreq_t2w(ts)

    @show diff(ωs_cum)[1] / unit_to_aru(:meV)
    @show ωs_cum[end] / unit_to_aru(:meV)

    As_cum = zeros(length(ωs_dense), el_i.n, length(occ));

    iT_and_ind = collect(Iterators.product(1:length(occ), 1:el_i.n))

    # We need to extrapolate ωs above ωs[end] to compute the cumulant up to ωs[end]
    # Real part : 1/ω, Imaginary part : 1/ω^2
    ωs_large = vcat(ωs, range(ωs[end], ωs[end] + 1.0 * unit_to_aru(:eV), step = ωs[end] - ωs[end-1])[2:end])
    basis = LinearSplineBasis(ωs_large .- e_Fermi)

    @views tmap(iT_and_ind) do (iT, ind)
        if Threads.threadid() == 1
            @info "iT = $iT / $(length(occ)), ind = $ind / $(el_i.n)"
        end

        # Add rest self-energy to ek not Σs (because it is purely static)
        Σ_static = Σs_rest[ind, iT]
        ek = el_i[ind].e + real(Σ_static)
        Σs_k = Σs[:, ind, iT] .- Σ_static

        # Extrapolate self-energy above ωs[end]
        Σs_k_itp = linear_interpolation(ωs, Σs_k)
        Σs_k_large = zeros(ComplexF64, length(ωs_large))
        for (iω, ω) in enumerate(ωs_large)
            if ω < ωs[1]
                Σs_k_large[iω] = 0.0im
            elseif ω > ωs[end]
                x0 = ωs[end] - e_Fermi
                x1 = ω - e_Fermi
                Σs_k_large[iω] = real(Σs_k[end]) * (x0 / x1) + im * imag(Σs_k[end]) * (x0 / x1)^2
            else
                Σs_k_large[iω] = Σs_k_itp(ω)
            end
        end

        βs = .-imag.(linear_interpolation(ωs_large, Σs_k_large; extrapolation_bc=0).(basis.xs[2:end-1] .+ ek)) ./ π
        As_cum_ik = EPSpectral.run_cumulant(basis, βs, ek - e_Fermi, ts, ts)

        # Interpolate back to the original grid
        As_cum_itp = linear_interpolation(ωs_cum .+ e_Fermi, As_cum_ik; extrapolation_bc = 0)
        As_cum[:, ind, iT] .= As_cum_itp.(ωs_dense)

        # Normalize
        As_sum = sum(As_cum[:, ind, iT] .* weights_ωs_dense)
        As_cum[:, ind, iT] ./= As_sum
    end

    As_cum
end

function cumulant_compute_chemical_potential(As, occ, el_i, ωs_dense)
    dω = diff(ωs_dense)[1]
    @assert all(diff(ωs_dense) .≈ dω)

    function _compute_n(μ, dos, occ, iT, el_i)
        T = occ.Tlist[iT]
        if occ.type === :Semiconductor && occ.nlist[iT] < 0
            # Count holes instead of electrons
            count_holes = true
            fermi = @. occ_fermion(μ - ωs_dense, T)
        else
            count_holes = false
            fermi = @. occ_fermion(ωs_dense - μ, T)
        end

        n = sum(dos .* fermi) * dω

        if count_holes
            n = sum(el_i.k_weight) - n
        end

        return (n + el_i.nstates_base) * occ.spin_degeneracy - occ.nelec
    end

    μlist = zeros(length(occ))
    for iT in 1:length(occ)
        dos = sum(As[:, i, iT] .* el_i[i].k_weight for i in 1:el_i.n)
        μlist[iT] = EPSpectral.find_zero(μ -> _compute_n(μ, dos, occ, iT, el_i) - occ.nlist[iT], extrema(ωs_dense) .+ (-0.5, +0.5))
    end

    μlist
end

function cumulant_compute_conductivity(Ω, As, occ, el_i, ωs_dense)
    dω = ωs_dense[2] - ωs_dense[1]
    @assert all(diff(ωs_dense) .≈ dω)

    nlist = zeros(length(occ))
    σ = zeros(3, 3, length(occ))

    for (iT, (; T, μ)) in enumerate(occ)
        if occ.type === :Semiconductor && occ.nlist[iT] < 0
            # Count holes instead of electrons
            count_holes = true
        else
            count_holes = false
        end

        if count_holes
            fermi = @. occ_fermion(μ - ωs_dense, T)
        else
            fermi = @. occ_fermion(ωs_dense - μ, T)
        end

        dfermi = @. -docc_fermion(ωs_dense - μ, T)

        n = tmapreduce(+, 1:el_i.n) do i
            n_k = sum(As[:, i, iT] .* fermi) * dω
            n_k * el_i[i].k_weight
        end
        nlist[iT] = (sum(el_i.k_weight) - n + el_i.nstates_base) * occ.spin_degeneracy - occ.nelec

        σ_iT = tmapreduce(+, 1:el_i.n) do i
            vk = el_i[i].vdiag            
            if Ω == 0
                jdos = sum(dfermi .* As[:, i, iT].^2) .* dω
            else
                As_itp = LinearInterpolation(ωs_dense, As[:, i, iT]; extrapolation_bc = 0)
                f1 = @. occ_fermion(ωs_dense - Ω / 2 - μ, T)
                f2 = @. occ_fermion(ωs_dense + Ω / 2 - μ, T)
                A1 = @. As_itp(ωs_dense - Ω / 2)
                A2 = @. As_itp(ωs_dense + Ω / 2)
                jdos = sum((f1 - f2) / Ω .* A1 .* A2) .* dω
            end
            jdos * real.(vk * vk') * π * el_i[i].k_weight
        end
        
        # Symmetrize conductivity tensor
        σ_iT = symmetrize(σ_iT, model.symmetry)
        σ_iT *= occ.spin_degeneracy / model.volume

        σ[:, :, iT] .= σ_iT
    end

    (; nlist, σ)
end
