# Run self-consistent Fan-Migdal (scFM) calculation

function run_scFM(nk, wfpt, ind_doping;
        η_init = 5.0 * unit_to_aru(:meV),
        save_intermediate = false,
        maxiter = 20,
        nk_tuple = (nk, nk, nk),
    )

    postfix = "doping$ind_doping.wfpt$wfpt.nk$nk"

    n_doping = n_doping_list[ind_doping]
    occ = setup_occupation_params(n_doping)

    # Compute Fermi level
    kpts, _, _, nstates_base = filter_kpoints(nk_tuple, model.nw, model.el_ham, window_active; model.symmetry)

    if !ElectronPhonon.chemical_potential_is_computed(occ)
        el_states = compute_electron_states(model, kpts, ["eigenvalue", "velocity_diagonal"], window_active; fourier_mode = "gridopt")
        el, _ = ElectronPhonon.electron_states_to_BTStates(el_states, kpts, nstates_base)
        ElectronPhonon.bte_compute_μ!(occ, el; do_print=true)
    end

    kpts, = filter_kpoints(nk_tuple, model.nw, model.el_ham, window_active; model.symmetry)
    kpts = GridKpoints(kpts)
    kqpts, ik_to_ikirr_isym = ElectronPhonon.unfold_kpoints(kpts, model.symmetry)
    ikq_to_ik = getindex.(ik_to_ikirr_isym, 1)
    @info "kpts.n = $(kpts.n)"
    @info "kqpts.n = $(kqpts.n)"

    el_k_save = compute_electron_states(model, kpts, ["eigenvalue", "velocity_diagonal"]; fourier_mode = "gridopt")
    eks = hcat(getproperty.(el_k_save, :e)...)
    vks = hcat(getproperty.(el_k_save, :vdiag)...)

    if wfpt
        # Load rest space contribution computed using WFPT
        filename_wfpt = "$folder_output/data_wfpt.nk$nk.jld2"
        isfile(filename_wfpt) || error("WFPT data file not found : $filename_wfpt")
        Σs_rest = load(filename_wfpt, "Σs_rest")
        @assert size(Σs_rest) == (kpts.n, model.nw, length(occ))
    else
        Σs_rest = zeros(ComplexF64, kpts.n, model.nw, length(occ))
    end

    S = AbInitioElectronPhononSolverX2X(model, ωs, occ, kpts, kqpts, ikq_to_ik, Σs_rest, window_active; ωs_dense, η = η_init, nstates_base);

    # Run 1 : Initialize self-energy from a one-shot calculation
    println("nk = $nk, wfpt = $wfpt, Dyson")
    flush(stdout)
    flush(stderr)
    # Reset self-energy to 0 to do Dyson
    S.Σs .= 0
    S.Σs_dense .= 0
    compute_self_energy!(S);

    # Run 2 : Update self-energy and chemical potential

    Σs_prev = copy(S.Σs)
    S.η = 0.0 * unit_to_aru(:meV)

    for iter in 1:maxiter
        println("nk = $nk, wfpt = $wfpt, scFM iteration $iter")
        flush(stdout)
        flush(stderr)

        compute_self_energy!(S);

        if save_intermediate
            jldsave("$folder_output/data_scFM.$postfix.iter$iter.jld2";
                S.kpts, S.kqpts, S.εks, S.εks_dense, S.window_active, S.Σs, S.ωs, S.ωs_dense, S.occ, S.η, S.Σs_rest, S.ikq_to_ik, eks, vks)
        end

        if n_doping != 0
            # Update chemical potential
            println("Occupation before chemical potential update : $([compute_occupation!(S, iT) - occ.nelec for iT in 1:length(S.occ)])")
            update_chemical_potential!(S)
            println("Updated chemical potential (meV) : $(_e_to_eV.(S.occ.μlist) .* 1e3)")
        end

        Σ_err_meV = maximum.(eachslice(abs.(S.Σs .- Σs_prev), dims = 4)) ./ unit_to_aru(:meV)
        println("Σ_err (meV) = $Σ_err_meV")
        if maximum(Σ_err_meV) < 0.05
            break
        end
        Σs_prev = copy(S.Σs)
    end

    jldsave("$folder_output/data_scFM.$postfix.jld2";
            S.kpts, S.kqpts, S.εks, S.εks_dense, S.window_active, S.Σs, S.ωs, S.ωs_dense,
            S.occ, S.η, S.Σs_rest, S.ikq_to_ik, eks, vks
    )
end;
