using ChunkSplitters
using SparseArrays
using StaticArrays
using Base.Threads
using Dates
using HDF5
import Interpolations: tweight

include("/home/ucl/modl/jmlihm/eph_self_consistent/dev/scKubo_bubble.jl")

# https://github.com/JuliaMath/Interpolations.jl/issues/450
Interpolations.tweight(::AbstractArray{SVector{3, ComplexF64}}) = Float64


function amputate_current_correlator(J_c, Gm_c, Gp_c)
    v_c = zero.(J_c)
    for i in eachindex(J_c)
        for c0 in 1:2, iω in axes(J_c[i], 1)
            Gm = SMatrix{2, 2}(Gm_c[i][iω, :, :])
            Gp = SMatrix{2, 2}(Gp_c[i][iω, :, :])
            J = SMatrix{2, 2}(J_c[i][iω, :, :, c0])
            v_c[i][iω, :, :, c0] = inv(Gm) * J * inv(Gp)
        end
    end
    v_c
end

function compute_conductivity(Ω, el, basis, v0_c, J_c, T)
    # Mobility using the current correlator

    # Basis factor for ∫ dν bᵢ(ν) / 2πi
    basis_sum = @. (basis.xs[3:end] .- basis.xs[1:end-2]) / 2 / (2π * im)

    χ_bubble = zeros(ComplexF64, 3, 3, 2, 2, el.n)

    for i in 1:el.n
        weight = el[i].k_weight

        for iω in eachindex(grid_points(basis))
            # Use bare current vertex v0_c, renormalized correlator J_c
            for c2 in 1:2, c1 in 1:2
                vL = v0_c[i][iω]
                JR = J_c[i][iω, c1, c1, c2]
                χ_bubble[:, :, c1, c2, i] += (conj.(vL) * transpose(JR)) * basis_sum[iω] * weight
            end
        end
    end

    if Ω == 0.
        freq_fac = im / 4T
    else
        freq_fac = -1 / (2im * Ω * (1 + 2 * occ_boson(Ω, T)))
    end
    σk = @. real((χ_bubble[:, :, 1, 2, :] + χ_bubble[:, :, 2, 1, :]) * freq_fac)

    σ = dropdims(sum(σk, dims=3), dims=3)

    (; σ, σk)
end


"""
Build a map that unfolds a velocity (polar, time-reversal odd) on `el_i` to `el_f` by `symmetry`.
``S(ik_and_isym_map[i][2]) * el_i[ik_and_isym_map[i][1]] -> el_f[i]``
"""
function find_unfolding_indices(el_i :: ElectronPhonon.BTStates, el_f :: ElectronPhonon.BTStates, symmetry)
    ik_and_isym_map = fill((0, 0), el_f.n)

    for i in 1:el_f.n
        xk_f = el_f[i].xks
        ib = el_f[i].iband

        found = false

        for (isym, (; S, is_tr)) in enumerate(symmetry)
            for j in 1:el_i.n
                if el_i[j].iband == ib
                    k_i = el_i[j].xks
                    Sk_i = is_tr ? -S * k_i : S * k_i
                    dk = Sk_i - xk_f
                    if all(abs.(dk .- round.(dk)) .< 1e-6)
                        ik_and_isym_map[i] = (j, isym)
                        found = true
                        break
                    end
                end
            end
        end

        found || error("Cannot find the corresponding state in el_i for i = $i, e_f = $(el_f[i])")
    end

    ik_and_isym_map
end

function iterate_current_Dyson(J_c, filename_btedata, ik_and_isym_map, el_i, el_f, ph, basis, Gm_c, Gp_c, model, T, kpts, kqpts)
    (; nmodes) = model
    fid = h5open(filename_btedata, "r")
    group_scattering = open_group(fid, "scattering")

    scat_list = ElectronPhonon.ElPhScatteringData{Float64}[]
    @time for ik in 1:kpts.n
        key = "ik$ik"
        g = open_group(group_scattering, key)
        push!(scat_list, load_BTData(g, ElectronPhonon.ElPhScatteringData{Float64}))
    end

    nq = div(ph.n, nmodes)
    @assert ph.n == nq * nmodes

    @time tmp_DJ = tmapreduce(.+, chunks(1:nq; n = nthreads(), split = :scatter); chunking = false) do iqs

        tmp_DJ_buffer = zero.(J_c)

        ω_ph_list = zeros(nmodes)
        D_c_modes = [[zeros(ComplexF64, basis.N, basis.N) for _ in 1:2, _ in 1:2] for _ in 1:nmodes]
        D_map_buffer = EPSpectral.get_D_map_opt_buffers_v2(basis)
        D_c = zeros(ComplexF64, basis.N, basis.N, 2, 2)
        J_c_f = zeros(Vec3{ComplexF64}, basis.N, 2, 2, 2)

        for iq in iqs
            if threadid() == 1
                @info "$(now()) iq = $iq / $nq"
                flush(stdout)
                flush(stderr)
            end

            ind_ph_list = (iq - 1) * nmodes .+ (1:nmodes)
            xq = ph[ind_ph_list[1]].xks

            # Compute D_c for each mode

            for imode in 1:nmodes
                ind_ph = (iq - 1) * nmodes + imode
                @assert ph.iband[ind_ph] == imode

                ω_ph = ph.e[ind_ph]
                ω_ph_list[imode] = ω_ph
                EPSpectral.get_D_map_opt_v2!(D_c_modes[imode], basis, ω_ph, T, D_map_buffer)
            end

            for ik in 1:kpts.n
                scat = scat_list[ik]

                # number of bands for the same k-point
                xk = kpts.vectors[ik]

                ind_el_i_list = filter(i -> el_i[i].xks ≈ xk, 1:el_i.n)

                ikq = xk_to_ik(xk + xq, kqpts)
                ikq === nothing && continue
                ind_el_f_list = filter(i -> el_f[i].xks ≈ kqpts.vectors[ikq], 1:el_f.n)

                for ind_el_i in ind_el_i_list, ind_el_f in ind_el_f_list

                    D_c .= 0

                    # Accumulate D_c for same ind_el_i and ind_el_f (summing over nmodes)

                    for s in scat
                        s.sign_ph == 1 || continue
                        s.ind_el_i == ind_el_i || continue
                        s.ind_el_f == ind_el_f || continue
                        s.ind_ph ∈ ind_ph_list || continue

                        (; ind_ph) = s

                        imode = ph.iband[ind_ph]
                        ω_ph = ph.e[ind_ph]
                        e_f = el_f[ind_el_f]
                        g2 = s.mel

                        @assert ω_ph_list[imode] ≈ ω_ph
                        ω_ph < ElectronPhonon.omega_acoustic && continue

                        fac = e_f.k_weight * g2
                        @views for c2 in 1:2, c1 in 1:2
                            D_c[:, :, c1, c2] .+= D_c_modes[imode][c1, c2] .* fac
                        end
                    end

                    # Multiply D_c to J_c_f at (k+q, jbnd)

                    i, isym = ik_and_isym_map[ind_el_f]
                    symop = model.symmetry[isym]
                    if symop.is_tr
                        J_c_f .= Ref(-symop.Scart) .* J_c[i]
                    else
                        J_c_f .= Ref(+symop.Scart) .* J_c[i]
                    end

                    @views for c0 in 1:2, c2 in 1:2, c1 in 1:2
                        mul!(tmp_DJ_buffer[ind_el_i][:, c1, c2, c0], D_c[:, :, c2, c1], J_c_f[:, c1, c2, c0], 1, 1)
                    end

                end
            end  # key
        end  # iq
        tmp_DJ_buffer
    end  # iqs

    close(fid)

    J_new_c = zero.(J_c)
    for i in 1:el_i.n
        # (-1)^{δ_{c,+}} sign factor when attaching Green function to vertex
        tmp_DJ[i][:, 2, :, :] .*= -1
        tmp_DJ[i][:, :, 2, :] .*= -1

        @views for c0 in 1:2, c1 in 1:2, c2 in 1:2, cp1 in 1:2, cp2 in 1:2
            @. J_new_c[i][:, c1, c2, c0] += Gm_c[i][:, c1, cp1] * tmp_DJ[i][:, cp1, cp2, c0] * Gp_c[i][:, cp2, c2]
        end
    end

    J_new_c .*= -1

    J_new_c
end;

function iterate_current_Dyson_setup_scattering(filename_btedata, el_i, el_f, ph, kpts, nmodes)
    # For each q point, find list of ind_el_i and ind_el_f mapped by q and store the
    # corresponding matrix elements

    nq = div(ph.n, nmodes)
    nk = kpts.n

    # Read all of the scattering data to memory
    fid = h5open(filename_btedata, "r")
    group_scattering = open_group(fid, "scattering")
    scat_list = ElectronPhonon.ElPhScatteringData{Float64}[]
    @time for ik in 1:kpts.n
        @info "$(now()) Reading ik = $ik / $(nk)"
        flush(stdout)
        flush(stderr)
        key = "ik$ik"
        g = open_group(group_scattering, key)
        push!(scat_list, load_BTData(g, ElectronPhonon.ElPhScatteringData{Float64}))
    end
    close(fid)

    ind_el_i_list_allk = [[Int[] for _ in 1:nk] for _ in 1:nq]
    ind_el_f_list_allk = [[Int[] for _ in 1:nk] for _ in 1:nq]

    mels_buffer_dict_allk = [Dict{NTuple{2, Int}, Vector{Float64}}() for _ in 1:nk]

    @time @threads for ik in 1:nk

        if threadid() == 1 && mod(ik, 10) == 0
            @info "$(now()) ik = $ik / $(nk)"
            flush(stdout)
            flush(stderr)
        end

        for s in scat_list[ik]
            s.sign_ph == 1 || continue

            (; ind_el_i, ind_el_f, ind_ph, mel) = s

            if ph.e[s.ind_ph] < ElectronPhonon.omega_acoustic
                mel = 0.0
            end

            iq = div(ind_ph - 1, nmodes) + 1
            imode = mod1(ind_ph, nmodes)

            @assert imode == ph.iband[ind_ph]

            if imode == 1
                push!(ind_el_i_list_allk[iq][ik], ind_el_i)
                push!(ind_el_f_list_allk[iq][ik], ind_el_f)
                mels_buffer_dict_allk[ik][(ind_el_i, ind_el_f)] = zeros(nmodes)
            end

            mels_buffer_dict_allk[ik][(ind_el_i, ind_el_f)][imode] = mel * el_f[ind_el_f].k_weight
        end
    end

    @time res = tmap(1:nq) do iq
        ind_el_i_list_iq = reduce(vcat, ind_el_i_list_allk[iq])
        ind_el_f_list_iq = reduce(vcat, ind_el_f_list_allk[iq])

        nscat = length(ind_el_i_list_iq)
        mels_list_iq = zeros(nmodes, nscat)
        iscat = 0
        @views for ik in 1:nk
            for i in eachindex(ind_el_i_list_allk[iq][ik])
                ind_el_i = ind_el_i_list_allk[iq][ik][i]
                ind_el_f = ind_el_f_list_allk[iq][ik][i]
                iscat += 1
                mels_list_iq[:, iscat] .= mels_buffer_dict_allk[ik][(ind_el_i, ind_el_f)]
            end
        end
        @assert iscat == nscat

        (ind_el_i_list_iq, ind_el_f_list_iq, mels_list_iq)
    end

    ind_el_i_list = getindex.(res, 1)
    ind_el_f_list = getindex.(res, 2)
    mels_list = getindex.(res, 3)

    (; ind_el_i_list, ind_el_f_list, mels_list)
end

function iterate_current_Dyson_distribute_qpts(btedata_for_dyson, nchunks = nthreads())
    # To fix node imbalence, partition iqs to equally distribute the scatterings

    nq = length(btedata_for_dyson.ind_el_i_list)
    nscat_tot = sum(length.(btedata_for_dyson.ind_el_i_list))
    nscat_per_thread_ideal = div(nscat_tot, nchunks)

    iqs_list = Vector{Int}[]

    iq = 0

    for tid in 1:nchunks
        push!(iqs_list, Int[])
        nscat_tid = 0.0

        while nscat_tid < nscat_per_thread_ideal
            iq += 1
            if iq > nq
                break
            end
            nscat_tid += length(btedata_for_dyson.ind_el_i_list[iq])
            push!(iqs_list[tid], iq)
        end
    end

    iqs_list
end

function iterate_current_Dyson_new4(J_c, ik_and_isym_map, el_i, el_f, ph, basis, Gm_c, Gp_c, model, T, btedata_for_dyson)
    (; nmodes) = model

    nq = div(ph.n, nmodes)
    @assert ph.n == nq * nmodes

    # Distribute q points to threads with load balancing
    nchunks = 1 * nthreads()
    iqs_list = iterate_current_Dyson_distribute_qpts(btedata_for_dyson, nchunks)

    tmp_DJ_buffers = [[zeros(ComplexF64, basis.N, 3 * 2, 2, 2) for _ in 1:el_i.n] for _ in 1:nchunks]

    J_c_perm = permutedims.(J_c, Ref((1, 4, 2, 3)))

    # Unfold J_c to el_f
    J_c_f_list = [zeros(ComplexF64, basis.N, 3*2, 2, 2) for _ in 1:el_f.n]
    for ind_el_f in 1:el_f.n
        i, isym = ik_and_isym_map[ind_el_f]

        symop = model.symmetry[isym]
        if symop.is_tr
            J_c_f = Ref(-symop.Scart) .* J_c_perm[i]
        else
            J_c_f = Ref(+symop.Scart) .* J_c_perm[i]
        end

        # Reinterpret (basis.N, 2, 2, 2) Array of Vec3{ComplexF64}
        # to a (basis.N, 3, 2, 2, 2) Array of ComplexF64
        J_c_f_list[ind_el_f] .= reshape(
            permutedims(reshape(
                reinterpret(ComplexF64, J_c_f),
                (3, basis.N, 2, 2, 2)), (2, 1, 3, 4, 5)
            ), (basis.N, 3*2, 2, 2)
        )
    end

    @threads :static for ichunk in 1:nchunks
    # for ichunk in 1:1

        iqs = iqs_list[ichunk]

        ω_ph_list = zeros(nmodes)
        D_c_modes = zeros(ComplexF64, basis.N, basis.N, nmodes, 2, 2)
        D_map_buffer = EPSpectral.get_D_map_opt_buffers_v2(basis)

        nbatch = 10
        J_c_f_mel_batch = [zeros(ComplexF64, basis.N, nmodes, 3 * 2, nbatch) for _ in 1:2, _ in 1:2]
        J_c_i_batch = [zeros(ComplexF64, basis.N, 3 * 2, nbatch) for _ in 1:2, _ in 1:2]

        for iq in iqs
            if threadid() == 1 && mod(iq, 10) == 0
                @info "$(now()) iq = $iq / $(length(iqs)) (total $nq)"
                flush(stdout)
                flush(stderr)
            end

            # Compute D_c for each mode

            for imode in 1:nmodes
                ind_ph = (iq - 1) * nmodes + imode
                @assert ph.iband[ind_ph] == imode

                ω_ph = ph.e[ind_ph]
                ω_ph_list[imode] = ω_ph
                EPSpectral.get_D_map_opt_v2_array!(view(D_c_modes, :, :, imode, :, :), basis, ω_ph, T, D_map_buffer)
            end

            ind_el_i_list = btedata_for_dyson.ind_el_i_list[iq]
            ind_el_f_list = btedata_for_dyson.ind_el_f_list[iq]
            mels_list = btedata_for_dyson.mels_list[iq]
            nscat = length(ind_el_i_list)

            # DJ[:, ind_el_i] += D_c_mode[:, :, imode] * mels[imode] * J_c_f[:, ind_el_f]

            @views for iscat in 1:nscat
                ibatch = mod1(iscat, nbatch)
                ind_el_f = ind_el_f_list[iscat]
                for c2 in 1:2, c1 in 1:2
                    @views for imode in 1:nmodes
                        J_c_f_mel_batch[c1, c2][:, imode, :, ibatch] .= mels_list[imode, iscat] .* J_c_f_list[ind_el_f][:, :, c1, c2]
                    end
                end

                if mod(iscat, nbatch) == 0 || iscat == nscat
                    @views for c2 in 1:2, c1 in 1:2
                        # (N, 3*2*nbatch) <- (N, N*nmodes) * (N*nmodes, 3*2*nbatch)
                        C = Base.ReshapedArray(J_c_i_batch[c1, c2], (basis.N, 3*2*nbatch), ())
                        A = Base.ReshapedArray(D_c_modes[:, :, :, c2, c1], (basis.N, basis.N*nmodes), ())
                        B = Base.ReshapedArray(J_c_f_mel_batch[c1, c2], (basis.N*nmodes, 3*2*nbatch), ())
                        mul!(C, A, B)
                    end

                    iscat_from = div(iscat - 1, nbatch) * nbatch + 1
                    for (ibatch, iscat_batch) in enumerate(iscat_from : iscat)
                        ind_el_i = ind_el_i_list[iscat_batch]
                        @views for c2 in 1:2, c1 in 1:2
                            tmp_DJ_buffers[ichunk][ind_el_i][:, :, c1, c2] .+= J_c_i_batch[c1, c2][:, :, ibatch]
                        end
                    end
                end
            end
        end  # iq
    end  # iqs

    tmp_DJ_tmp = reduce(+, tmp_DJ_buffers)

    # Reinterpret a (basis.N, 3 * 2, 2, 2) Array of ComplexF64
    # back to a (basis.N, 2, 2, 2) Array of Vec3{ComplexF64}
    tmp_DJ = [reshape(reinterpret(Vec3{ComplexF64}, permutedims(reshape(x, basis.N, 3, 2, 2, 2), (2, 1, 3, 4, 5))), basis.N, 2, 2, 2) for x in tmp_DJ_tmp]

    tmp_DJ = permutedims.(tmp_DJ, Ref((1, 3, 4, 2)))

    J_new_c = zero.(J_c)
    for i in 1:el_i.n
        # (-1)^{δ_{c,+}} sign factor when attaching Green function to vertex
        tmp_DJ[i][:, 2, :, :] .*= -1
        tmp_DJ[i][:, :, 2, :] .*= -1

        @views for c0 in 1:2, c1 in 1:2, c2 in 1:2, cp1 in 1:2, cp2 in 1:2
            @. J_new_c[i][:, c1, c2, c0] += Gm_c[i][:, c1, cp1] * tmp_DJ[i][:, cp1, cp2, c0] * Gp_c[i][:, cp2, c2]
        end
    end

    J_new_c .*= -1

    J_new_c
end;

function get_D_map_delta(basis :: LinearSplineBasis, ω₀)
    D_c = [zeros(ComplexF64, basis.N, basis.N) for _ in 1:2]
    get_D_map_delta!(D_c, basis, ω₀)
end

function get_D_map_delta!(D_c, basis :: LinearSplineBasis, ω₀)
    for j in 1:basis.N, i in 1:basis.N
        ImD_emi = EPSpectral.get_D_map_imag(basis, +ω₀, i, j) / 2π
        ImD_abs = EPSpectral.get_D_map_imag(basis, -ω₀, i, j) / 2π

        D_c[1][i, j] = ImD_emi
        D_c[2][i, j] = ImD_abs
    end

    D_c
end

function setup_scKubo_D_irr(basis, ph, nmodes, model)
    qpts = GridKpoints(ElectronPhonon.Kpoints(ph.xks[1:nmodes:end]; ph.ngrid))
    qpts_irr, iq_to_iqirr_isym = ElectronPhonon.fold_kpoints(qpts, model.symmetry);
    qpts_irr

    ω_ph_irr = compute_eigenvalues_ph(model, qpts_irr)

    D_tmp = [zeros(basis.N, basis.N) for _ in 1:2]

    D_irr = map(Iterators.product(1:nmodes, 1:qpts_irr.n)) do (imode, iqirr)
        if imode == 1 && mod(iqirr, 2_000) == 0
            println("$(now()) iq = $iqirr / $(qpts_irr.n)")
            flush(stdout)
            flush(stderr)
        end
        ω_ph = ω_ph_irr[imode, iqirr]
        get_D_map_delta!(D_tmp, basis, ω_ph)
        sparse.(D_tmp)
    end

    (; D_irr, iq_to_iqirr_isym)
end

function setup_scKubo_D_irr_dense(basis, ph, nmodes, model, T)
    qpts = GridKpoints(ElectronPhonon.Kpoints(ph.xks[1:nmodes:end]; ph.ngrid))
    qpts_irr, iq_to_iqirr_isym = ElectronPhonon.fold_kpoints(qpts, model.symmetry);
    qpts_irr

    ω_ph_irr = compute_eigenvalues_ph(model, qpts_irr)

    D_tmp = [zeros(basis.N, basis.N) for _ in 1:2]

    D_irr = map(Iterators.product(1:nmodes, 1:qpts_irr.n)) do (imode, iqirr)
        if imode == 1 && mod(iqirr, 10) == 0
            println("$(now()) iq = $iqirr / $(qpts_irr.n)")
            flush(stdout)
            flush(stderr)
        end
        ω_ph = ω_ph_irr[imode, iqirr]
        EPSpectral.get_D_map(basis, ω_ph, T)
    end

    (; D_irr, iq_to_iqirr_isym)
end

function iterate_current_Dyson_new5(J_c, ik_and_isym_map, el_i, el_f, ph, basis, Gm_c, Gp_c, model, T, btedata_for_dyson)
    # TODO: The current btedata_for_dyson data outline is not optimal.
    # One would want the data to be stored as list of ind_el_i.
    # But, the curret data is stored as list of q points due to a historical reason.
    (; nmodes) = model

    nq = div(ph.n, nmodes)
    @assert ph.n == nq * nmodes

    # Distribute q points to threads with load balancing
    nchunks = 1 * nthreads()
    ind_el_i_chunks = collect(chunks(1:el_i.n; n = nchunks))

    tmp_DJ = [zeros(Vec3{ComplexF64}, basis.N, 6, 2) for _ in 1:el_i.n]

    # Unfold J_c to el_f
    J_c_f_list = [zeros(Vec3{ComplexF64}, basis.N, 2, 2, 2) for _ in 1:el_f.n]
    for ind_el_f in 1:el_f.n
        i, isym = ik_and_isym_map[ind_el_f]

        symop = model.symmetry[isym]
        if symop.is_tr
            J_c_f_list[ind_el_f] .= Ref(-symop.Scart) .* J_c[i]
        else
            J_c_f_list[ind_el_f] .= Ref(+symop.Scart) .* J_c[i]
        end
    end

    # Compute D_c for each mode using the irreducible q points
    D_c_irr, iq_to_iqirr_isym = setup_scKubo_D_irr(basis, ph, nmodes, model)

    @threads for ichunk in eachindex(ind_el_i_chunks)
    # @time for ichunk in 1:1

        for iq in 1:nq
            if threadid() == 1 && mod(iq, 2_000) == 0
                @info "$(now()) iq = $iq / $nq"
                flush(stdout)
                flush(stderr)
            end

            ind_el_i_list = btedata_for_dyson.ind_el_i_list[iq]
            ind_el_f_list = btedata_for_dyson.ind_el_f_list[iq]
            mels_list = btedata_for_dyson.mels_list[iq]
            nscat = length(ind_el_i_list)

            @views for iscat in 1:nscat
                ind_el_i = ind_el_i_list[iscat]
                ind_el_i ∈ ind_el_i_chunks[ichunk] || continue

                ind_el_f = ind_el_f_list[iscat]

                J_c_f = J_c_f_list[ind_el_f]
                tmp_DJ_i = tmp_DJ[ind_el_i]

                for imode in 1:nmodes, c0 in 1:2
                    mel = mels_list[imode, iscat]

                    ind_ph = (iq - 1) * nmodes + imode
                    ω_ph = ph.e[ind_ph]
                    nocc = occ_boson(ω_ph, T)

                    iqirr = iq_to_iqirr_isym[iq][1]
                    D_emi, D_abs = D_c_irr[imode, iqirr]

                    SparseArrays._spmatmul!(tmp_DJ_i[:, 1, c0], D_emi, J_c_f[:, 1, 1, c0], +mel, 1)
                    SparseArrays._spmatmul!(tmp_DJ_i[:, 1, c0], D_abs, J_c_f[:, 1, 1, c0], -mel, 1)
                    SparseArrays._spmatmul!(tmp_DJ_i[:, 2, c0], D_emi, J_c_f[:, 1, 1, c0], -(2nocc+1) * mel, 1)
                    SparseArrays._spmatmul!(tmp_DJ_i[:, 2, c0], D_abs, J_c_f[:, 1, 1, c0], -(2nocc+1) * mel, 1)
                    SparseArrays._spmatmul!(tmp_DJ_i[:, 3, c0], D_emi, J_c_f[:, 2, 2, c0], +mel, 1)
                    SparseArrays._spmatmul!(tmp_DJ_i[:, 3, c0], D_abs, J_c_f[:, 2, 2, c0], -mel, 1)
                    SparseArrays._spmatmul!(tmp_DJ_i[:, 4, c0], D_emi, J_c_f[:, 2, 2, c0], -(2nocc+1) * mel, 1)
                    SparseArrays._spmatmul!(tmp_DJ_i[:, 4, c0], D_abs, J_c_f[:, 2, 2, c0], -(2nocc+1) * mel, 1)
                    SparseArrays._spmatmul!(tmp_DJ_i[:, 5, c0], D_emi, J_c_f[:, 2, 1, c0], -2nocc * mel, 1)
                    SparseArrays._spmatmul!(tmp_DJ_i[:, 5, c0], D_abs, J_c_f[:, 2, 1, c0], -2*(nocc+1) * mel, 1)
                    SparseArrays._spmatmul!(tmp_DJ_i[:, 6, c0], D_emi, J_c_f[:, 1, 2, c0], -2*(nocc+1) * mel, 1)
                    SparseArrays._spmatmul!(tmp_DJ_i[:, 6, c0], D_abs, J_c_f[:, 1, 2, c0], -2nocc * mel, 1)
                end
            end
        end  # iq
    end  # iqs

    xs = basis.xs
    map_KK = [EPSpectral.linear_spline_kramers_kronig(xs[i+1], xs[j], xs[j+1], xs[j+2]) for i in 1:basis.N, j in 1:basis.N] ./ (im * π)

    DJ_c = [zeros(Vec3{ComplexF64}, basis.N, 2, 2, 2) for _ in 1:el_i.n]
    @views for i in 1:el_i.n, c0 in 1:2
        mul!(DJ_c[i][:, 1, 1, c0], map_KK, tmp_DJ[i][:, 1, c0], +1, 0)
        mul!(DJ_c[i][:, 2, 2, c0], map_KK, tmp_DJ[i][:, 3, c0], -1, 0)
        DJ_c[i][:, 1, 1, c0] .+= tmp_DJ[i][:, 2, c0]
        DJ_c[i][:, 2, 2, c0] .+= tmp_DJ[i][:, 4, c0]
        DJ_c[i][:, 2, 1, c0] .= tmp_DJ[i][:, 5, c0]
        DJ_c[i][:, 1, 2, c0] .= tmp_DJ[i][:, 6, c0]
    end

    J_new_c = zero.(J_c)
    for i in 1:el_i.n
        # (-1)^{δ_{c,+}} sign factor when attaching Green function to vertex
        DJ_c[i][:, 2, :, :] .*= -1
        DJ_c[i][:, :, 2, :] .*= -1

        @views for c0 in 1:2, c1 in 1:2, c2 in 1:2, cp1 in 1:2, cp2 in 1:2
            @. J_new_c[i][:, c1, c2, c0] += Gm_c[i][:, c1, cp1] * DJ_c[i][:, cp1, cp2, c0] * Gp_c[i][:, cp2, c2]
        end
    end

    J_new_c .*= -1

    J_new_c
end;


function iterate_current_Dyson_new6(J_c, ik_and_isym_map, el_i, el_f, ph, basis, Gm_c, Gp_c, model, T, filename_btedata_h5)
    (; nmodes) = model

    nq = div(ph.n, nmodes)
    @assert ph.n == nq * nmodes

    tmp_DJ = [zeros(Vec3{ComplexF64}, basis.N, 6, 2) for _ in 1:el_i.n]

    # Unfold J_c to el_f
    J_c_f_list = [zeros(Vec3{ComplexF64}, basis.N, 2, 2, 2) for _ in 1:el_f.n]
    for ind_el_f in 1:el_f.n
        i, isym = ik_and_isym_map[ind_el_f]

        symop = model.symmetry[isym]
        if symop.is_tr
            J_c_f_list[ind_el_f] .= Ref(-symop.Scart) .* J_c[i]
        else
            J_c_f_list[ind_el_f] .= Ref(+symop.Scart) .* J_c[i]
        end
    end

    # Compute D_c for each mode using the irreducible q points
    D_c_irr, iq_to_iqirr_isym = setup_scKubo_D_irr(basis, ph, nmodes, model)

    @threads for ind_el_i in 1:el_i.n

        if threadid() == 1 && mod(ind_el_i, 1) == 0
            @info "$(now()) ind_el_i = $ind_el_i / $(el_i.n)"
            flush(stdout)
            flush(stderr)
        end

        tmp_DJ_i = tmp_DJ[ind_el_i]

        ind_el_f_list, iq_list, mels_list = read_scKubo_btedata_hdf_new3_ind(filename_btedata_h5, ind_el_i)
        nscat = length(ind_el_f_list)

        @views for iscat in 1:nscat

            ind_el_f = ind_el_f_list[iscat]
            iq = iq_list[iscat]

            J_c_f = J_c_f_list[ind_el_f]

            for imode in 1:nmodes, c0 in 1:2
                mel = mels_list[imode, iscat]

                ind_ph = (iq - 1) * nmodes + imode
                ω_ph = ph.e[ind_ph]
                nocc = occ_boson(ω_ph, T)

                iqirr = iq_to_iqirr_isym[iq][1]
                D_emi, D_abs = D_c_irr[imode, iqirr]

                SparseArrays._spmatmul!(tmp_DJ_i[:, 1, c0], D_emi, J_c_f[:, 1, 1, c0], +mel, 1)
                SparseArrays._spmatmul!(tmp_DJ_i[:, 1, c0], D_abs, J_c_f[:, 1, 1, c0], -mel, 1)
                SparseArrays._spmatmul!(tmp_DJ_i[:, 2, c0], D_emi, J_c_f[:, 1, 1, c0], -(2nocc+1) * mel, 1)
                SparseArrays._spmatmul!(tmp_DJ_i[:, 2, c0], D_abs, J_c_f[:, 1, 1, c0], -(2nocc+1) * mel, 1)
                SparseArrays._spmatmul!(tmp_DJ_i[:, 3, c0], D_emi, J_c_f[:, 2, 2, c0], +mel, 1)
                SparseArrays._spmatmul!(tmp_DJ_i[:, 3, c0], D_abs, J_c_f[:, 2, 2, c0], -mel, 1)
                SparseArrays._spmatmul!(tmp_DJ_i[:, 4, c0], D_emi, J_c_f[:, 2, 2, c0], -(2nocc+1) * mel, 1)
                SparseArrays._spmatmul!(tmp_DJ_i[:, 4, c0], D_abs, J_c_f[:, 2, 2, c0], -(2nocc+1) * mel, 1)
                SparseArrays._spmatmul!(tmp_DJ_i[:, 5, c0], D_emi, J_c_f[:, 2, 1, c0], -2nocc * mel, 1)
                SparseArrays._spmatmul!(tmp_DJ_i[:, 5, c0], D_abs, J_c_f[:, 2, 1, c0], -2*(nocc+1) * mel, 1)
                SparseArrays._spmatmul!(tmp_DJ_i[:, 6, c0], D_emi, J_c_f[:, 1, 2, c0], -2*(nocc+1) * mel, 1)
                SparseArrays._spmatmul!(tmp_DJ_i[:, 6, c0], D_abs, J_c_f[:, 1, 2, c0], -2nocc * mel, 1)
            end
        end  # iscat
    end  # ind_el_i

    xs = basis.xs
    map_KK = [EPSpectral.linear_spline_kramers_kronig(xs[i+1], xs[j], xs[j+1], xs[j+2]) for i in 1:basis.N, j in 1:basis.N] ./ (im * π)

    DJ_c = [zeros(Vec3{ComplexF64}, basis.N, 2, 2, 2) for _ in 1:el_i.n]
    @views for i in 1:el_i.n, c0 in 1:2
        mul!(DJ_c[i][:, 1, 1, c0], map_KK, tmp_DJ[i][:, 1, c0], +1, 0)
        mul!(DJ_c[i][:, 2, 2, c0], map_KK, tmp_DJ[i][:, 3, c0], -1, 0)
        DJ_c[i][:, 1, 1, c0] .+= tmp_DJ[i][:, 2, c0]
        DJ_c[i][:, 2, 2, c0] .+= tmp_DJ[i][:, 4, c0]
        DJ_c[i][:, 2, 1, c0] .= tmp_DJ[i][:, 5, c0]
        DJ_c[i][:, 1, 2, c0] .= tmp_DJ[i][:, 6, c0]
    end

    J_new_c = zero.(J_c)
    for i in 1:el_i.n
        # (-1)^{δ_{c,+}} sign factor when attaching Green function to vertex
        DJ_c[i][:, 2, :, :] .*= -1
        DJ_c[i][:, :, 2, :] .*= -1

        @views for c0 in 1:2, c1 in 1:2, c2 in 1:2, cp1 in 1:2, cp2 in 1:2
            @. J_new_c[i][:, c1, c2, c0] += Gm_c[i][:, c1, cp1] * DJ_c[i][:, cp1, cp2, c0] * Gp_c[i][:, cp2, c2]
        end
    end

    J_new_c .*= -1

    J_new_c
end;


"""
Perform the convolution of D matrix with the current vertex J.

Usage:
```
map_KK = EPSpectral.get_Kramers_Kronig_map(basis)
tmp_DJ = zeros(ComplexF64, basis.N, 6)
D_convolution_step1!(tmp_DJ, D_emi, D_abs, F, nq, mel)
DJ = zeros(ComplexF64, basis.N, 2, 2)
D_convolution_step2!(DJ, map_KK, tmp_DJ)
``

# This should give the same result using dense D matrix as
```
D = EPSpectral.get_D_map(basis, ω_ph, T)
DJ = zeros(ComplexF64, basis.N, 2, 2)
@views for c2 in 1:2, c1 in 1:2
    mul!(DJ[:, c1, c2], D[c2, c1], F[:, c1, c2], mel, 0)
end
```
"""
@views function D_convolution_step1!(tmp_DJ, D_emi, D_abs, F, nq, mel)
    SparseArrays._spmatmul!(tmp_DJ[:, 1], D_emi, F[:, 1, 1], +mel, 1)
    SparseArrays._spmatmul!(tmp_DJ[:, 1], D_abs, F[:, 1, 1], -mel, 1)
    SparseArrays._spmatmul!(tmp_DJ[:, 2], D_emi, F[:, 1, 1], -(2nq+1) * mel, 1)
    SparseArrays._spmatmul!(tmp_DJ[:, 2], D_abs, F[:, 1, 1], -(2nq+1) * mel, 1)
    SparseArrays._spmatmul!(tmp_DJ[:, 3], D_emi, F[:, 2, 2], +mel, 1)
    SparseArrays._spmatmul!(tmp_DJ[:, 3], D_abs, F[:, 2, 2], -mel, 1)
    SparseArrays._spmatmul!(tmp_DJ[:, 4], D_emi, F[:, 2, 2], -(2nq+1) * mel, 1)
    SparseArrays._spmatmul!(tmp_DJ[:, 4], D_abs, F[:, 2, 2], -(2nq+1) * mel, 1)
    SparseArrays._spmatmul!(tmp_DJ[:, 5], D_emi, F[:, 2, 1], -2nq * mel, 1)
    SparseArrays._spmatmul!(tmp_DJ[:, 5], D_abs, F[:, 2, 1], -2*(nq+1) * mel, 1)
    SparseArrays._spmatmul!(tmp_DJ[:, 6], D_emi, F[:, 1, 2], -2*(nq+1) * mel, 1)
    SparseArrays._spmatmul!(tmp_DJ[:, 6], D_abs, F[:, 1, 2], -2nq * mel, 1)
    tmp_DJ
end

@views function D_convolution_step2!(DJ, map_KK, tmp_DJ)
    # map_KK performs Kramers-Kronig transformation without 1/π factor, so we divide here.
    mul!(DJ[:, 1, 1], map_KK, tmp_DJ[:, 1], -im / π, 1)
    mul!(DJ[:, 2, 2], map_KK, tmp_DJ[:, 3], +im / π, 1)
    DJ[:, 1, 1] .+= tmp_DJ[:, 2]
    DJ[:, 2, 2] .+= tmp_DJ[:, 4]
    DJ[:, 2, 1] .+= tmp_DJ[:, 5]
    DJ[:, 1, 2] .+= tmp_DJ[:, 6]
    DJ
end

function iterate_current_Dyson_new7(J, ik_and_isym_map, el_i, el_f, ph, basis, Gm, Gp, model, T, filename_btedata_h5)
    (; nmodes) = model

    nq = div(ph.n, nmodes)
    @assert ph.n == nq * nmodes

    tmp_DJ = [zeros(Vec3{ComplexF64}, basis.N, 6, 2) for _ in 1:el_i.n]

    # Unfold J to el_f
    J_f_list = [zeros(Vec3{ComplexF64}, basis.N, 2, 2, 2) for _ in 1:el_f.n]
    for ind_el_f in 1:el_f.n
        i, isym = ik_and_isym_map[ind_el_f]

        symop = model.symmetry[isym]
        if symop.is_tr
            J_f_list[ind_el_f] .= Ref(-symop.Scart) .* J[i]
        else
            J_f_list[ind_el_f] .= Ref(+symop.Scart) .* J[i]
        end
    end

    # Compute D_c for each mode using the irreducible q points
    D_c_irr, iq_to_iqirr_isym = setup_scKubo_D_irr(basis, ph, nmodes, model)

    @threads for ind_el_i in 1:el_i.n

        if threadid() == 1 && mod(ind_el_i, 1) == 0
            @info "$(now()) ind_el_i = $ind_el_i / $(el_i.n)"
            flush(stdout)
            flush(stderr)
        end

        tmp_DJ_i = tmp_DJ[ind_el_i]

        (; ind_el_f_list, iq_list, g_list, dg_list) = scKubo_read_matrix_elements(filename_btedata_h5, ind_el_i)
        nscat = length(ind_el_f_list)

        @views for iscat in 1:nscat

            ind_el_f = ind_el_f_list[iscat]
            el_kq = el_f[ind_el_f]
            iq = iq_list[iscat]

            J_f = J_f_list[ind_el_f]

            for imode in 1:nmodes
                ind_ph = (iq - 1) * nmodes + imode
                ωq = ph.e[ind_ph]
                ωq < ElectronPhonon.omega_acoustic && continue

                mel = abs2(g_list[imode, iscat]) / (2ωq) * el_kq.k_weight
                nq = occ_boson(ωq, T)

                iqirr = iq_to_iqirr_isym[iq][1]
                D_emi, D_abs = D_c_irr[imode, iqirr]

                @views for c0 in 1:2
                    D_convolution_step1!(tmp_DJ_i[:, :, c0], D_emi, D_abs, J_f[:, :, :, c0], nq, mel)
                end
            end
        end  # iscat
    end  # ind_el_i

    map_KK = EPSpectral.get_Kramers_Kronig_map(basis)

    DJ = [zeros(Vec3{ComplexF64}, basis.N, 2, 2, 2) for _ in 1:el_i.n]
    @views for i in 1:el_i.n, c0 in 1:2
        D_convolution_step2!(DJ[i][:, :, :, c0], map_KK, tmp_DJ[i][:, :, c0])
    end

    for i in 1:el_i.n
        # (-1)^{δ_{c,+}} sign factor when attaching Green function to vertex
        DJ[i][:, 2, :, :] .*= -1
        DJ[i][:, :, 2, :] .*= -1
    end

    J_new = scKubo_attach_propagator(DJ, Gm, Gp)
    J_new .*= -1

    J_new
end;



function read_scKubo_btedata_hdf(filename, nq)
    ind_el_i_list = Vector{Int}[]
    ind_el_f_list = Vector{Int}[]
    mels_list = Matrix{Float64}[]

    f = h5open(filename, "r")
    for iq in 1:nq
        if mod(iq, 10_000) == 0
            @info "$(now()) iq = $iq / $nq"
            flush(stdout)
            flush(stderr)
            GC.gc()
        end

        push!(ind_el_i_list, read(f["ind_el_i/iq$iq"], Int))
        push!(ind_el_f_list, read(f["ind_el_f/iq$iq"], Int))
        push!(mels_list, read(f["mels/iq$iq"], Float64))
    end
    close(f)

    (; ind_el_i_list, ind_el_f_list, mels_list)
end

function read_scKubo_btedata_hdf_new3_ind(filename, ind_el_i)
    f = h5open(filename, "r")
    ind_el_f_list = read(f["ind_el_f/ind_el_i.$ind_el_i"], Int) :: Vector{Int}
    iq_list = read(f["iq/ind_el_i.$ind_el_i"], Int) :: Vector{Int}
    mels_list = read(f["mels/ind_el_i.$ind_el_i"], Float64) :: Matrix{Float64}
    close(f)

    (; ind_el_f_list, iq_list, mels_list)
end

function scKubo_read_matrix_elements(filename, ind_el_i)
    # Read HDF5 file computed using SCKuboCovDerDataCalculator
    f = h5open(filename, "r")
    ind_el_f_list = read(f["ind_el_f/ind_el_i.$ind_el_i"], Int) :: Vector{Int}
    iq_list = read(f["iq/ind_el_i.$ind_el_i"], Int) :: Vector{Int}
    g_list = read(f["g/ind_el_i.$ind_el_i"], ComplexF64) :: Matrix{ComplexF64}
    dg_list_ = read(f["dg/ind_el_i.$ind_el_i"], ComplexF64) :: Matrix{ComplexF64}
    close(f)

    dg_list = Matrix(reinterpret(Vec3{ComplexF64}, dg_list_))

    return (; ind_el_f_list, iq_list, g_list, dg_list)
end

function read_scKubo_btedata_hdf_new3(filename, num_el_i)
    ind_el_f_list = Vector{Int}[]
    iq_list = Vector{Int}[]
    mels_list = Matrix{Float64}[]

    f = h5open(filename, "r")
    for i in 1:num_el_i
        if mod(i, 1_000) == 0
            @info "$(now()) ind_el_i = $i / $num_el_i"
            flush(stdout)
            flush(stderr)
            GC.gc()
        end

        push!(ind_el_f_list, read(f, "ind_el_f/ind_el_i.$i"))
        push!(iq_list, read(f, "iq/ind_el_i.$i"))
        push!(mels_list, read(f, "mels/ind_el_i.$i"))
    end
    close(f)

    (; ind_el_f_list, iq_list, mels_list)
end

# function iterate_current_Dyson(J_c, filename_btedata, ik_and_isym_map, el_i, el_f, ph, basis, Gm_c, Gp_c, model, T, kpts, kqpts)
#     fid = h5open(filename_btedata, "r")

#     nbands_of_ikq = [count(el_f.xks .≈ Ref(kq)) for kq in kqpts.vectors]
#     @assert sum(nbands_of_ikq) == el_f.n
#     max_nbands = maximum(nbands_of_ikq)

#     tmp_DJ = zero.(J_c)
#     J_new_c = zero.(J_c)

#     group_scattering = open_group(fid, "scattering")
#     mpi_isroot() && println("Original grid: Total $(length(group_scattering)) groups of scattering")

#     @threads :static for ind_el_i in 1:el_i.n
#         if threadid() == 1
#             @info "$(now()) ind_el_i = $ind_el_i / $(el_i.n)"
#             flush(stdout)
#             flush(stderr)
#         end

#         e_i = el_i[ind_el_i]
#         ik = xk_to_ik(e_i.xks, kpts)
#         key = "ik$ik"

#         # number of bands for the same k-point
#         ibands_of_ik = el_i.iband[filter(i -> el_i[i].xks ≈ el_i[ind_el_i].xks, 1:el_i.n)]
#         nband_of_ik = length(ibands_of_ik)
#         my_iband_ind = findfirst(ibands_of_ik .== el_i[ind_el_i].iband)

#         g = open_group(group_scattering, key)
#         scat = load_BTData(g, ElectronPhonon.ElPhScatteringData{Float64})

#         # Indices of scat: [sign_ph, el_i_band, el_f_iband, imode, ikq]
#         # We only consider the case of sign_ph == 1 and el_i_band == iband
#         @assert length(scat) == 2 * nband_of_ik * el_f.n * model.nmodes
#         iscat_selected = reshape(1:length(scat), 2, nband_of_ik, el_f.n * model.nmodes)[1, my_iband_ind, :]

#         D_c = [zeros(ComplexF64, basis.N, basis.N, 2, 2) for _ in 1:max_nbands]
#         D_c_imode = [zeros(ComplexF64, basis.N, basis.N) for _ in 1:2, _ in 1:2]
#         D_map_buffer = zeros(length(basis.xs))

#         J_c_f = zeros(Vec3{ComplexF64}, basis.N, 2, 2, 2)

#         ω_ph_prev = NaN

#         @views for ikq in 1:kqpts.n

#             for x in D_c
#                 x .= 0
#             end

#             # Compute D_c for each jbnd by summing over phonon modes

#             for imode in 1:model.nmodes
#                 for jbnd in 1:nbands_of_ikq[ikq]

#                     ind_el_f = sum(nbands_of_ikq[1:ikq-1]) + jbnd
#                     iscat = iscat_selected[sum(nbands_of_ikq[1:ikq-1]) * model.nmodes + (imode - 1) * nbands_of_ikq[ikq] + jbnd]

#                     s = scat[iscat]
#                     (; ind_ph) = s

#                     @assert s.sign_ph == -1
#                     @assert s.ind_el_i == ind_el_i
#                     @assert s.ind_el_f == ind_el_f
#                     @assert ph.iband[ind_ph] == imode

#                     e_f = el_f[ind_el_f]
#                     ω_ph = ph.e[ind_ph]
#                     g2 = s.mel

#                     ω_ph < ElectronPhonon.omega_acoustic && continue

#                     if !(ω_ph ≈ ω_ph_prev)
#                         EPSpectral.get_D_map_opt!(D_c_imode, basis, ω_ph, T, D_map_buffer)
#                         ω_ph_prev = ω_ph
#                     end

#                     @views for c2 in 1:2, c1 in 1:2
#                         D_c[jbnd][:, :, c1, c2] .+= D_c_imode[c1, c2] .* e_f.k_weight .* g2
#                     end
#                 end
#             end

#             # Multiply D_c to J_c_f at (k+q, jbnd)

#             for jbnd in 1:nbands_of_ikq[ikq]
#                 ind_el_f = sum(nbands_of_ikq[1:ikq-1]) + jbnd

#                 i, isym = ik_and_isym_map[ind_el_f]
#                 symop = model.symmetry[isym]
#                 if symop.is_tr
#                     J_c_f .= Ref(-symop.Scart) .* J_c[i]
#                 else
#                     J_c_f .= Ref(+symop.Scart) .* J_c[i]
#                 end

#                 @views for c0 in 1:2, c2 in 1:2, c1 in 1:2
#                     mul!(tmp_DJ[ind_el_i][:, c1, c2, c0], D_c[jbnd][:, :, c2, c1], J_c_f[:, c1, c2, c0], 1, 1)
#                 end
#             end
#         end
#     end
#     close(fid)

#     for i in 1:el_i.n
#         # (-1)^{δ_{c,+}} sign factor when attaching Green function to vertex
#         tmp_DJ[i][:, 2, :, :] .*= -1
#         tmp_DJ[i][:, :, 2, :] .*= -1

#         @views for c0 in 1:2, c1 in 1:2, c2 in 1:2, cp1 in 1:2, cp2 in 1:2
#             @. J_new_c[i][:, c1, c2, c0] += Gm_c[i][:, c1, cp1] * tmp_DJ[i][:, cp1, cp2, c0] * Gp_c[i][:, cp2, c2]
#         end
#     end

#     J_new_c .*= -1

#     J_new_c
# end;



function scKubo_trim_Σs_to_BTStates(Σs :: Array{ComplexF64, 4}, el_i, kpts)
    # Σs uses band and k point indexing (iw, ib, ik, iT)
    # Trim Σs to the bands and k points of el_i
    @assert size(Σs, 3) == kpts.n
    Σs_out = zeros(eltype(Σs), size(Σs, 1), el_i.n, size(Σs, 4))
    @views for i in 1:el_i.n
        ib = el_i[i].iband
        ik = xk_to_ik(el_i[i].xks, kpts)
        Σs_out[:, i, :] .= Σs[:, ib, ik, :]
    end
    Σs_out
end

function scKubo_trim_Σs_rest_to_BTStates(Σs_rest :: Array{ComplexF64, 3}, el_i, kpts)
    # Σs_rest uses band and k point indexing (ik, ib, iT)
    # Trim Σs to the bands and k points of el_i
    @assert size(Σs_rest, 1) == kpts.n
    Σs_out = zeros(eltype(Σs_rest), el_i.n, size(Σs_rest, 3))
    @views for i in 1:el_i.n
        ib = el_i[i].iband
        ik = xk_to_ik(el_i[i].xks, kpts)
        Σs_out[i, :] .= Σs_rest[ik, ib, :]
    end
    Σs_out
end


function scKubo_setup_Green(el_i :: ElectronPhonon.BTStates, basis, ωs, Σs :: Array{ComplexF64, 3}, occ, Ω, iT)
    # Setup Gm and Gp for an ab initio calculation
    (; μ, T) = occ[iT]

    Gm = [zeros(ComplexF64, basis.N, 2, 2) for _ in 1:el_i.n]
    Gp = [zeros(ComplexF64, basis.N, 2, 2) for _ in 1:el_i.n]

    Σk_itp = extrapolate(interpolate!((ωs,), zeros(ComplexF64, length(ωs)), Gridded(Linear())), Flat())

    for i in 1:el_i.n
        ek = el_i[i].e
        parent(parent(Σk_itp)) .= Σs[:, i, iT]

        @views for (iω, ω) in enumerate(grid_points(basis))
            ωm = ω - Ω/2
            ωp = ω + Ω/2
            Gm[i][iω, :, :] .= EPSpectral.get_G_contour(ωm, ek, Σk_itp(ωm), μ, T)
            Gp[i][iω, :, :] .= EPSpectral.get_G_contour(ωp, ek, Σk_itp(ωp), μ, T)
        end
    end
    (; Gm, Gp)
end



function compute_phonon_assisted_current_vertex(Gm, Gp, basis, filename_mel, el_i, el_f, ik_and_isym_map, ph)
    D_irr, iq_to_iqirr_isym = setup_scKubo_D_irr(basis, ph, model.nmodes, model);
    map_KK = EPSpectral.get_Kramers_Kronig_map(basis)

    tmap(1:el_i.n) do ind_el_i

        if threadid() == 1
            @info "$(now()) ind_el_i = $ind_el_i / $(el_i.n)"
            flush(stdout)
            flush(stderr)
        end

        (; ind_el_f_list, iq_list, g_list, dg_list) = scKubo_read_matrix_elements(filename_mel, ind_el_i)
        nscat = length(ind_el_f_list)
        nmodes = size(g_list, 1)

        tmp_DJ_1 = zeros(Vec3{ComplexF64}, basis.N, 6)
        tmp_DJ_2 = zeros(Vec3{ComplexF64}, basis.N, 6)
        V0_ep_k_1 = zeros(Vec3{ComplexF64}, basis.N, 2, 2)
        V0_ep_k_2 = zeros(Vec3{ComplexF64}, basis.N, 2, 2)

        @views for iscat in 1:nscat
            ind_el_f = ind_el_f_list[iscat]
            el_kq = el_f[ind_el_f]
            iq = iq_list[iscat]

            ind_i_of_f = ik_and_isym_map[ind_el_f][1]

            for imode in 1:nmodes
                ind_ph = (iq - 1) * nmodes + imode
                ωq = ph.e[ind_ph]
                ωq < ElectronPhonon.omega_acoustic && continue

                nq = occ_boson(ωq, T)
                g = g_list[imode, iscat]
                dg = dg_list[imode, iscat]

                factor_m = conj(g) * dg * el_kq.k_weight / (2ωq)
                factor_p = g * conj(dg) * el_kq.k_weight / (2ωq)

                iqirr = iq_to_iqirr_isym[iq][1]
                D_emi, D_abs = D_irr[imode, iqirr]

                # Convolution step 1 : Frequency shift
                D_convolution_step1!(tmp_DJ_1, D_emi, D_abs, Gm[ind_i_of_f], nq, factor_m)
                D_convolution_step1!(tmp_DJ_2, D_emi, D_abs, Gp[ind_i_of_f], nq, factor_p)
            end
        end

        # Convolution step 2 : Kramers-Kronig
        D_convolution_step2!(V0_ep_k_1, map_KK, tmp_DJ_1)
        D_convolution_step2!(V0_ep_k_2, map_KK, tmp_DJ_2)

        V0_ep_k = zeros(Vec3{ComplexF64}, basis.N, 2, 2, 2)
        @views for c2 in 1:2, c1 in 1:2
            V0_ep_k[:, c1, c2, c2] .+= V0_ep_k_1[:, c1, c2]
            V0_ep_k[:, c1, c2, c1] .+= V0_ep_k_2[:, c1, c2]
        end

        # (-1)^{δ_{c,+}} sign factor in e-ph vertex
        @views V0_ep_k[:, 2, :, :] .*= -1
        @views V0_ep_k[:, :, 2, :] .*= -1

        # (-1)^{δ_{c,+}} sign factor in current vertex
        # FIXME: One should include this in bare current vertex and remove it from here.
        @views V0_ep_k[:, :, :, 2] .*= -1

        # Overall -1 sign (from self-energy loop)
        V0_ep_k .*= -1

        V0_ep_k
    end
end;
