using OhMyThreads
using ChunkSplitters

function compute_phonon_assisted_bubble(Ω, occ, basis, model, el_i, el_f, ph, filename_mel, ωs, Σs, ik_and_isym_map; truncate_spectral = false)

    if truncate_spectral
        error("implementation is wrong. must set to 0 whenever argument of A exceeds box, not Fermi")
    end

    # Phonon-assisted bubble
    ωgrid = grid_points(basis)
    dω = ωgrid[2] - ωgrid[1]
    @assert all(diff(ωgrid) .≈ dω)

    # Cutoff energies to truncate spectral function
    emin_list = fill(-Inf, length(occ))
    emax_list = fill(+Inf, length(occ))
    if truncate_spectral
        for iT in 1:length(occ)
            emin, emax = spectral_window_for_semiconductors(ωs, Σs, occ, el_i, iT)
            emin_list[iT] = emin
            emax_list[iT] = emax
        end
    end

    σ = zeros(ComplexF64, 3, 3, length(occ))

    for ind_el_i in 1:el_i.n
        if mod(ind_el_i, 50) == 0
            println("ind_el_i = $ind_el_i / $(el_i.n)")
        end

        el = el_i[ind_el_i]
        ek = el.e

        # Read matrix elements from file
        (; ind_el_f_list, iq_list, dg_list) = scKubo_read_matrix_elements(filename_mel, ind_el_i)

        # Precompute spectral function at k
        Ak = zeros(length(ωgrid), length(occ))
        @views for iT in 1:length(occ)
            Σk = Σs[:, ind_el_i, iT]
            Σk_itp = extrapolate(interpolate!((ωs,), Σk, Gridded(Linear())), Flat())
            for (iω, ω) in enumerate(ωgrid)
                Ak[iω, iT] = -imag(1 / (ω - ek - Σk_itp(ω))) / π
            end

            if truncate_spectral
                Ak[ωgrid .< emin_list[iT], iT] .= 0
                Ak[ωgrid .> emax_list[iT], iT] .= 0
            end
        end

        iscat_list_chunks = collect(chunks(1:length(ind_el_f_list); n = Threads.nthreads(), split = :batch))

        σ_k_list = zeros(ComplexF64, 3, 3, length(occ), length(iscat_list_chunks))

        Threads.@threads for ichunk in eachindex(iscat_list_chunks)
            iscat_list = iscat_list_chunks[ichunk]

            σ_k_chunk = zeros(ComplexF64, 3, 3, length(occ))
            Σkq_itp = extrapolate(interpolate!((ωs,), zeros(ComplexF64, length(ωs)), Gridded(Linear())), Flat())

            @views for iscat in iscat_list
                ind_el_f = ind_el_f_list[iscat]
                iq = iq_list[iscat]

                el_kq = el_f[ind_el_f]
                ekq = el_kq.e
                ind_el_i_of_f = ik_and_isym_map[ind_el_f][1]

                for (iT, (; μ, T)) in enumerate(occ)
                    parent(parent(Σkq_itp)) .= Σs[:, ind_el_i_of_f, iT]

                    for imode in 1:model.nmodes
                        ind_ph = (iq - 1) * model.nmodes + imode
                        ωq = ph[ind_ph].e
                        ωq < ElectronPhonon.omega_acoustic && continue

                        dg = dg_list[imode, iscat]
                        dg² = (dg * dg') / (2ωq)

                        σ_kq_tmp = _phonon_assisted_bubble_inner(Ω, ωgrid, ωq, μ, T, ekq, Σkq_itp, view(Ak, :, iT), emin_list[iT], emax_list[iT])

                        σ_k_chunk[:, :, iT] .+= σ_kq_tmp * π * dω * dg² * el_kq.k_weight
                    end
                end  # iscat
            end  # iT

            σ_k_list[:, :, :, ichunk] .= σ_k_chunk
        end  # iscat chunks

        σ_k = dropdims(sum(σ_k_list, dims = 4), dims = 4)

        σ .+= σ_k * el.k_weight
    end

    # Symmetrize conductivity tensor
    σ = symmetrize_array(σ, model.symmetry; order = 2)

    σ .*= occ.spin_degeneracy ./ model.volume

    return real.(σ)
end

function _phonon_assisted_bubble_inner(Ω, ωgrid, ωq, μ, T, ekq, Σkq_itp, Ak, emin = -Inf, emax = Inf)
    out = 0.0
    nq = occ_boson(ωq, T)

    for (iω, ω) in enumerate(ωgrid)
        f = occ_fermion(ω - μ, T)

        for s in (+1, -1)
            emin <= ω - Ω + s*ωq <= emax || continue  # Truncate spectral function outside [emin, emax]

            if Ω == 0
                df = -docc_fermion(ω + s*ωq - μ, T)
            else
                f1 = occ_fermion(ω + s*ωq - Ω - μ, T)
                f2 = occ_fermion(ω + s*ωq - μ, T)
                df = (f1 - f2) / Ω
            end

            if s == +1
                f_ph = nq + 1 - f
            else
                f_ph = nq + f
            end

            # Manual calculation using real and imaginary parts for efficiency
            Σkq_r, Σkq_i = reim(Σkq_itp(ω - Ω + s*ωq))
            denom = (ω - Ω + s*ωq - ekq - Σkq_r)^2 + Σkq_i^2
            Akq = -Σkq_i / (denom * π)

            out += df * f_ph * Ak[iω] * Akq
        end
    end

    out
end




"""
- `truncate_spectral` : Truncate the spectral function outside the window. Used for one-shot FM
calculations for lightly doped semiconductors. Can be computed using `spectral_window_for_semiconductors`.
"""
function compute_conductivity_bubble(Ω, occ, basis, model, el_i :: ElectronPhonon.BTStates, ωs, Σs; truncate_spectral = false)

    ωgrid = grid_points(basis)
    # Basis factor for ∫ dν bᵢ(ν)
    basis_sum = @. (basis.xs[3:end] .- basis.xs[1:end-2]) / 2

    dfermi = zeros(length(ωgrid))
    σks = [zeros(3, 3, length(occ)) for _ in 1:el_i.n]
    jdosks = [zeros(length(occ)) for _ in 1:el_i.n]

    @views for (iT, (; μ, T)) in enumerate(occ)

        if Ω == 0
            @. dfermi = -docc_fermion(ωgrid - μ, T)
        else
            @. dfermi = (occ_fermion(ωgrid - Ω/2 - μ, T) - occ_fermion(ωgrid + Ω/2 - μ, T)) / Ω
        end

        if truncate_spectral
            emin, emax = spectral_window_for_semiconductors(ωs, Σs, occ, el_i, iT)
            dfermi[ωgrid .- Ω/2 .< emin] .= 0
            dfermi[ωgrid .- Ω/2 .> emax] .= 0
            dfermi[ωgrid .+ Ω/2 .< emin] .= 0
            dfermi[ωgrid .+ Ω/2 .> emax] .= 0
        end

        for ind in 1:el_i.n

            el = el_i[ind]
            ek = el.e
            vk = el.vdiag

            Σ_itp = extrapolate(interpolate!((ωs,), Σs[:, ind, iT], Gridded(Linear())), Flat())

            σk_int = 0.0
            @views for (iω, ω) in enumerate(ωgrid)
                ωm = ω - Ω/2
                ωp = ω + Ω/2
                Am = -imag(1 / (ωm - ek - Σ_itp(ωm))) / π
                Ap = -imag(1 / (ωp - ek - Σ_itp(ωp))) / π
                σk_int += Am * Ap * dfermi[iω] * basis_sum[iω] * π
            end

            # Get a 3*3 conductivity tensor
            σk = real.(vk * vk') * σk_int

            σks[ind][:, :, iT] .= σk * el.k_weight
            jdosks[ind][iT] = σk_int * el.k_weight

        end
    end

    σ = sum(σks)
    jdos = sum(jdosks)

    # Symmetrize conductivity tensor
    σ = symmetrize_array(σ, model.symmetry; order = 2)

    σ .*= occ.spin_degeneracy ./ model.volume
    jdos .*= occ.spin_degeneracy ./ model.volume

    (; σ, jdos, σks, jdosks)
end


function compute_occupation(occ, basis, els :: ElectronPhonon.BTStates, ωs, Σs; truncate_spectral = false)

    ωgrid = grid_points(basis)
    # Basis factor for ∫ dν bᵢ(ν)
    basis_sum = @. (basis.xs[3:end] .- basis.xs[1:end-2]) / 2

    fermi = zeros(length(ωgrid))
    nlist = zeros(length(occ))

    @views for (iT, (; μ, T)) in enumerate(occ)

        if occ.type === :Semiconductor && occ.nlist[iT] < 0
            # Count holes instead of electrons
            count_holes = true
            @. fermi = occ_fermion(μ - ωgrid, T)
        else
            count_holes = false
            @. fermi = occ_fermion(ωgrid - μ, T)
        end

        if truncate_spectral
            emin, emax = spectral_window_for_semiconductors(ωs, Σs, occ, els, iT)
            fermi[ωgrid .< emin] .= 0
            fermi[ωgrid .> emax] .= 0
        end

        n = mapreduce(+, 1:els.n) do ind
            el = els[ind]
            ek = el.e

            Σ_itp = extrapolate(interpolate!((ωs,), Σs[:, ind, iT], Gridded(Linear())), Flat())

            nk_int = 0.0
            @views for (iω, ω) in enumerate(ωgrid)
                A = -imag(1 / (ω - ek - Σ_itp(ω))) / π
                nk_int += A * fermi[iω] * basis_sum[iω]
            end
            nk_int * el.k_weight
        end

        if count_holes
            n = sum(els.k_weight) - n + els.nstates_base
        else
            n = n + els.nstates_base
        end

        nlist[iT] = n
    end

    @. nlist = nlist * occ.spin_degeneracy - occ.nelec

    nlist
end
